"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * Copyright 2016 Simon Edwards <simon@simonzone.com>
 */
const CodeMirror = require("codemirror");
/**
 * Command to extend the selection vertically in both directions as multiple cursors.
 *
 * @param cm the CodeMirror instance to operate on
 */
function verticalMultiCursor(cm) {
    const doc = cm.getDoc();
    const cursorAnchor = doc.getCursor('anchor');
    const cursorHead = doc.getCursor('head');
    const anchorCol = Math.min(cursorHead.ch, cursorAnchor.ch);
    const headCol = Math.max(cursorHead.ch, cursorAnchor.ch);
    const newSelections = [];
    const wholeHeight = cursorAnchor.line === cursorHead.line;
    const start = wholeHeight ? 0 : Math.min(cursorAnchor.line, cursorHead.line);
    const end = wholeHeight ? doc.lineCount() : Math.max(cursorAnchor.line, cursorHead.line) + 1;
    let primaryIndex = 0;
    for (let i = start; i < end; i++) {
        const lineStr = doc.getLine(i);
        if (lineStr.length >= anchorCol) {
            newSelections.push({
                anchor: { line: i, ch: anchorCol },
                head: { line: i, ch: Math.min(lineStr.length, headCol) }
            });
            if (i === cursorHead.line) {
                primaryIndex = newSelections.length - 1;
            }
        }
    }
    doc.setSelections(newSelections, primaryIndex);
}
exports.COMMAND_VERTICAL_MULTICURSOR = "verticalMultiCursor";
CodeMirror.commands[exports.COMMAND_VERTICAL_MULTICURSOR] = verticalMultiCursor;
/**
 * Replaces text in the CodeMirror selection using a RegExp.
 *
 * @param cm the CodeMirror instance to operate on
 * @param searchExp the regular expression matching the strings to replace
 * @param replacement the replacement string
 */
function replaceInSelection(cm, searchExp, replacement) {
    const selectionsText = cm.getDoc().getSelections();
    const replacements = selectionsText.map((str) => {
        return str.replace(searchExp, replacement);
    });
    cm.getDoc().replaceSelections(replacements, "around");
}
/**
 * Command to replace forward slashes with back slashes in the selection.
 *
 * @param cm the CodeMirror instance to operate on
 */
function forwardSlashesToBack(cm) {
    replaceInSelection(cm, /\//g, "\\");
}
exports.COMMAND_FORWARDSLASHES_TO_BACK = "forwardSlashesToBack";
CodeMirror.commands[exports.COMMAND_FORWARDSLASHES_TO_BACK] = forwardSlashesToBack;
/**
 * Command to replace back slashes with forward slashes in the selection.
 *
 * @param cm the CodeMirror instance to operate on
 */
function backSlashesToForward(cm) {
    replaceInSelection(cm, /\\/g, "/");
}
exports.COMMAND_BACKSLASHES_TO_FORWARD = "backSlashesToForward";
CodeMirror.commands[exports.COMMAND_BACKSLASHES_TO_FORWARD] = backSlashesToForward;
/**
 * Command to backslash escape possible shell special characters in the selection.
 *
 * @param cm the CodeMirror instance to operate on
 */
function escapeShellChars(cm) {
    replaceInSelection(cm, /([^a-zA-Z0-9_./,])/g, "\\$1");
}
exports.COMMAND_ESCAPE_SHELL_CHARS = "escapeShellChars";
CodeMirror.commands[exports.COMMAND_ESCAPE_SHELL_CHARS] = escapeShellChars;
/**
 * Command to unescape backslash escaped shell special characters in the selection.
 *
 * @param cm the CodeMirror instance to operate on
 */
function unescapeShellChars(cm) {
    replaceInSelection(cm, /\\(.)/g, "$1");
}
exports.COMMAND_UNESCAPE_SHELL_CHARS = "unescapeShellChars";
CodeMirror.commands[exports.COMMAND_UNESCAPE_SHELL_CHARS] = unescapeShellChars;
function toggleLineNumbers(cm) {
    const isSet = cm.getOption("lineNumbers");
    cm.setOption("lineNumbers", !isSet);
}
exports.COMMAND_TOGGLE_LINE_NUMBERS = "toggleLineNumbers";
function init() {
    // This is needed to make sure that the TypeScript compile sees that that module is used and should be included.
}
exports.init = init;
function isCommand(command) {
    return [exports.COMMAND_FORWARDSLASHES_TO_BACK,
        exports.COMMAND_BACKSLASHES_TO_FORWARD,
        exports.COMMAND_ESCAPE_SHELL_CHARS,
        exports.COMMAND_UNESCAPE_SHELL_CHARS,
        exports.COMMAND_VERTICAL_MULTICURSOR,
        exports.COMMAND_TOGGLE_LINE_NUMBERS,
    ].indexOf(command) !== -1;
}
exports.isCommand = isCommand;
function executeCommand(cm, command) {
    if (command === exports.COMMAND_TOGGLE_LINE_NUMBERS) {
        toggleLineNumbers(cm);
    }
    else {
        cm.execCommand(command);
    }
}
exports.executeCommand = executeCommand;
function commandDescriptions(cm) {
    const descriptions = [
        { command: exports.COMMAND_FORWARDSLASHES_TO_BACK, iconRight: "", label: "Forward Slashes to Backslashes" },
        { command: exports.COMMAND_BACKSLASHES_TO_FORWARD, iconRight: "", label: "Backslashes to Forward Slashes" },
        { command: exports.COMMAND_ESCAPE_SHELL_CHARS, iconRight: "", label: "Escape Shell Characters" },
        { command: exports.COMMAND_UNESCAPE_SHELL_CHARS, iconRight: "", label: "Unescape Shell Characters" },
        { command: exports.COMMAND_VERTICAL_MULTICURSOR, iconRight: "", label: "Selection to Multiple Cursors" },
        { command: exports.COMMAND_TOGGLE_LINE_NUMBERS, iconLeft: cm.getOption("lineNumbers") ? "check-square-o" : "square-o",
            iconRight: "list-ol", label: "Line Numbers" },
    ];
    return descriptions;
}
exports.commandDescriptions = commandDescriptions;
//# sourceMappingURL=CodeMirrorCommands.js.map