"use strict";
/*
 * Copyright 2014-2016 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const Util = require("./gui/Util");
/**
 * Convert an array-like object to a real array.
 *
 * @param {Object} fakeArray An array-like object with get() and length support.
 * @return {Array} A real array object with the elements as fakeArray.
 */
function toArray(fakeArray) {
    var result = [];
    var len = fakeArray.length;
    for (var i = 0; i < len; i++) {
        result.push(fakeArray[i]);
    }
    return result;
}
exports.toArray = toArray;
/**
 * Split child nodes at a character offset.
 *
 * @param {Node} rootNode The node whose child nodes are to be split.
 * @param {number} charOffset The character offset to split at.
 * @return {Object} An object of the form {left: array, right: array}
 *     containing arrays of the child nodes left of, and right of the
 *     character offset.
 */
function splitNodeContentsAtChar(rootNode, charOffset) {
    var newoffset;
    var newNode;
    var kids = toArray(rootNode.childNodes);
    var len = kids.length;
    var offset = 0;
    if (charOffset === 0) {
        return { left: [], right: kids };
    }
    for (var i = 0; i < len; i++) {
        var kid = kids[i];
        if (kid.nodeName === "#text") {
            newoffset = offset + kid.nodeValue.length;
            if (newoffset === charOffset) {
                return { left: kids.slice(0, i + 1), right: kids.slice(i + 1) };
            }
            if (newoffset > charOffset) {
                newNode = kid.ownerDocument.createTextNode(kid.nodeValue.slice(0, charOffset - offset));
                kid.nodeValue = kid.nodeValue.slice(charOffset - offset);
                rootNode.insertBefore(newNode, kid);
                return { left: kids.slice(0, i).concat(newNode), right: kids.slice(i) };
            }
        }
        else {
            // Assume a normal node.
            newoffset = offset + kid.textContent.length;
            if (newoffset === charOffset) {
                return { left: kids.slice(0, i + 1), right: kids.slice(i + 1) };
            }
            if (newoffset > charOffset) {
                var parts = splitNodeContentsAtChar(kid, charOffset - offset);
                newNode = kid.cloneNode(false);
                parts.left.forEach(function (node) {
                    newNode.appendChild(node);
                });
                rootNode.insertBefore(newNode, kid);
                return { left: kids.slice(0, i).concat(newNode), right: kids.slice(i) };
            }
        }
        offset = newoffset;
    }
    return { left: kids, right: [] };
}
exports.splitNodeContentsAtChar = splitNodeContentsAtChar;
function encloseCharRange(rootNode, startOffset, endOffset, newNode) {
    var split1Results = splitNodeContentsAtChar(rootNode, startOffset);
    var nodeStartOffset = split1Results.left.length;
    var split2Results = splitNodeContentsAtChar(rootNode, endOffset);
    var nodeEndOffset = split2Results.left.length;
    var kids = toArray(rootNode.childNodes).slice(nodeStartOffset, nodeEndOffset);
    kids.forEach(function (kid) {
        newNode.appendChild(kid);
    });
    insertNodeAt(rootNode, nodeStartOffset, newNode);
}
exports.encloseCharRange = encloseCharRange;
/**
 * Insert a node under a parent node at an offset.
 *
 * @param {Node} rootNode
 * @param {number} offset
 * @param {Node} newNode
 */
function insertNodeAt(rootNode, offset, newNode) {
    if (offset >= rootNode.childNodes.length) {
        rootNode.appendChild(newNode);
    }
    else {
        rootNode.insertBefore(newNode, rootNode.childNodes[offset]);
    }
}
exports.insertNodeAt = insertNodeAt;
/**
 * Extracts formatted plain text from a document range.
 *
 * @param range the range to scan and extract the text from.
 * @return the formatted plain text representation.
 */
function extractTextFromRange(range) {
    const nbspRegexp = /\u00a0/g;
    const startContainer = range.startContainer;
    const endContainer = range.endContainer;
    let result = "";
    // Use case where the start and end of the range are inside the same text node.
    if (startContainer === range.endContainer) {
        if (startContainer.nodeType === Node.TEXT_NODE) {
            const textNode = startContainer;
            return textNode.data.slice(range.startOffset, range.endOffset).replace(nbspRegexp, " ");
        }
    }
    let currentNode;
    if (startContainer.nodeType === Node.TEXT_NODE) {
        result += startContainer.data.slice(range.startOffset).replace(nbspRegexp, " ");
        currentNode = nextDocumentOrderNode(startContainer);
    }
    else {
        currentNode = startContainer.childNodes[range.startOffset];
    }
    const endNode = endContainer.nodeType === Node.TEXT_NODE ? endContainer : endContainer.childNodes[range.endOffset];
    while (currentNode !== endNode && currentNode !== null) {
        if (currentNode.nodeType === Node.TEXT_NODE) {
            result += currentNode.data.replace(nbspRegexp, " ");
        }
        else if (currentNode.nodeName === "DIV") {
            const divElement = currentNode;
            if (divElement.classList.contains('terminal-active') || divElement.classList.contains('terminal-scrollback')) {
                result = Util.trimRight(result) + "\n";
            }
        }
        currentNode = nextDocumentOrderNode(currentNode);
    }
    if (endContainer.nodeType === Node.TEXT_NODE) {
        result += endContainer.data.slice(0, range.endOffset).replace(nbspRegexp, " ");
    }
    return result;
}
exports.extractTextFromRange = extractTextFromRange;
function nextDocumentOrderNode(currentNode) {
    if (currentNode.childNodes.length !== 0) {
        return currentNode.childNodes[0];
    }
    if (currentNode.nextSibling !== null) {
        return currentNode.nextSibling;
    }
    return nextDocumentOrderNodeUp(currentNode.parentNode);
}
function nextDocumentOrderNodeUp(currentNode) {
    if (currentNode.nextSibling !== null) {
        return currentNode.nextSibling;
    }
    return nextDocumentOrderNodeUp(currentNode.parentNode);
}
function getShadowRoot(self) {
    return self.webkitShadowRoot ? self.webkitShadowRoot : self.shadowRoot;
}
exports.getShadowRoot = getShadowRoot;
function getShadowId(el, id) {
    const shadowRoot = getShadowRoot(el);
    if (shadowRoot === null) {
        return null;
    }
    return shadowRoot.querySelector('#' + id);
}
exports.getShadowId = getShadowId;
/**
 * Converts a node list to a real array.
 *
 * @param  nodeList the node list to convert.
 * @return          a new array holding the same contents as the node list.
 */
function nodeListToArray(nodeList) {
    let i = 0;
    const result = [];
    const len = nodeList.length;
    for (i = 0; i < len; i++) {
        result.push(nodeList[i]);
    }
    return result;
}
exports.nodeListToArray = nodeListToArray;
/**
 * Create a KeyboardEvent.
 *
 * This works around a bunch of bugs in Blink.
 *
 * @param  eventName event name, one of 'keypress', 'keydown' and 'keyup'
 * @param  initMap map of values to use to fill the event
 * @return new keyboard event
 */
function newKeyboardEvent(eventName, initMap) {
    const fakeKeyDownEvent = new KeyboardEvent(eventName, initMap);
    // https://stackoverflow.com/questions/12937391/cannot-initialize-keycode-in-keyboard-event-init-method      
    Object.defineProperty(fakeKeyDownEvent, 'keyCode', {
        get: function () {
            return initMap.keyCode;
        }
    });
    Object.defineProperty(fakeKeyDownEvent, 'charCode', {
        get: function () {
            return initMap.charCode;
        }
    });
    Object.defineProperty(fakeKeyDownEvent, 'code', {
        get: function () {
            return initMap.code;
        }
    });
    Object.defineProperty(fakeKeyDownEvent, 'which', {
        get: function () {
            return initMap.which;
        }
    });
    return fakeKeyDownEvent;
}
exports.newKeyboardEvent = newKeyboardEvent;
/**
 * Add an event listener which blocks an event and retransmits it.
 *
 * @param listenTarget    the object on which to intercept the custom event
 * @param eventName the name of the Custom Event to intercept and retransmit
 * @param retransmitTarget (Optional) the element on which to retransmit the event
 */
function addCustomEventResender(listenTarget, eventName, retransmitTarget = null) {
    listenTarget.addEventListener(eventName, (ev) => {
        const inflightResentEvent = inflightResentEvents.get(ev.target);
        if (inflightResentEvent === ev && inflightResentEvent.type === eventName) {
            return;
        }
        ev.stopPropagation();
        const detail = ev.detail;
        const bubbles = ev.bubbles;
        const event = new CustomEvent(eventName, { bubbles: bubbles, detail: detail });
        inflightResentEvents.set(listenTarget, event);
        (retransmitTarget === null ? listenTarget : retransmitTarget).dispatchEvent(event);
        inflightResentEvents.delete(listenTarget);
    });
}
exports.addCustomEventResender = addCustomEventResender;
const inflightResentEvents = new Map();
let doLaterId = -1;
let laterList = [];
function doLaterTimeoutHandler() {
    doLaterId = -1;
    const workingList = [...laterList];
    laterList = [];
    workingList.forEach(f => f());
}
/**
 * Schedule a function to be executed later.
 *
 * @param  func the function to be executed later
 * @param  msec Optional time delay in ms. Default is 0.
 * @return {LaterHandle} This object can be used to cancel the scheduled execution.
 */
function doLater(func, msec = 0) {
    laterList.push(func);
    if (doLaterId === -1) {
        doLaterId = window.setTimeout(doLaterTimeoutHandler, msec);
    }
    return { cancel: () => {
            laterList = laterList.filter(f => f !== func);
        } };
}
exports.doLater = doLater;
let doLaterFrameId = -1;
let laterFrameList = [];
/**
 * Schedule a function to run at the next animation frame.
 */
function doLaterFrameHandler() {
    const workingList = [...laterFrameList];
    laterFrameList = [];
    window.cancelAnimationFrame(doLaterFrameId);
    doLaterFrameId = -1;
    workingList.forEach(f => f());
}
function doLaterFrame(func) {
    laterFrameList.push(func);
    if (doLaterFrameId === -1) {
        doLaterFrameId = window.requestAnimationFrame(doLaterFrameHandler);
    }
    return { cancel: () => {
            laterFrameList = laterFrameList.filter(f => f !== func);
        } };
}
exports.doLaterFrame = doLaterFrame;
//-------------------------------------------------------------------------
/**
 * Format a Uint8Array and mimetype as a data url.
 *
 * @param  buffer   [description]
 * @param  mimeType [description]
 * @return          [description]
 */
function CreateDataUrl(buffer, mimeType) {
    const base64Data = buffer.toString('base64');
    return "data:" + mimeType + ";base64," + base64Data;
}
exports.CreateDataUrl = CreateDataUrl;
function getEventDeepPath(ev) {
    return ev.composedPath();
}
exports.getEventDeepPath = getEventDeepPath;
/**
 * Set the input focus on an element and prevent any scrolling for occuring.
 *
 * @param el the element to focus
 */
function focusWithoutScroll(el) {
    const preScrollTops = [];
    let p = el;
    do {
        preScrollTops.push({ element: p, top: p.scrollTop });
        let parent = p.parentElement;
        if (parent == null) {
            const nodeParent = p.parentNode;
            if (nodeParent != null && nodeParent.nodeName === "#document-fragment") {
                parent = nodeParent.host;
            }
        }
        p = parent;
    } while (p != null);
    el.focus();
    if (preScrollTops.length !== 0) {
        // Restore the previous scroll top values.
        preScrollTops.forEach((pair) => {
            if (pair.element.scrollTop !== pair.top) {
                pair.element.scrollTop = pair.top;
            }
        });
    }
}
exports.focusWithoutScroll = focusWithoutScroll;
/**
 * Prevent an Element from scrolling.
 *
 * @param el the element to prevent all scrolling on.
 */
function preventScroll(el) {
    el.addEventListener('scroll', (ev) => {
        el.scrollTop = 0;
    }, true);
    el.addEventListener('scroll', (ev) => {
        el.scrollTop = 0;
    });
}
exports.preventScroll = preventScroll;
/**
 * Convert a length with 'px' suffix to a plain integer.
 *
 * @param length the length value as a string
 * @return the length as a number
 */
function pixelLengthToInt(length) {
    if (typeof length === "string") {
        const lengthStr = length.indexOf("px") !== -1 ? length.substr(0, length.length - 2) : length;
        return parseInt(lengthStr, 10);
    }
    else {
        return length;
    }
}
exports.pixelLengthToInt = pixelLengthToInt;
/**
 * Expand a HTML string into Document Fragment.
 *
 * @param {html} the HTML code to use.
 * @return A Document Fragment containing the nodes defined by the `html` parameter.
 */
function htmlToFragment(html) {
    const div = document.createElement("DIV");
    div.innerHTML = html;
    const frag = document.createDocumentFragment();
    while (div.childNodes.length !== 0) {
        frag.appendChild(div.childNodes[0]);
    }
    return frag;
}
exports.htmlToFragment = htmlToFragment;
/**
 * Update the list of children on an element.
 *
 * This function updates the children while trying to do as few changes on
 * the DOM as possible.
 *
 * @param el the Element to update.
 * @param targetChildrenList the desired list of child elements.
 */
function setElementChildren(el, targetChildrenList) {
    // Delete phase
    const unneededChildrenSet = new Set(toArray(el.children));
    for (const kid of targetChildrenList) {
        unneededChildrenSet.delete(kid);
    }
    for (const kid of unneededChildrenSet) {
        el.removeChild(kid);
    }
    // Insert the missing children and fix the order.
    for (let i = 0; i < targetChildrenList.length; i++) {
        if (el.children.length <= i) {
            el.appendChild(targetChildrenList[i]);
        }
        else {
            if (el.children.item(i) !== targetChildrenList[i]) {
                el.insertBefore(targetChildrenList[i], el.children.item(i));
            }
        }
    }
}
exports.setElementChildren = setElementChildren;
/**
 * Test if a node is in the DOM tree.
 *
 * @param node the node to test
 * @return true if the node is attached somewhere inside its DOM tree.
 */
function isNodeInDom(node) {
    let currentNode = node;
    let nextNode = node;
    while (true) {
        currentNode = nextNode;
        if (currentNode.parentNode != null) {
            nextNode = currentNode.parentNode;
        }
        else if (currentNode.nodeType == Node.DOCUMENT_FRAGMENT_NODE && currentNode.host != null) {
            nextNode = currentNode.host;
        }
        else {
            break;
        }
    }
    return currentNode.nodeType === Node.DOCUMENT_NODE;
}
exports.isNodeInDom = isNodeInDom;
/**
 * Remove all CSS classes from an element.
 */
function removeAllClasses(el) {
    while (el.classList.length !== 0) {
        el.classList.remove(el.classList[0]);
    }
}
exports.removeAllClasses = removeAllClasses;
/**
 * Get the path from the node through its parents, to the root.
 *
 * This function will also traverse shadow DOM boundaries.
 *
 * @param node the node
 * @return the path from the node to its root parent with nodes closer to the
 *         start node first and the ultimate parent node last.
 */
function nodePathToRoot(node) {
    let currentNode = node;
    let nextNode = node;
    const path = [];
    while (true) {
        currentNode = nextNode;
        if (currentNode.parentNode != null) {
            nextNode = currentNode.parentNode;
            path.push(nextNode);
        }
        else if (currentNode.nodeType == Node.DOCUMENT_FRAGMENT_NODE && currentNode.host != null) {
            nextNode = currentNode.host;
            path.push(nextNode);
        }
        else {
            break;
        }
    }
    return path;
}
exports.nodePathToRoot = nodePathToRoot;
/**
 * Get the list of active/focused elements and their nested active elements.
 *
 * This function traverses Shadow DOM boundaries to find nested
 * active/focused elements starting from the top window.
 *
 * @returns list of active elements
 */
function activeNestedElements() {
    const result = [];
    let activeElement = window.document.activeElement;
    if (activeElement != null) {
        while (true) {
            result.push(activeElement);
            const shadowRoot = getShadowRoot(activeElement);
            if (shadowRoot == null) {
                break;
            }
            const nextActiveElement = shadowRoot.activeElement;
            if (nextActiveElement == null) {
                break;
            }
            activeElement = nextActiveElement;
        }
    }
    return result;
}
exports.activeNestedElements = activeNestedElements;
//# sourceMappingURL=DomUtils.js.map