"use strict";
/*
 * Copyright 2014-2016 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const MenuItem_1 = require("./gui/MenuItem");
const CheckboxMenuItem_1 = require("./gui/CheckboxMenuItem");
const DomUtils = require("./DomUtils");
const Util = require("./gui/Util");
const ViewerElement_1 = require("./ViewerElement");
const ViewerElementTypes = require("./ViewerElementTypes");
const ThemeableElementBase_1 = require("./ThemeableElementBase");
const VirtualScrollArea = require("./VirtualScrollArea");
const ThemeTypes = require("./Theme");
const CommandPaletteRequestTypes_1 = require("./CommandPaletteRequestTypes");
const Logger_1 = require("./Logger");
const SupportsClipboardPaste = require("./SupportsClipboardPaste");
const InternalMimeTypes_1 = require("./InternalMimeTypes");
const GeneralEvents_1 = require("./GeneralEvents");
MenuItem_1.MenuItem.init();
CheckboxMenuItem_1.CheckboxMenuItem.init();
const ID = "EtEmbeddedViewerTemplate";
const ID_CONTAINER = "ID_CONTAINER";
const ID_HEADER = "ID_HEADER";
const ID_OUTPUT = "ID_OUTPUT";
const ID_OUTPUT_CONTAINER = "ID_OUTPUT_CONTAINER";
const ID_ICON = "ID_ICON";
const ID_ICON_DIV = "ID_ICON_DIV";
const ID_COMMAND_LINE = "ID_COMMAND_LINE";
const ID_TAG_NAME = "ID_TAG_NAME";
const ID_SCROLL_ICON = "ID_SCROLL_ICON";
const ID_SCROLL_NAME = "ID_SCROLL_NAME";
// const ID_EXPAND_BUTTON = "expand_button";
// const ID_EXPAND_ICON = "expand_icon";
// const ID_EXPAND_MENU_ITEM = "expandmenuitem";
const ID_CLOSE_BUTTON = "ID_CLOSE_BUTTON";
const ID_POP_OUT_BUTTON = "ID_POP_OUT_BUTTON";
const ID_TAG_ICON = "ID_TAG_ICON";
const CLASS_SCROLLING = "scrolling";
const CLASS_NOT_SCROLLING = "not-scrolling";
const CLASS_BOTTOM_VISIBLE = "bottom-visible";
const CLASS_BOTTOM_NOT_VISIBLE = "bottom-not-visible";
const CLASS_COMMAND_RUNNING = "running";
const CLASS_COMMAND_FAILED = "fail";
const CLASS_COMMAND_SUCCEEDED = "success";
let registered = false;
const DEBUG_SIZE = false;
/**
 * A visual frame which contains another element and can be shown directly inside a terminal.
 */
class EmbeddedViewer extends ViewerElement_1.ViewerElement {
    //-----------------------------------------------------------------------
    //
    //   #                                                         
    //   #       # ###### ######  ####  #   #  ####  #      ###### 
    //   #       # #      #      #    #  # #  #    # #      #      
    //   #       # #####  #####  #        #   #      #      #####  
    //   #       # #      #      #        #   #      #      #      
    //   #       # #      #      #    #   #   #    # #      #      
    //   ####### # #      ######  ####    #    ####  ###### ###### 
    //
    //-----------------------------------------------------------------------
    constructor() {
        super();
        this._initProperties();
    }
    /**
     * Initialize the EtEmbeddedViewer class and resources.
     *
     * When EtEmbeddedViewer is imported into a render process, this static method
     * must be called before an instances may be created. This is can be safely
     * called multiple times.
     */
    static init() {
        if (registered === false) {
            window.customElements.define(EmbeddedViewer.TAG_NAME.toLowerCase(), EmbeddedViewer);
            registered = true;
        }
    }
    /**
     * Type guard for detecting a EtEmbeddedViewer instance.
     *
     * @param  node the node to test
     * @return      True if the node is a EtEmbeddedViewer.
     */
    static is(node) {
        return node !== null && node !== undefined && node instanceof EmbeddedViewer;
    }
    _initProperties() {
        this._log = Logger_1.getLogger(EmbeddedViewer.TAG_NAME, this);
        this._visualState = ViewerElementTypes.VisualState.AUTO;
        this._mode = ViewerElementTypes.Mode.DEFAULT;
        this._virtualScrollArea = new VirtualScrollArea.VirtualScrollArea();
        this._childFocusHandlerFunc = this._handleChildFocus.bind(this);
        this._requestContainerHeight = false;
        this._requestContainerScroll = false;
        this._requestContainerYScroll = 0;
        this._headerTop = 0;
        this._headerBottom = 0;
    }
    //-----------------------------------------------------------------------
    //
    // ######                                
    // #     # #    # #####  #      #  ####  
    // #     # #    # #    # #      # #    # 
    // ######  #    # #####  #      # #      
    // #       #    # #    # #      # #      
    // #       #    # #    # #      # #    # 
    // #        ####  #####  ###### #  ####  
    //
    //-----------------------------------------------------------------------
    setViewerElement(element) {
        const oldViewer = this._getViewerElement();
        if (oldViewer != null) {
            oldViewer.removeEventListener('focus', this._childFocusHandlerFunc);
        }
        if (this.childNodes.length !== 0) {
            this.innerHTML = "";
        }
        if (element !== null) {
            element.setVisualState(this._visualState);
            element.setMode(this._mode);
            element.addEventListener('focus', this._childFocusHandlerFunc);
            this.appendChild(element);
            this._virtualScrollArea.appendScrollable(element);
        }
    }
    getViewerElement() {
        return this._getViewerElement();
    }
    setVisualState(newVisualState) {
        this._visualState = newVisualState;
        const viewerElement = this.getViewerElement();
        if (viewerElement !== null) {
            viewerElement.setVisualState(newVisualState);
        }
    }
    getVisualState() {
        return this._visualState;
    }
    // See VirtualScrollable
    getMinHeight() {
        if (DEBUG_SIZE) {
            this._log.debug("getMinHeight() => ", this.getReserveViewportHeight(0));
        }
        return this.getReserveViewportHeight(0);
    }
    // See VirtualScrollable
    getVirtualHeight(containerHeight) {
        const viewerElement = this.getViewerElement();
        let result = 0;
        if (viewerElement !== null) {
            result = this._virtualScrollArea.getVirtualHeight();
        }
        if (DEBUG_SIZE) {
            this._log.debug("getVirtualHeight() => ", result);
        }
        return result;
    }
    // See VirtualScrollable
    getReserveViewportHeight(containerHeight) {
        const { top, bottom } = this._borderSize();
        const result = top + bottom;
        if (DEBUG_SIZE) {
            this._log.debug("getReserveViewportHeight() => ", result);
        }
        return result;
    }
    // See VirtualScrollable
    setDimensionsAndScroll(setterState) {
        if (DEBUG_SIZE) {
            this._log.debug(`setDimensionsAndScroll(): height=${setterState.height}, ` +
                `heightChanged=${setterState.heightChanged}, yOffset=${setterState.yOffset}, ` +
                `yOffsetChanged=${setterState.yOffsetChanged}, physicalTop=${setterState.physicalTop}, ` +
                `containerHeight=${setterState.containerHeight}`);
        }
        if (setterState.heightChanged) {
            this.style.height = "" + setterState.height + "px";
        }
        const containerDiv = this._getById(ID_CONTAINER);
        if (setterState.yOffset === 0) {
            containerDiv.classList.remove(CLASS_SCROLLING);
            containerDiv.classList.add(CLASS_NOT_SCROLLING);
        }
        else {
            containerDiv.classList.add(CLASS_SCROLLING);
            containerDiv.classList.remove(CLASS_NOT_SCROLLING);
        }
        const headerDiv = this._getById(ID_HEADER);
        const rect = headerDiv.getBoundingClientRect();
        headerDiv.style.top = Math.min(Math.max(setterState.physicalTop, 0), setterState.height - rect.height) + 'px';
        const outputContainerDiv = this._getById(ID_OUTPUT_CONTAINER);
        outputContainerDiv.style.top = "" + rect.height + "px";
        if (setterState.physicalTop > 0 || setterState.height < setterState.containerHeight) {
            // Bottom part is visible
            containerDiv.classList.remove(CLASS_BOTTOM_NOT_VISIBLE);
            containerDiv.classList.add(CLASS_BOTTOM_VISIBLE);
        }
        else {
            containerDiv.classList.add(CLASS_BOTTOM_NOT_VISIBLE);
            containerDiv.classList.remove(CLASS_BOTTOM_VISIBLE);
        }
        const scrollNameDiv = this._getById(ID_SCROLL_NAME);
        const percent = Math.floor(setterState.yOffset / this.getVirtualHeight(0) * 100);
        scrollNameDiv.innerHTML = "" + percent + "%";
        if (setterState.heightChanged) {
            this._requestContainerHeight = true;
        }
        this._requestContainerScroll = true;
        this._requestContainerYScroll = setterState.yOffset;
        if (this.parentElement != null) {
            this._applyContainerChanges();
        }
        if (setterState.physicalTopChanged || setterState.containerHeight || setterState.heightChanged) {
            const viewportBottomOffset = setterState.physicalTop + setterState.containerHeight - setterState.height;
            this._virtualScrollArea.setViewportBottomOffset(viewportBottomOffset);
        }
    }
    markVisible(visible) {
        if (visible) {
            if (DEBUG_SIZE) {
                this._log.debug("markVisible()");
            }
            this._applyContainerChanges();
            this._virtualScrollArea.reapplyState();
        }
    }
    _applyContainerChanges() {
        if (this._requestContainerHeight) {
            this._requestContainerHeight = false;
            const outputContainerDiv = this._getById(ID_OUTPUT);
            this._virtualScrollArea.updateContainerHeight(outputContainerDiv.getBoundingClientRect().height);
        }
        if (this._requestContainerScroll) {
            this._requestContainerScroll = false;
            this._virtualScrollArea.scrollTo(this._requestContainerYScroll);
        }
    }
    getSelectionText() {
        const viewerElement = this.getViewerElement();
        return viewerElement === null ? null : viewerElement.getSelectionText();
    }
    getText() {
        const viewerElement = this.getViewerElement();
        if (viewerElement === null) {
            return "";
        }
        return viewerElement.getText();
    }
    setTag(tag) {
        this.setAttribute(EmbeddedViewer.ATTR_TAG, tag);
    }
    getTag() {
        return this.getAttribute(EmbeddedViewer.ATTR_TAG);
    }
    setTitle(newTitle) {
        this.setAttribute(EmbeddedViewer.ATTR_FRAME_TITLE, newTitle);
    }
    getTitle() {
        return this.getAttribute(EmbeddedViewer.ATTR_FRAME_TITLE);
    }
    hasFocus() {
        const el = this.getViewerElement();
        if (el == null) {
            return false;
        }
        return el.hasFocus();
    }
    setReturnCode(returnCode) {
        this.setAttribute(EmbeddedViewer.ATTR_RETURN_CODE, returnCode === null || returnCode === undefined ? null : "" + returnCode);
    }
    getReturnCode() {
        const rcString = this.getAttribute(EmbeddedViewer.ATTR_RETURN_CODE);
        return rcString === null || rcString === undefined ? null : parseInt(rcString, 10);
    }
    setAwesomeIcon(iconName) {
        this.setAttribute(EmbeddedViewer.ATTR_AWESOME_ICON, iconName);
    }
    getAwesomeIcon() {
        return this.getAttribute(EmbeddedViewer.ATTR_AWESOME_ICON);
    }
    canPaste() {
        const el = this.getViewerElement();
        if (el == null) {
            return false;
        }
        return SupportsClipboardPaste.isSupportsClipboardPaste(el) && el.canPaste();
    }
    pasteText(text) {
        if (!this.canPaste()) {
            return;
        }
        const el = this.getViewerElement();
        if (SupportsClipboardPaste.isSupportsClipboardPaste(el)) {
            el.pasteText(text);
        }
    }
    clearSelection() {
        const viewerElement = this.getViewerElement();
        if (viewerElement === null) {
            return;
        }
        viewerElement.clearSelection();
    }
    setMode(newMode) {
        this._mode = newMode;
        const viewerElement = this.getViewerElement();
        if (viewerElement !== null) {
            return viewerElement.setMode(newMode);
        }
    }
    getMode() {
        return this._mode;
    }
    focus() {
        const viewerElement = this.getViewerElement();
        if (viewerElement !== null) {
            return viewerElement.focus();
        }
        else {
            super.focus();
        }
    }
    getCursorPosition() {
        const viewerElement = this.getViewerElement();
        if (viewerElement !== null) {
            const borderSize = this._borderSize();
            const { left, top, bottom, viewPortTop } = viewerElement.getCursorPosition();
            return {
                left,
                top: top + borderSize.top,
                bottom: bottom + borderSize.top,
                viewPortTop: viewPortTop + borderSize.top
            };
        }
        return null;
    }
    setCursorPositionTop(x) {
        const viewerElement = this.getViewerElement();
        if (viewerElement !== null) {
            return viewerElement.setCursorPositionTop(x);
        }
        return false;
    }
    setCursorPositionBottom(x) {
        const viewerElement = this.getViewerElement();
        if (viewerElement !== null) {
            return viewerElement.setCursorPositionBottom(x);
        }
        return false;
    }
    /**
     * Custom Element 'connected' life cycle hook.
     */
    connectedCallback() {
        super.connectedCallback();
        if (DomUtils.getShadowRoot(this) !== null) {
            return;
        }
        this._setUpShadowDom();
        this._setUpDefaultAttributes();
        this.installThemeCss();
        this._setUpEventHandlers();
        this._setUpVirtualScrollArea();
        // Remove the anti-flicker style.
        DomUtils.getShadowId(this, ID_CONTAINER).setAttribute('style', '');
    }
    _setUpShadowDom() {
        const shadow = this.attachShadow({ mode: 'open', delegatesFocus: true });
        const clone = this._createClone();
        shadow.appendChild(clone);
    }
    _setUpDefaultAttributes() {
        this._setAttr(EmbeddedViewer.ATTR_FRAME_TITLE, this.getAttribute(EmbeddedViewer.ATTR_FRAME_TITLE));
        this._setAttr(EmbeddedViewer.ATTR_RETURN_CODE, this.getAttribute(EmbeddedViewer.ATTR_RETURN_CODE));
        this._setAttr(EmbeddedViewer.ATTR_EXPAND, this.getAttribute(EmbeddedViewer.ATTR_EXPAND));
        this._setAttr(EmbeddedViewer.ATTR_TAG, this.getAttribute(EmbeddedViewer.ATTR_TAG));
        this._setAttr(EmbeddedViewer.ATTR_TOOL_TIP, this.getAttribute(EmbeddedViewer.ATTR_TOOL_TIP));
        this._setAttr(EmbeddedViewer.ATTR_AWESOME_ICON, this.getAttribute(EmbeddedViewer.ATTR_AWESOME_ICON));
    }
    _setUpEventHandlers() {
        DomUtils.getShadowId(this, ID_POP_OUT_BUTTON).addEventListener('click', this._emitFramePopOut.bind(this));
        DomUtils.getShadowId(this, ID_CLOSE_BUTTON).addEventListener('click', this._emitCloseRequest.bind(this));
        const headerDiv = DomUtils.getShadowId(this, ID_HEADER);
        headerDiv.addEventListener('dragstart', this._handleDragStart.bind(this), false);
        headerDiv.addEventListener('dragend', this._handleDragEnd.bind(this), false);
        const outputDiv = DomUtils.getShadowId(this, ID_OUTPUT);
        outputDiv.addEventListener('mousedown', this.focus.bind(this));
        outputDiv.addEventListener('click', this.focus.bind(this));
        outputDiv.addEventListener('keydown', this._handleKeyDown.bind(this));
        outputDiv.addEventListener(VirtualScrollArea.EVENT_RESIZE, this._handleVirtualScrollableResize.bind(this));
        DomUtils.addCustomEventResender(this, ViewerElement_1.ViewerElement.EVENT_BEFORE_SELECTION_CHANGE);
        DomUtils.addCustomEventResender(this, ViewerElement_1.ViewerElement.EVENT_CURSOR_MOVE);
        DomUtils.addCustomEventResender(this, ViewerElement_1.ViewerElement.EVENT_CURSOR_EDGE);
        // Right mouse button click opens up the command palette.
        DomUtils.getShadowId(this, ID_CONTAINER).addEventListener('contextmenu', (ev) => {
            ev.stopPropagation();
            ev.preventDefault();
            const viewerElement = this.getViewerElement();
            if (viewerElement === null) {
                return;
            }
            if (CommandPaletteRequestTypes_1.isCommandable(viewerElement)) {
                viewerElement.executeCommand(CommandPaletteRequestTypes_1.COMMAND_OPEN_COMMAND_PALETTE);
            }
            else {
                this.executeCommand(CommandPaletteRequestTypes_1.COMMAND_OPEN_COMMAND_PALETTE);
            }
        });
    }
    _setUpVirtualScrollArea() {
        const outputContainerDiv = DomUtils.getShadowId(this, ID_OUTPUT_CONTAINER);
        DomUtils.preventScroll(outputContainerDiv);
        this._virtualScrollArea.setScrollFunction((offset) => {
            const outputDiv = DomUtils.getShadowId(this, ID_OUTPUT);
            outputDiv.style.top = "-" + offset + "px";
        });
        // const expandbutton = this._getById(ID_EXPAND_BUTTON);
        // expandbutton.addEventListener('click', (): void => {
        //   const expanded = util.htmlValueToBool(this.getAttribute(EtEmbeddedViewer.ATTR_EXPAND), true);
        //   this.setAttribute(EtEmbeddedViewer.ATTR_EXPAND, "" + !expanded);
        // });
        const setterState = {
            height: this.getMinHeight(),
            heightChanged: true,
            yOffset: 0,
            yOffsetChanged: true,
            physicalTop: 0,
            physicalTopChanged: true,
            containerHeight: this.getMinHeight(),
            containerHeightChanged: true,
            visibleBottomOffset: 0,
            visibleBottomOffsetChanged: true
        };
        this.setDimensionsAndScroll(setterState);
    }
    static get observedAttributes() {
        return [
            EmbeddedViewer.ATTR_FRAME_TITLE,
            EmbeddedViewer.ATTR_RETURN_CODE,
            EmbeddedViewer.ATTR_EXPAND,
            EmbeddedViewer.ATTR_TAG,
            EmbeddedViewer.ATTR_TOOL_TIP,
            EmbeddedViewer.ATTR_AWESOME_ICON
        ];
    }
    _handleDragStart(ev) {
        ev.stopPropagation();
        const target = ev.target;
        if (target.getAttribute("draggable") == null || this.getViewerElement() == null) {
            ev.preventDefault();
            return;
        }
        ev.dataTransfer.setData(InternalMimeTypes_1.FrameMimeType.MIMETYPE, "" + this.getTag());
        ev.dataTransfer.setData("text/plain", this.getText());
        ev.dataTransfer.setDragImage(target, -10, -10);
        ev.dataTransfer.effectAllowed = 'move';
        ev.dataTransfer.dropEffect = 'move';
        const dragStartedEvent = new CustomEvent(GeneralEvents_1.EVENT_DRAG_STARTED, { bubbles: true });
        this.dispatchEvent(dragStartedEvent);
    }
    _handleDragEnd(ev) {
        const dragEndedEvent = new CustomEvent(GeneralEvents_1.EVENT_DRAG_ENDED, { bubbles: true });
        this.dispatchEvent(dragEndedEvent);
    }
    /**
     * Custom Element 'attribute changed' hook.
     */
    attributeChangedCallback(attrName, oldValue, newValue) {
        this._setAttr(attrName, newValue);
    }
    _themeCssFiles() {
        return [ThemeTypes.CssFile.GUI_CONTROLS, ThemeTypes.CssFile.FONT_AWESOME, ThemeTypes.CssFile.EMBEDDED_FRAME];
    }
    executeCommand(commandId) {
        this._executeCommand(commandId);
    }
    //-----------------------------------------------------------------------
    //
    // ######                                      
    // #     # #####  # #    #   ##   ##### ###### 
    // #     # #    # # #    #  #  #    #   #      
    // ######  #    # # #    # #    #   #   #####  
    // #       #####  # #    # ######   #   #      
    // #       #   #  #  #  #  #    #   #   #      
    // #       #    # #   ##   #    #   #   ###### 
    //
    //-----------------------------------------------------------------------
    _createClone() {
        let template = window.document.getElementById(ID);
        if (template === null) {
            template = window.document.createElement('template');
            template.id = ID;
            template.innerHTML = `
        <style id=${ThemeableElementBase_1.ThemeableElementBase.ID_THEME}></style>
        <div id='${ID_CONTAINER}' style='display: none;' class='${CLASS_COMMAND_RUNNING}'>
          <div id='${ID_HEADER}' tabindex='0' draggable='true'>
            <div class='left_block'>
              <div id='${ID_ICON_DIV}'><i id='${ID_ICON}'></i></div>
              <div id='${ID_COMMAND_LINE}'></div>
            </div>
            <div class='header_spacer'></div>
            <div class='right_block'>
              <div id='${ID_SCROLL_ICON}'><i class='fa fa-arrows-v'></i></div>
              <div id='${ID_SCROLL_NAME}'></div>
              <div id='${ID_TAG_ICON}'><i class='fa fa-tag'></i></div>
              <div id='${ID_TAG_NAME}'></div>
              <div class='spacer'></div>
` +
                //              <div class='spacer'></div>
                `              <button id='${ID_POP_OUT_BUTTON}'><i class='fa fa-external-link'></i></button>
              <div class='spacer'></div>
              <button id='${ID_CLOSE_BUTTON}' title='Close'><i class='fa fa-times-circle'></i></button>` +
                `</div>` +
                `</div>
          <div id='${ID_OUTPUT_CONTAINER}'><div id='${ID_OUTPUT}'><slot></slot></div></div>
        </div>`;
            window.document.body.appendChild(template);
        }
        return window.document.importNode(template.content, true);
    }
    _getById(id) {
        return DomUtils.getShadowRoot(this).querySelector('#' + id);
    }
    /**
     * Process an attribute value change.
     */
    _setAttr(attrName, newValue) {
        if (DomUtils.getShadowRoot(this) === null) {
            return;
        }
        if (attrName === EmbeddedViewer.ATTR_FRAME_TITLE) {
            this._getById(ID_COMMAND_LINE).innerText = newValue;
            return;
        }
        if (attrName === EmbeddedViewer.ATTR_RETURN_CODE) {
            const container = this._getById(ID_CONTAINER);
            if (newValue === null || newValue === undefined || newValue === "") {
                container.classList.add(CLASS_COMMAND_RUNNING);
                container.classList.remove(CLASS_COMMAND_SUCCEEDED);
                container.classList.remove(CLASS_COMMAND_FAILED);
            }
            else {
                const rc = parseInt(newValue, 10);
                container.classList.remove(CLASS_COMMAND_RUNNING);
                if (rc === 0) {
                    container.classList.add(CLASS_COMMAND_SUCCEEDED);
                }
                else {
                    container.classList.add(CLASS_COMMAND_FAILED);
                }
            }
            return;
        }
        if (attrName === EmbeddedViewer.ATTR_EXPAND) {
            const output = this._getById(ID_OUTPUT);
            // const expandicon = <HTMLDivElement>this._getById(ID_EXPAND_ICON);
            if (Util.htmlValueToBool(newValue, true)) {
                // Expanded.
                output.classList.remove('closed');
                // expandicon.classList.remove('fa-plus-square-o');
                // expandicon.classList.add('fa-minus-square-o');
                // (<checkboxmenuitem>this._getById(ID_EXPAND_MENU_ITEM)).setAttribute('checked', "true");
            }
            else {
                // Collapsed.
                output.classList.add('closed');
                // expandicon.classList.add('fa-plus-square-o');
                // expandicon.classList.remove('fa-minus-square-o');
                // (<checkboxmenuitem>this._getById(ID_EXPAND_MENU_ITEM)).setAttribute('checked', "false");
            }
            return;
        }
        if (attrName === EmbeddedViewer.ATTR_TAG) {
            const tagName = this._getById(ID_TAG_NAME);
            tagName.innerText = newValue;
        }
        if (attrName === EmbeddedViewer.ATTR_TOOL_TIP) {
            const iconDiv = this._getById(ID_ICON_DIV);
            if (newValue !== null) {
                iconDiv.setAttribute('title', newValue);
            }
        }
        if (attrName === EmbeddedViewer.ATTR_AWESOME_ICON) {
            const icon = this._getById(ID_ICON);
            icon.className = "fa " + (newValue !== null && newValue !== undefined && newValue !== "" ? "fa-" : "") + newValue;
        }
    }
    _borderSize() {
        const headerDiv = this._getById(ID_HEADER);
        const outputContainerDiv = this._getById(ID_OUTPUT_CONTAINER);
        const outputContainerStyle = window.getComputedStyle(outputContainerDiv);
        const headerRect = headerDiv.getBoundingClientRect();
        if (headerRect.width === 0) {
            // Bogus info. This element most likely isn't in the DOM tree proper. Fall back to the last good read.
            return { top: this._headerTop, bottom: this._headerBottom };
        }
        const top = headerRect.height + DomUtils.pixelLengthToInt(outputContainerStyle.borderTopWidth);
        const bottom = DomUtils.pixelLengthToInt(outputContainerStyle.borderBottomWidth);
        this._headerTop = top;
        this._headerBottom = bottom;
        return { top, bottom };
    }
    _getViewerElement() {
        if (this.firstElementChild !== null && this.firstElementChild instanceof ViewerElement_1.ViewerElement) {
            return this.firstElementChild;
        }
        else {
            return null;
        }
    }
    _handleChildFocus(ev) {
        const focusEvent = new FocusEvent('focus', {});
        this.dispatchEvent(focusEvent);
    }
    _handleKeyDown(ev) {
        if (ev.keyCode === 79 && ev.ctrlKey && ev.shiftKey) {
            ev.stopPropagation();
            ev.preventDefault();
            this._emitFramePopOut();
            return;
        }
        if (ev.keyCode === 87 && ev.ctrlKey && ev.shiftKey) {
            ev.stopPropagation();
            ev.preventDefault();
            this._emitCloseRequest();
            return;
        }
    }
    _executeCommand(command) {
        switch (command) {
            case CommandPaletteRequestTypes_1.COMMAND_OPEN_COMMAND_PALETTE:
                CommandPaletteRequestTypes_1.dispatchCommandPaletteRequest(this);
                break;
            default:
                return false;
        }
        return true;
    }
    getCommandPaletteEntries(commandableStack) {
        return [];
    }
    _emitManualScroll() {
        const event = new CustomEvent(EmbeddedViewer.EVENT_SCROLL_MOVE);
        event.initCustomEvent(EmbeddedViewer.EVENT_SCROLL_MOVE, true, true, null);
        this.dispatchEvent(event);
    }
    _emitFramePopOut() {
        const event = new CustomEvent(EmbeddedViewer.EVENT_FRAME_POP_OUT);
        event.initCustomEvent(EmbeddedViewer.EVENT_FRAME_POP_OUT, true, true, this);
        this.dispatchEvent(event);
    }
    _emitCloseRequest() {
        const event = new CustomEvent(EmbeddedViewer.EVENT_CLOSE_REQUEST);
        event.initCustomEvent(EmbeddedViewer.EVENT_CLOSE_REQUEST, true, true, null);
        this.dispatchEvent(event);
    }
    _handleVirtualScrollableResize(ev) {
        const scrollable = ev.target;
        const height = this._virtualScrollArea.getVirtualHeight();
        this._virtualScrollArea.updateScrollableSize(scrollable);
        const newHeight = this._virtualScrollArea.getVirtualHeight();
        if (height !== newHeight) {
            VirtualScrollArea.emitResizeEvent(this);
        }
    }
}
/**
 * The HTML tag name of this element.
 */
EmbeddedViewer.TAG_NAME = 'ET-EMBEDDEDVIEWER';
EmbeddedViewer.EVENT_CLOSE_REQUEST = 'close-request';
EmbeddedViewer.EVENT_FRAME_POP_OUT = 'frame-pop-out';
EmbeddedViewer.EVENT_SCROLL_MOVE = 'scroll-move';
EmbeddedViewer.ATTR_FRAME_TITLE = 'frame-title';
EmbeddedViewer.ATTR_RETURN_CODE = "return-code";
EmbeddedViewer.ATTR_EXPAND = 'expand';
EmbeddedViewer.ATTR_TAG = 'tag';
EmbeddedViewer.ATTR_TOOL_TIP = 'tool-tip';
EmbeddedViewer.ATTR_AWESOME_ICON = 'awesome-icon';
exports.EmbeddedViewer = EmbeddedViewer;
//# sourceMappingURL=EmbeddedViewer.js.map