"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/*
 * Copyright 2017 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */
const fs = require("fs");
const path = require("path");
const _ = require("lodash");
const Logger_1 = require("./Logger");
class ExtensionLoader {
    constructor(extensionPaths) {
        this.extensionPaths = extensionPaths;
        this._log = null;
        this._extensions = [];
        this._log = Logger_1.getLogger("ExtensionLoader", this);
    }
    scan() {
        this._extensions = _.flatten(this.extensionPaths.map(p => this._scanPath(p)));
    }
    getExtensions() {
        return this._extensions;
    }
    load(extension) {
        if (extension.module !== null) {
            return true;
        }
        const mainJsPath = path.join(extension.path, extension.main);
        try {
            const module = require(mainJsPath);
            extension.module = module;
            return true;
        }
        catch (ex) {
            this._log.warn(`Unable to load ${mainJsPath}. ${ex}`);
            return false;
        }
    }
    _scanPath(extensionPath) {
        if (fs.existsSync(extensionPath)) {
            const result = [];
            const contents = fs.readdirSync(extensionPath);
            for (const item of contents) {
                const packageJsonPath = path.join(extensionPath, item, "package.json");
                if (fs.existsSync(packageJsonPath)) {
                    const extensionInfo = this._loadPackageJson(path.join(extensionPath, item));
                    if (extensionInfo !== null) {
                        result.push(extensionInfo);
                    }
                }
                else {
                    this._log.warn(`Unable to read ${packageJsonPath}, skipping`);
                }
            }
            return result;
        }
        else {
            this._log.warn(`Extension path ${extensionPath} doesn't exist.`);
            return [];
        }
    }
    _loadPackageJson(extensionPath) {
        const packageJsonPath = path.join(extensionPath, "package.json");
        const packageJsonString = fs.readFileSync(packageJsonPath, "UTF8");
        try {
            const packageJson = JSON.parse(packageJsonString);
            if (packageJson.name === undefined) {
                this._log.warn(`${packageJsonPath} didn't contain a 'name' field, skipping.`);
                return null;
            }
            const result = {
                name: packageJson.name,
                path: extensionPath,
                main: packageJson.main !== undefined ? packageJson.main : "main.js",
                version: packageJson.version,
                description: packageJson.description,
                module: null
            };
            return result;
        }
        catch (ex) {
            this._log.warn(`An error occurred while processing ${packageJsonPath}.`, ex);
            return null;
        }
    }
}
exports.ExtensionLoader = ExtensionLoader;
//# sourceMappingURL=ExtensionLoader.js.map