"use strict";
/*
 * Copyright 2014-2016 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * Logging support for inside Electron apps.
 *
 * Instances of this class can be made for different parts of the program
 * code. Logger instances with the same name are made unique by use a counter.
 */
const instanceCounter = new Map();
const instanceNames = new WeakMap(); // Maps objects to the names used by their loggers.
function getLogger(name, instance) {
    return new LoggerImpl(name, instance);
}
exports.getLogger = getLogger;
/**
 * Try to map an object to a human understandable name.
 *
 * @param obj the object
 * @return the name of the object or null if one could not be found.
 */
function objectName(obj) {
    const name = instanceNames.get(obj);
    return name === undefined ? null : name;
}
exports.objectName = objectName;
const logWriters = [];
function addLogWriter(writer) {
    logWriters.push(writer);
}
exports.addLogWriter = addLogWriter;
class LoggerImpl {
    /**
     * Contruct a logger.
     *
     * @param  name the name of the code or class associated with this logger instance.
     * @return the new logger instance
     */
    constructor(name, instance) {
        this._recording = false;
        this._messageLog = [];
        const baseName = name === undefined ? "(unknown)" : name;
        const instanceCount = instanceCounter.has(baseName) ? instanceCounter.get(baseName) + 1 : 0;
        instanceCounter.set(baseName, instanceCount);
        this._name = baseName + " #" + instanceCount;
        if (instance != null) {
            instanceNames.set(instance, this._name);
        }
    }
    getName() {
        return this._name;
    }
    debug(msg, ...opts) {
        this._logMessage("DEBUG", msg, opts);
    }
    info(msg, ...opts) {
        this._logMessage("INFO", msg, opts);
    }
    warn(msg, ...opts) {
        this._logMessage("WARN", msg, opts);
    }
    severe(msg, ...opts) {
        this._logMessage("SEVERE", msg, opts);
    }
    startRecording() {
        this._recording = true;
    }
    stopRecording() {
        this._recording = false;
    }
    startTime(label) {
        console.time(label);
    }
    endTime(label) {
        console.timeEnd(label);
    }
    getFormattedLogMessages() {
        return this._messageLog.reduce((accu, logMessage) => accu + "\n" + logMessage.msg, "");
    }
    _logMessage(level, msg, opts) {
        const formatted = this._format(level, msg);
        if (this._recording) {
            this._messageLog.push({ level, msg: formatted + opts.reduce((x, accu) => accu + x + ", ", "") });
        }
        for (let i = 0; i < logWriters.length; i++) {
            logWriters[i].write(level, formatted, ...opts);
        }
    }
    _format(level, msg) {
        return `${(new Date()).toISOString().replace(/(T|Z)/g, " ").trim()} ${level} [${this._name}] ${msg}`;
    }
}
class ConsoleLogWriter {
    write(level, msg, ...opts) {
        switch (level) {
            case "DEBUG":
            case "INFO":
                console.log(msg, ...opts);
                break;
            case "WARN":
                console.warn(msg, ...opts);
                break;
            case "SEVERE":
                console.error(msg, ...opts);
                break;
        }
    }
}
addLogWriter(new ConsoleLogWriter());
//# sourceMappingURL=Logger.js.map