"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/*
 * Copyright 2014-2016 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */
const DomUtils = require("./DomUtils");
const ThemeableElementBase_1 = require("./ThemeableElementBase");
const TabWidget_1 = require("./gui/TabWidget");
const Terminal_1 = require("./Terminal");
const SettingsTab_1 = require("./settings/SettingsTab");
const AboutTab_1 = require("./AboutTab");
const KeyBindingsTab_1 = require("./KeyBindingsTab");
const ViewerTab_1 = require("./ViewerTab");
const Tab_1 = require("./gui/Tab");
const ViewerElement_1 = require("./ViewerElement");
const Splitter_1 = require("./gui/Splitter");
const SnapDropContainer_1 = require("./gui/SnapDropContainer");
const EmptyPaneMenu_1 = require("./EmptyPaneMenu");
const ViewerElementTypes = require("./ViewerElementTypes");
const ThemeTypes = require("./Theme");
const ResizeRefreshElementBase = require("./ResizeRefreshElementBase");
const WebIpc = require("./WebIpc");
const Messages = require("./WindowMessages");
const _ = require("lodash");
const SplitLayout_1 = require("./SplitLayout");
const SupportsClipboardPaste = require("./SupportsClipboardPaste");
const config = require("./Config");
const he = require("he");
const keybindingmanager = require("./KeyBindingManager");
const GeneralEvents_1 = require("./GeneralEvents");
const InternalMimeTypes_1 = require("./InternalMimeTypes");
const Logger_1 = require("./Logger");
const VisualState = ViewerElementTypes.VisualState;
const ID = "ExtratermMainWebUITemplate";
const ID_TOP_LAYOUT = "ID_TOP_LAYOUT";
const ID_MAIN_CONTENTS = "ID_MAIN_CONTENTS";
const ID_TITLE_BAR = "ID_TITLE_BAR";
const ID_TITLE_BAR_SPACE = "ID_TITLE_BAR_SPACE";
const ID_DRAG_BAR = "ID_DRAG_BAR";
const ID_TOP_RESIZE_BAR = "ID_TOP_RESIZE_BAR";
const ID_MINIMIZE_BUTTON = "ID_MINIMIZE_BUTTON";
const ID_MAXIMIZE_BUTTON = "ID_MAXIMIZE_BUTTON";
const ID_CLOSE_BUTTON = "ID_CLOSE_BUTTON";
const ID_OSX_MINIMIZE_BUTTON = "ID_OSX_MINIMIZE_BUTTON";
const ID_OSX_MAXIMIZE_BUTTON = "ID_OSX_MAXIMIZE_BUTTON";
const ID_OSX_CLOSE_BUTTON = "ID_OSX_CLOSE_BUTTON";
const ID_REST_SLOT = "ID_REST_SLOT";
const ID_REST_DIV_LEFT = "ID_REST_DIV_LEFT";
const CLASS_SPLIT = "split";
const CLASS_TAB_HEADER_CONTAINER = "tab_header_container";
const CLASS_TAB_HEADER_ICON = "tab_header_icon";
const CLASS_TAB_HEADER_MIDDLE = "tab_header_middle";
const CLASS_TAB_HEADER_TAG = "tab_header_tag";
const CLASS_TAB_HEADER_CLOSE = "tab_header_close";
const CLASS_TAB_CONTENT = "tab_content";
const CLASS_NEW_BUTTON_CONTAINER = "CLASS_NEW_BUTTON_CONTAINER";
const CLASS_NEW_TAB_BUTTON = "CLASS_NEW_TAB_BUTTON";
const CLASS_SPACE = "CLASS_SPACE";
const CLASS_MAIN_DRAGGING = "CLASS_MAIN_DRAGGING";
const CLASS_MAIN_NOT_DRAGGING = "CLASS_MAIN_NOT_DRAGGING";
const KEYBINDINGS_MAIN_UI = "main-ui";
const PALETTE_GROUP = "mainwebui";
const COMMAND_SELECT_TAB_LEFT = "selectTabLeft";
const COMMAND_SELECT_TAB_RIGHT = "selectTabRight";
const COMMAND_SELECT_PANE_LEFT = "selectPaneLeft";
const COMMAND_SELECT_PANE_RIGHT = "selectPaneRight";
const COMMAND_SELECT_PANE_ABOVE = "selectPaneAbove";
const COMMAND_SELECT_PANE_BELOW = "selectPaneBelow";
const COMMAND_MOVE_TAB_LEFT = "moveTabLeft";
const COMMAND_MOVE_TAB_RIGHT = "moveTabRight";
const COMMAND_MOVE_TAB_UP = "moveTabUp";
const COMMAND_MOVE_TAB_DOWN = "moveTabDown";
const COMMAND_NEW_TERMINAL = "newTerminal";
const COMMAND_CLOSE_TAB = "closeTab";
const COMMAND_HORIZONTAL_SPLIT = "horizontalSplit";
const COMMAND_VERTICAL_SPLIT = "verticalSplit";
const COMMAND_CLOSE_PANE = "closePane";
let registered = false;
const LAST_FOCUS = "last_focus";
const staticLog = Logger_1.getLogger("Static ExtratermMainWebUI");
// Theme management
const activeInstances = new Set();
let themeCss = "";
/**
 * Top level UI component for a normal terminal window
 *
 */
class MainWebUi extends ThemeableElementBase_1.ThemeableElementBase {
    //-----------------------------------------------------------------------
    //
    //   #                                                         
    //   #       # ###### ######  ####  #   #  ####  #      ###### 
    //   #       # #      #      #    #  # #  #    # #      #      
    //   #       # #####  #####  #        #   #      #      #####  
    //   #       # #      #      #        #   #      #      #      
    //   #       # #      #      #    #   #   #    # #      #      
    //   ####### # #      ######  ####    #    ####  ###### ###### 
    //
    //-----------------------------------------------------------------------
    constructor() {
        super();
        this._initProperties(); // Initialise our properties. The constructor was not called.
    }
    //-----------------------------------------------------------------------
    // Statics
    static init() {
        Tab_1.Tab.init();
        TabWidget_1.TabWidget.init();
        Terminal_1.EtTerminal.init();
        SettingsTab_1.SettingsTab.init();
        KeyBindingsTab_1.EtKeyBindingsTab.init();
        AboutTab_1.AboutTab.init();
        ViewerTab_1.EtViewerTab.init();
        Splitter_1.Splitter.init();
        EmptyPaneMenu_1.EmptyPaneMenu.init();
        if (registered === false) {
            window.customElements.define(MainWebUi.TAG_NAME.toLowerCase(), MainWebUi);
            registered = true;
        }
    }
    _initProperties() {
        this._log = Logger_1.getLogger("ExtratermMainWebUI", this);
        this._terminalPtyIdMap = new Map();
        this._ptyIdTerminalMap = new Map();
        this._lastFocus = null;
        this._tabIdCounter = 0;
        this._configManager = null;
        this._keyBindingManager = null;
        this._themes = [];
        this._splitLayout = new SplitLayout_1.SplitLayout();
    }
    focus() {
        if (this._lastFocus != null) {
            this._focusTabContent(this._lastFocus);
        }
        else {
            const allContentElements = this._splitLayout.getAllTabContents();
            if (allContentElements.length !== 0) {
                this._focusTabContent(allContentElements[0]);
            }
        }
    }
    setConfigManager(configManager) {
        this._configManager = configManager;
    }
    setKeyBindingManager(keyBindingManager) {
        this._keyBindingManager = keyBindingManager;
    }
    setThemes(themes) {
        this._themes = themes;
    }
    getTabCount() {
        return this._splitLayout.getAllTabContents().filter((el) => !(el instanceof EmptyPaneMenu_1.EmptyPaneMenu)).length;
    }
    refresh(level) {
        this._refresh(level);
    }
    connectedCallback() {
        super.connectedCallback();
        this._setUpShadowDom();
        this._setUpMainContainer();
        this._setUpSplitLayout();
        this._setUpWindowControls();
        this._setupIpc();
    }
    _setUpShadowDom() {
        const shadow = this.attachShadow({ mode: 'open', delegatesFocus: true });
        const clone = this._createClone();
        shadow.appendChild(clone);
        this.updateThemeCss();
    }
    _setUpMainContainer() {
        const mainContainer = DomUtils.getShadowId(this, ID_MAIN_CONTENTS);
        mainContainer.classList.add(CLASS_MAIN_NOT_DRAGGING);
        mainContainer.addEventListener(TabWidget_1.TabWidget.EVENT_TAB_SWITCH, this._handleTabSwitchEvent.bind(this));
        mainContainer.addEventListener(TabWidget_1.TabWidget.EVENT_DROPPED, this._handleTabWidgetDroppedEvent.bind(this));
        mainContainer.addEventListener(SnapDropContainer_1.SnapDropContainer.EVENT_DROPPED, this._handleTabWidgetSnapDroppedEvent.bind(this));
        DomUtils.addCustomEventResender(mainContainer, GeneralEvents_1.EVENT_DRAG_STARTED, this);
        DomUtils.addCustomEventResender(mainContainer, GeneralEvents_1.EVENT_DRAG_ENDED, this);
        mainContainer.addEventListener(GeneralEvents_1.EVENT_DRAG_STARTED, this._handleDragStartedEvent.bind(this));
        mainContainer.addEventListener(GeneralEvents_1.EVENT_DRAG_ENDED, this._handleDragEndedEvent.bind(this));
        mainContainer.addEventListener('click', this._handleMainContainerClickEvent.bind(this));
    }
    _handleTabWidgetDroppedEvent(ev) {
        const detail = ev.detail;
        if (detail.mimeType === InternalMimeTypes_1.ElementMimeType.MIMETYPE) {
            this._handleElementDroppedEvent(detail.targetTabWidget, detail.tabIndex, detail.dropData);
        }
        else if (detail.mimeType === InternalMimeTypes_1.FrameMimeType.MIMETYPE) {
            this._handleFrameDroppedEvent(detail.targetTabWidget, detail.tabIndex, detail.dropData);
        }
    }
    _handleElementDroppedEvent(targetTabWidget, tabIndex, dropData) {
        if (InternalMimeTypes_1.ElementMimeType.tagNameFromData(dropData) === Tab_1.Tab.TAG_NAME) {
            const tabElement = DomUtils.getShadowId(this, InternalMimeTypes_1.ElementMimeType.elementIdFromData(dropData));
            this._splitLayout.moveTabToTabWidget(tabElement, targetTabWidget, tabIndex);
            this._splitLayout.update();
            const tabContent = this._splitLayout.getTabContentByTab(tabElement);
            targetTabWidget.setSelectedIndex(tabIndex);
            this._focusTabContent(tabContent);
        }
    }
    _handleFrameDroppedEvent(targetTabWidget, tabIndex, dropData) {
        for (const el of this._splitLayout.getAllTabContents()) {
            if (el instanceof Terminal_1.EtTerminal) {
                const embeddedViewer = el.getEmbeddedViewerByFrameId(dropData);
                if (embeddedViewer != null) {
                    const viewerTab = this._popOutEmbeddedViewer(embeddedViewer, el);
                    const tab = this._splitLayout.getTabByTabContent(viewerTab);
                    this._splitLayout.moveTabToTabWidget(tab, targetTabWidget, tabIndex);
                    this._splitLayout.update();
                    this._switchToTab(viewerTab);
                    return;
                }
            }
        }
    }
    _handleTabWidgetSnapDroppedEvent(ev) {
        const detail = ev.detail;
        const target = ev.target;
        if (target instanceof TabWidget_1.TabWidget) {
            switch (detail.dropLocation) {
                case SnapDropContainer_1.DropLocation.MIDDLE:
                    this._handleTabWidgetSnapDroppedMiddleEvent(ev);
                    break;
                case SnapDropContainer_1.DropLocation.NORTH:
                    this._handleTabWidgetSnapDroppedDirectionEvent(ev, Splitter_1.SplitOrientation.HORIZONTAL, true);
                    break;
                case SnapDropContainer_1.DropLocation.SOUTH:
                    this._handleTabWidgetSnapDroppedDirectionEvent(ev, Splitter_1.SplitOrientation.HORIZONTAL, false);
                    break;
                case SnapDropContainer_1.DropLocation.WEST:
                    this._handleTabWidgetSnapDroppedDirectionEvent(ev, Splitter_1.SplitOrientation.VERTICAL, true);
                    break;
                case SnapDropContainer_1.DropLocation.EAST:
                    this._handleTabWidgetSnapDroppedDirectionEvent(ev, Splitter_1.SplitOrientation.VERTICAL, false);
                    break;
            }
        }
    }
    _handleTabWidgetSnapDroppedMiddleEvent(ev) {
        const detail = ev.detail;
        const target = ev.target;
        if (target instanceof TabWidget_1.TabWidget) {
            if (detail.mimeType === InternalMimeTypes_1.ElementMimeType.MIMETYPE) {
                this._handleElementDroppedEvent(target, target.getSelectedIndex() + 1, detail.dropData);
            }
            else if (detail.mimeType === InternalMimeTypes_1.FrameMimeType.MIMETYPE) {
                this._handleFrameDroppedEvent(target, target.getSelectedIndex() + 1, detail.dropData);
            }
        }
    }
    _handleTabWidgetSnapDroppedDirectionEvent(ev, orientation, splitBefore) {
        const detail = ev.detail;
        const target = ev.target;
        if (target instanceof TabWidget_1.TabWidget) {
            const newTabWidget = splitBefore ? this._splitLayout.splitBeforeTabWidget(target, orientation) : this._splitLayout.splitAfterTabWidget(target, orientation);
            this._splitLayout.update();
            if (detail.mimeType === InternalMimeTypes_1.ElementMimeType.MIMETYPE) {
                this._handleElementDroppedEvent(newTabWidget, 0, detail.dropData);
            }
            else if (detail.mimeType === InternalMimeTypes_1.FrameMimeType.MIMETYPE) {
                this._handleFrameDroppedEvent(newTabWidget, 0, detail.dropData);
            }
            this._refreshSplitLayout();
        }
    }
    _handleDragStartedEvent(ev) {
        const mainContainer = DomUtils.getShadowId(this, ID_MAIN_CONTENTS);
        mainContainer.classList.add(CLASS_MAIN_DRAGGING);
        mainContainer.classList.remove(CLASS_MAIN_NOT_DRAGGING);
    }
    _handleDragEndedEvent(ev) {
        const mainContainer = DomUtils.getShadowId(this, ID_MAIN_CONTENTS);
        mainContainer.classList.remove(CLASS_MAIN_DRAGGING);
        mainContainer.classList.add(CLASS_MAIN_NOT_DRAGGING);
    }
    _handleMainContainerClickEvent(ev) {
        for (const part of ev.path) {
            if (part instanceof HTMLButtonElement) {
                if (part.classList.contains(CLASS_NEW_TAB_BUTTON)) {
                    let el = part;
                    while (el != null && !(el instanceof TabWidget_1.TabWidget)) {
                        el = el.parentElement;
                    }
                    const newTerminal = this.newTerminalTab(el);
                    this._switchToTab(newTerminal);
                }
            }
        }
    }
    _setUpSplitLayout() {
        const mainContainer = DomUtils.getShadowId(this, ID_MAIN_CONTENTS);
        this._splitLayout.setRootContainer(mainContainer);
        this._splitLayout.setTabContainerFactory((tabWidget, tab, tabContent) => {
            const divContainer = document.createElement("DIV");
            divContainer.classList.add(CLASS_TAB_CONTENT);
            divContainer.addEventListener('keydown', this._handleKeyDownCapture.bind(this, tabContent), true);
            return divContainer;
        });
        this._splitLayout.setRightSpaceDefaultElementFactory(() => {
            const tempDiv = document.createElement("DIV");
            tempDiv.innerHTML = this._newTabRestAreaHtml();
            return tempDiv.children.item(0);
        });
        this._splitLayout.setTopLeftElement(this._leftControls());
        this._splitLayout.setTopRightElement(this._menuControls());
        this._splitLayout.setEmptySplitElementFactory(() => {
            const emptyPaneMenu = document.createElement(EmptyPaneMenu_1.EmptyPaneMenu.TAG_NAME);
            const commandList = [
                { id: COMMAND_NEW_TERMINAL, group: PALETTE_GROUP, iconRight: "plus", label: "New Terminal", commandExecutor: null },
                { id: COMMAND_HORIZONTAL_SPLIT, group: PALETTE_GROUP, iconRight: "extraicon-#xea08", label: "Horizontal Split", commandExecutor: null },
                { id: COMMAND_VERTICAL_SPLIT, group: PALETTE_GROUP, iconRight: "columns", label: "Vertical Split", commandExecutor: null },
                { id: COMMAND_CLOSE_PANE, group: PALETTE_GROUP, label: "Close Pane", commandExecutor: null }
            ];
            this._insertCommandKeyBindings(commandList);
            emptyPaneMenu.setEntries(commandList);
            emptyPaneMenu.addEventListener("selected", (ev) => {
                emptyPaneMenu.setFilter("");
                this.executeCommand(ev.detail.selected, { tabElement: emptyPaneMenu });
            });
            return emptyPaneMenu;
        });
    }
    _setUpWindowControls() {
        DomUtils.getShadowId(this, ID_MINIMIZE_BUTTON).addEventListener('click', () => {
            this.focus();
            this._sendWindowRequestEvent(MainWebUi.EVENT_MINIMIZE_WINDOW_REQUEST);
        });
        DomUtils.getShadowId(this, ID_MAXIMIZE_BUTTON).addEventListener('click', () => {
            this.focus();
            this._sendWindowRequestEvent(MainWebUi.EVENT_MAXIMIZE_WINDOW_REQUEST);
        });
        DomUtils.getShadowId(this, ID_CLOSE_BUTTON).addEventListener('click', () => {
            this.focus();
            this._sendWindowRequestEvent(MainWebUi.EVENT_CLOSE_WINDOW_REQUEST);
        });
    }
    _menuControls() {
        const tempDiv = document.createElement("DIV");
        tempDiv.innerHTML = this._newTabRestAreaHtml(`<slot id="${ID_REST_SLOT}"></slot>`);
        return tempDiv.children.item(0);
    }
    _leftControls() {
        const tempDiv = document.createElement("DIV");
        tempDiv.innerHTML =
            `<div id="${ID_REST_DIV_LEFT}">` +
                `<button id="${ID_OSX_CLOSE_BUTTON}" tabindex="-1"></button>` +
                `<button id="${ID_OSX_MINIMIZE_BUTTON}" tabindex="-1"></button>` +
                `<button id="${ID_OSX_MAXIMIZE_BUTTON}" tabindex="-1"></button>` +
                `</div>`;
        tempDiv.querySelector("#" + ID_OSX_MINIMIZE_BUTTON).addEventListener('click', () => {
            this.focus();
            this._sendWindowRequestEvent(MainWebUi.EVENT_MINIMIZE_WINDOW_REQUEST);
        });
        tempDiv.querySelector("#" + ID_OSX_MAXIMIZE_BUTTON).addEventListener('click', () => {
            this.focus();
            this._sendWindowRequestEvent(MainWebUi.EVENT_MAXIMIZE_WINDOW_REQUEST);
        });
        tempDiv.querySelector("#" + ID_OSX_CLOSE_BUTTON).addEventListener('click', () => {
            this.focus();
            this._sendWindowRequestEvent(MainWebUi.EVENT_CLOSE_WINDOW_REQUEST);
        });
        return tempDiv.children.item(0);
    }
    _createClone() {
        var template = window.document.getElementById(ID);
        if (template === null) {
            template = window.document.createElement('template');
            template.id = ID;
            template.innerHTML = this._html();
            window.document.body.appendChild(template);
        }
        return window.document.importNode(template.content, true);
    }
    _html() {
        return `
    <style id="${ThemeableElementBase_1.ThemeableElementBase.ID_THEME}"></style>` +
            `<div id="${ID_TOP_LAYOUT}">` +
            `<div id="${ID_TITLE_BAR}">` +
            `<div id="${ID_TITLE_BAR_SPACE}">` +
            `<div id="${ID_TOP_RESIZE_BAR}"></div>` +
            `<div id="${ID_DRAG_BAR}"></div>` +
            `</div>` +
            `<button id="${ID_MINIMIZE_BUTTON}" tabindex="-1"></button>` +
            `<button id="${ID_MAXIMIZE_BUTTON}" tabindex="-1"></button>` +
            `<button id="${ID_CLOSE_BUTTON}" tabindex="-1"></button>` +
            `</div>` +
            `<div id="${ID_MAIN_CONTENTS}">` +
            `</div>` +
            `</div>`;
    }
    _newTabRestAreaHtml(extraContents = "") {
        return `<div class="${CLASS_NEW_BUTTON_CONTAINER}"><button class="btn btn-quiet ${CLASS_NEW_TAB_BUTTON}"><i class="fa fa-plus"></i></button>
    <div class="${CLASS_SPACE}"></div>${extraContents}</div>`;
    }
    _themeCssFiles() {
        return [ThemeTypes.CssFile.GUI_CONTROLS, ThemeTypes.CssFile.FONT_AWESOME, ThemeTypes.CssFile.MAIN_UI];
    }
    destroy() {
    }
    // ----------------------------------------------------------------------
    //
    //   #######                      
    //      #      ##   #####   ####  
    //      #     #  #  #    # #      
    //      #    #    # #####   ####  
    //      #    ###### #    #      # 
    //      #    #    # #    # #    # 
    //      #    #    # #####   ####  
    //
    // ----------------------------------------------------------------------
    /**
     * Initialise and insert a tab.
     *
     */
    _addTab(tabWidget, tabContentElement) {
        const newId = this._tabIdCounter;
        this._tabIdCounter++;
        const newTab = document.createElement(Tab_1.Tab.TAG_NAME);
        newTab.setAttribute('id', "tab_id_" + newId);
        newTab.tabIndex = -1;
        newTab.innerHTML =
            `<div class="${CLASS_TAB_HEADER_CONTAINER}">` +
                `<div class="${CLASS_TAB_HEADER_ICON}"></div>` +
                `<div class="${CLASS_TAB_HEADER_MIDDLE}">${newId}</div>` +
                `<div class="${CLASS_TAB_HEADER_TAG}"></div>` +
                `<div class="${CLASS_TAB_HEADER_CLOSE}">` +
                `<button id="close_tag_id_${newId}"><i class="fa fa-times"></i></button>` +
                `</div>` +
                `</div>`;
        this._splitLayout.appendTab(tabWidget, newTab, tabContentElement);
        this._splitLayout.update();
        const closeTabButton = DomUtils.getShadowRoot(this).getElementById("close_tag_id_" + newId);
        closeTabButton.addEventListener('click', (ev) => {
            this.closeTab(tabContentElement);
        });
        return newTab;
    }
    _tabFromElement(el) {
        return this._splitLayout.getTabByTabContent(el);
    }
    _tabWidgetFromElement(el) {
        return this._splitLayout.getTabWidgetByTabContent(el);
    }
    _firstTabWidget() {
        return this._splitLayout.firstTabWidget();
    }
    _handleTabSwitchEvent(ev) {
        if (ev.target instanceof TabWidget_1.TabWidget) {
            const el = this._splitLayout.getTabContentByTab(ev.target.getSelectedTab());
            let title = "";
            if (el instanceof Terminal_1.EtTerminal) {
                title = el.getTerminalTitle();
            }
            else if (el instanceof ViewerTab_1.EtViewerTab) {
                title = el.getTitle();
            }
            else if (el instanceof ViewerElement_1.ViewerElement) {
                title = el.getTitle();
            }
            this._sendTitleEvent(title);
            this._focusTabContent(el);
        }
    }
    newTerminalTab(tabWidget = null) {
        if (tabWidget == null) {
            tabWidget = this._splitLayout.firstTabWidget();
        }
        const newTerminal = document.createElement(Terminal_1.EtTerminal.TAG_NAME);
        config.injectConfigManager(newTerminal, this._configManager);
        keybindingmanager.injectKeyBindingManager(newTerminal, this._keyBindingManager);
        newTerminal.setFrameFinder(this._frameFinder.bind(this));
        this._terminalPtyIdMap.set(newTerminal, null);
        this._applyNewTerminalConfig(newTerminal, this._configManager.getConfig());
        this._addTab(tabWidget, newTerminal);
        this._setUpNewTerminalEventHandlers(newTerminal);
        this._createPtyForTerminal(newTerminal);
        this._updateTabTitle(newTerminal);
        this._sendTabOpenedEvent();
        newTerminal.refresh(ResizeRefreshElementBase.RefreshLevel.COMPLETE);
        return newTerminal;
    }
    _applyNewTerminalConfig(newTerminal, config) {
        newTerminal.setBlinkingCursor(config.blinkingCursor);
        newTerminal.setScrollbackSize(config.scrollbackLines);
    }
    _createPtyForTerminal(newTerminal) {
        const sessionProfile = this._configManager.getConfig().expandedProfiles[0];
        const newEnv = _.cloneDeep(process.env);
        const expandedExtra = sessionProfile.extraEnv;
        let prop;
        for (prop in expandedExtra) {
            newEnv[prop] = expandedExtra[prop];
        }
        WebIpc.requestPtyCreate(sessionProfile.command, sessionProfile.arguments, newTerminal.getColumns(), newTerminal.getRows(), newEnv)
            .then((msg) => {
            this._terminalPtyIdMap.set(newTerminal, msg.id);
            this._ptyIdTerminalMap.set(msg.id, newTerminal);
            WebIpc.ptyResize(msg.id, newTerminal.getColumns(), newTerminal.getRows());
            WebIpc.ptyOutputBufferSize(msg.id, 1024); // Just big enough to get things started. We don't need the exact buffer size.
        });
    }
    _setUpNewTerminalEventHandlers(newTerminal) {
        newTerminal.addEventListener('focus', (ev) => {
            this._lastFocus = newTerminal;
        });
        newTerminal.addEventListener(Terminal_1.EtTerminal.EVENT_USER_INPUT, (ev) => {
            const ptyId = this._terminalPtyIdMap.get(newTerminal);
            if (ptyId != null) {
                WebIpc.ptyInput(ptyId, ev.detail.data);
            }
        });
        newTerminal.addEventListener(Terminal_1.EtTerminal.EVENT_TERMINAL_RESIZE, (ev) => {
            const currentColumns = ev.detail.columns;
            const currentRows = ev.detail.rows;
            const ptyId = this._terminalPtyIdMap.get(newTerminal);
            if (ptyId != null) {
                WebIpc.ptyResize(ptyId, currentColumns, currentRows);
            }
        });
        newTerminal.addEventListener(Terminal_1.EtTerminal.EVENT_TERMINAL_BUFFER_SIZE, (ev) => {
            const status = ev.detail;
            const ptyId = this._terminalPtyIdMap.get(newTerminal);
            if (ptyId != null) {
                WebIpc.ptyOutputBufferSize(ptyId, status.bufferSize);
            }
        });
        newTerminal.addEventListener(Terminal_1.EtTerminal.EVENT_TITLE, (ev) => {
            this._updateTabTitle(newTerminal);
            this._sendTitleEvent(ev.detail.title);
        });
        newTerminal.addEventListener(Terminal_1.EtTerminal.EVENT_EMBEDDED_VIEWER_POP_OUT, this._handleEmbeddedViewerPopOutEvent.bind(this));
    }
    _handleEmbeddedViewerPopOutEvent(ev) {
        this._popOutEmbeddedViewer(ev.detail.embeddedViewer, ev.detail.terminal);
    }
    _popOutEmbeddedViewer(embeddedViewer, terminal) {
        const viewerTab = this.openViewerTab(embeddedViewer, terminal.getFontAdjust());
        this._switchToTab(viewerTab);
        terminal.deleteEmbeddedViewer(embeddedViewer);
        return viewerTab;
    }
    openViewerTab(embeddedViewer, fontAdjust) {
        const viewerElement = embeddedViewer.getViewerElement();
        const viewerTab = document.createElement(ViewerTab_1.EtViewerTab.TAG_NAME);
        viewerTab.setFontAdjust(fontAdjust);
        keybindingmanager.injectKeyBindingManager(viewerTab, this._keyBindingManager);
        viewerTab.setTitle(embeddedViewer.getTitle());
        viewerTab.setTag(embeddedViewer.getTag());
        viewerElement.setMode(ViewerElementTypes.Mode.CURSOR);
        viewerElement.setVisualState(VisualState.AUTO);
        const result = this._openViewerTab(this._firstTabWidget(), viewerTab);
        viewerTab.setViewerElement(viewerElement);
        this._updateTabTitle(viewerTab);
        return viewerTab;
    }
    _openViewerTab(tabWidget, viewerElement) {
        viewerElement.setFocusable(true);
        this._addTab(tabWidget, viewerElement);
        viewerElement.addEventListener('focus', (ev) => {
            this._lastFocus = viewerElement;
        });
        this._updateTabTitle(viewerElement);
        this._sendTabOpenedEvent();
    }
    _updateTabTitle(el) {
        const tab = this._splitLayout.getTabByTabContent(el);
        let title = "";
        let htmlTitle = "";
        let icon = null;
        let tag = "";
        if (el instanceof Terminal_1.EtTerminal) {
            title = el.getTerminalTitle();
            htmlTitle = he.escape(title);
            icon = "keyboard-o";
        }
        else if (el instanceof ViewerTab_1.EtViewerTab) {
            title = el.getTitle();
            htmlTitle = he.escape(title);
            icon = el.getAwesomeIcon();
            if (el.getTag() !== null) {
                tag = "<i class='fa fa-tag'></i>" + el.getTag();
            }
        }
        else if (el instanceof ViewerElement_1.ViewerElement) {
            title = el.getTitle();
            htmlTitle = he.escape(title);
            icon = el.getAwesomeIcon();
        }
        else {
            this._log.warn(`Unrecognized element type in _updateTabTitle(). ${el}`);
        }
        const iconDiv = tab.querySelector(`DIV.${CLASS_TAB_HEADER_ICON}`);
        iconDiv.innerHTML = icon !== null ? '<i class="fa fa-' + icon + '"></i>' : "";
        const middleDiv = tab.querySelector(`DIV.${CLASS_TAB_HEADER_MIDDLE}`);
        middleDiv.title = title;
        middleDiv.innerHTML = htmlTitle;
        const tabDiv = tab.querySelector(`DIV.${CLASS_TAB_HEADER_TAG}`);
        tabDiv.innerHTML = tag;
    }
    openSettingsTab() {
        const settingsTabs = this._splitLayout.getAllTabContents().filter((el) => el instanceof SettingsTab_1.SettingsTab);
        if (settingsTabs.length !== 0) {
            this._switchToTab(settingsTabs[0]);
        }
        else {
            const viewerElement = document.createElement(SettingsTab_1.SettingsTab.TAG_NAME);
            config.injectConfigManager(viewerElement, this._configManager);
            keybindingmanager.injectKeyBindingManager(viewerElement, this._keyBindingManager);
            viewerElement.setThemes(this._themes);
            this._openViewerTab(this._firstTabWidget(), viewerElement);
            this._switchToTab(viewerElement);
        }
    }
    openKeyBindingsTab() {
        const keyBindingsTabs = this._splitLayout.getAllTabContents().filter((el) => el instanceof KeyBindingsTab_1.EtKeyBindingsTab);
        if (keyBindingsTabs.length !== 0) {
            this._switchToTab(keyBindingsTabs[0]);
        }
        else {
            const viewerElement = document.createElement(KeyBindingsTab_1.EtKeyBindingsTab.TAG_NAME);
            config.injectConfigManager(viewerElement, this._configManager);
            keybindingmanager.injectKeyBindingManager(viewerElement, this._keyBindingManager);
            this._openViewerTab(this._firstTabWidget(), viewerElement);
            this._switchToTab(viewerElement);
        }
    }
    openAboutTab() {
        const aboutTabs = this._splitLayout.getAllTabContents().filter((el) => el instanceof AboutTab_1.AboutTab);
        if (aboutTabs.length !== 0) {
            this._switchToTab(aboutTabs[0]);
        }
        else {
            const viewerElement = document.createElement(AboutTab_1.AboutTab.TAG_NAME);
            keybindingmanager.injectKeyBindingManager(viewerElement, this._keyBindingManager);
            this._openViewerTab(this._firstTabWidget(), viewerElement);
            this._switchToTab(viewerElement);
        }
    }
    /**
     *
     */
    closeTab(tabContentElement) {
        const tabWidget = this._splitLayout.getTabWidgetByTabContent(tabContentElement);
        const tabWidgetContents = this._splitLayout.getTabContentsByTabWidget(tabWidget);
        this._splitLayout.removeTabContent(tabContentElement);
        this._splitLayout.update();
        if (tabContentElement instanceof Terminal_1.EtTerminal) {
            const ptyId = this._terminalPtyIdMap.get(tabContentElement);
            tabContentElement.destroy();
            if (ptyId !== null) {
                WebIpc.ptyClose(ptyId);
            }
            this._terminalPtyIdMap.delete(tabContentElement);
            this._ptyIdTerminalMap.delete(ptyId);
        }
        const oldIndex = tabWidgetContents.indexOf(tabContentElement);
        if (tabWidgetContents.length >= 2) {
            this._switchToTab(tabWidgetContents[oldIndex === 0 ? 1 : oldIndex - 1]);
        }
        else {
            this._switchToTab(this._splitLayout.getTabContentsByTabWidget(tabWidget)[0]);
        }
        this._sendTabClosedEvent();
    }
    _switchToTab(tabContentElement) {
        this._splitLayout.showTabByTabContent(tabContentElement);
        this._focusTabContent(tabContentElement);
    }
    _shiftTab(tabWidget, direction) {
        const contents = this._splitLayout.getTabContentsByTabWidget(tabWidget);
        const len = contents.length;
        if (len === 0) {
            return;
        }
        let i = tabWidget.getSelectedIndex();
        i = i + direction;
        if (i < 0) {
            i = len - 1;
        }
        else if (i >= len) {
            i = 0;
        }
        tabWidget.setSelectedIndex(i);
        this._focusTabContent(contents[i]);
    }
    focusPane() {
        // FIXME
    }
    _selectPaneLeft(tabElement) {
        return this._selectPaneInDirection(tabElement, this._splitLayout.getTabWidgetToLeft);
    }
    _selectPaneRight(tabElement) {
        return this._selectPaneInDirection(tabElement, this._splitLayout.getTabWidgetToRight);
    }
    _selectPaneAbove(tabElement) {
        return this._selectPaneInDirection(tabElement, this._splitLayout.getTabWidgetAbove);
    }
    _selectPaneBelow(tabElement) {
        return this._selectPaneInDirection(tabElement, this._splitLayout.getTabWidgetBelow);
    }
    _selectPaneInDirection(tabElement, directionFunc) {
        const currentTabWidget = this._splitLayout.getTabWidgetByTabContent(tabElement);
        const targetTabWidget = directionFunc.call(this._splitLayout, currentTabWidget);
        if (targetTabWidget != null) {
            targetTabWidget.focus();
            const content = this._splitLayout.getTabContentByTab(targetTabWidget.getSelectedTab());
            if (elementSupportsFocus(content)) {
                content.focus();
                return { tabWidget: targetTabWidget, tabContent: content };
            }
            return { tabWidget: targetTabWidget, tabContent: null };
        }
        return { tabWidget: null, tabContent: null };
    }
    _moveTabElementLeft(tabElement) {
        this._moveTabElementInDirection(tabElement, this._splitLayout.getTabWidgetToLeft);
    }
    _moveTabElementRight(tabElement) {
        this._moveTabElementInDirection(tabElement, this._splitLayout.getTabWidgetToRight);
    }
    _moveTabElementUp(tabElement) {
        this._moveTabElementInDirection(tabElement, this._splitLayout.getTabWidgetAbove);
    }
    _moveTabElementDown(tabElement) {
        this._moveTabElementInDirection(tabElement, this._splitLayout.getTabWidgetBelow);
    }
    _moveTabElementInDirection(tabElement, directionFunc) {
        const currentTabWidget = this._splitLayout.getTabWidgetByTabContent(tabElement);
        const targetTabWidget = directionFunc.call(this._splitLayout, currentTabWidget);
        if (targetTabWidget != null) {
            this._splitLayout.moveTabToTabWidget(this._splitLayout.getTabByTabContent(tabElement), targetTabWidget, 0);
            this._splitLayout.update();
            targetTabWidget.focus();
            if (elementSupportsFocus(tabElement)) {
                tabElement.focus();
            }
        }
    }
    _getTabElementWithFocus() {
        for (const el of this._splitLayout.getAllTabContents()) {
            if (elementSupportsFocus(el)) {
                if (el.hasFocus()) {
                    return el;
                }
            }
        }
        return null;
    }
    _focusTabContent(el) {
        if (el instanceof Terminal_1.EtTerminal) {
            el.resizeToContainer();
            el.focus();
        }
        else if (elementSupportsFocus(el)) {
            el.focus();
        }
    }
    _horizontalSplit(tabContentElement) {
        this._split(tabContentElement, Splitter_1.SplitOrientation.HORIZONTAL);
    }
    _verticalSplit(tabContentElement) {
        this._split(tabContentElement, Splitter_1.SplitOrientation.VERTICAL);
    }
    _split(tabContentElement, orientation) {
        const newTabWidget = this._splitLayout.splitAfterTabContent(tabContentElement, orientation);
        this._splitLayout.update();
        this._refreshSplitLayout();
        if (newTabWidget != null) {
            const element = this._splitLayout.getEmptyContentByTabWidget(newTabWidget);
            if (element != null) {
                newTabWidget.focus();
                if (element instanceof EmptyPaneMenu_1.EmptyPaneMenu) {
                    element.focus();
                }
            }
            else {
                const tabWidget = this._splitLayout.getTabWidgetByTabContent(tabContentElement);
                tabWidget.focus();
                if (elementSupportsFocus(tabContentElement)) {
                    tabContentElement.focus();
                }
            }
        }
    }
    _refreshSplitLayout() {
        const rootContainer = DomUtils.getShadowId(this, ID_MAIN_CONTENTS);
        const firstKid = rootContainer.children.item(0);
        if (firstKid instanceof Splitter_1.Splitter || firstKid instanceof TabWidget_1.TabWidget) {
            firstKid.refresh(ResizeRefreshElementBase.RefreshLevel.COMPLETE);
        }
    }
    _closeSplit(tabContentElement) {
        let focusInfo = null;
        if (tabContentElement instanceof EmptyPaneMenu_1.EmptyPaneMenu) {
            focusInfo = this._selectPaneLeft(tabContentElement);
            if (focusInfo.tabWidget == null) {
                focusInfo = this._selectPaneRight(tabContentElement);
                if (focusInfo.tabWidget == null) {
                    focusInfo = this._selectPaneAbove(tabContentElement);
                    if (focusInfo.tabWidget == null) {
                        focusInfo = this._selectPaneBelow(tabContentElement);
                    }
                }
            }
        }
        this._splitLayout.closeSplitAtTabContent(tabContentElement);
        this._splitLayout.update();
        this._refreshSplitLayout();
        if (focusInfo == null) {
            const tabWidget = this._splitLayout.getTabWidgetByTabContent(tabContentElement);
            focusInfo = { tabWidget, tabContent: tabContentElement };
        }
        if (focusInfo.tabWidget != null) {
            focusInfo.tabWidget.focus();
            if (focusInfo.tabContent != null) {
                if (elementSupportsFocus(focusInfo.tabContent)) {
                    focusInfo.tabContent.focus();
                }
            }
        }
    }
    //-----------------------------------------------------------------------
    //
    //    #####                                                     
    //   #     # #      # #####  #####   ####    ##   #####  #####  
    //   #       #      # #    # #    # #    #  #  #  #    # #    # 
    //   #       #      # #    # #####  #    # #    # #    # #    # 
    //   #       #      # #####  #    # #    # ###### #####  #    # 
    //   #     # #      # #      #    # #    # #    # #   #  #    # 
    //    #####  ###### # #      #####   ####  #    # #    # #####  
    //
    //-----------------------------------------------------------------------
    /**
     * Copys the selection in the focussed terminal to the clipboard.
     */
    copyToClipboard() {
        const elWithFocus = this._getTabElementWithFocus();
        if (elWithFocus != null) {
            if (elWithFocus instanceof Terminal_1.EtTerminal || elWithFocus instanceof ViewerTab_1.EtViewerTab) {
                elWithFocus.copyToClipboard();
            }
        }
    }
    /**
     * Pastes text into the terminal which has the input focus.
     *
     * @param text the text to paste.
     */
    pasteText(text) {
        const elWithFocus = this._getTabElementWithFocus();
        if (elWithFocus != null && SupportsClipboardPaste.isSupportsClipboardPaste(elWithFocus)) {
            elWithFocus.pasteText(text);
        }
    }
    //-----------------------------------------------------------------------
    _refresh(level) {
        ResizeRefreshElementBase.ResizeRefreshElementBase.refreshChildNodes(DomUtils.getShadowId(this, ID_MAIN_CONTENTS), level);
    }
    _sendTabOpenedEvent() {
        const event = new CustomEvent(MainWebUi.EVENT_TAB_OPENED, { detail: null });
        this.dispatchEvent(event);
    }
    _sendTabClosedEvent() {
        const event = new CustomEvent(MainWebUi.EVENT_TAB_CLOSED, { detail: null });
        this.dispatchEvent(event);
    }
    _sendTitleEvent(title) {
        const event = new CustomEvent(MainWebUi.EVENT_TITLE, { detail: { title: title } });
        this.dispatchEvent(event);
    }
    _sendWindowRequestEvent(eventName) {
        const event = new CustomEvent(eventName, {});
        this.dispatchEvent(event);
    }
    _frameFinder(frameId) {
        for (const el of this._splitLayout.getAllTabContents()) {
            let text = null;
            if (el instanceof ViewerTab_1.EtViewerTab && el.getTag() === frameId) {
                text = el.getFrameContents(frameId);
            }
            else if (el instanceof Terminal_1.EtTerminal) {
                text = el.getFrameContents(frameId);
            }
            if (text != null) {
                return text;
            }
        }
        return null;
    }
    // ----------------------------------------------------------------------
    //
    //   #    #                                                 
    //   #   #  ###### #   # #####   ####    ##   #####  #####  
    //   #  #   #       # #  #    # #    #  #  #  #    # #    # 
    //   ###    #####    #   #####  #    # #    # #    # #    # 
    //   #  #   #        #   #    # #    # ###### #####  #    # 
    //   #   #  #        #   #    # #    # #    # #   #  #    # 
    //   #    # ######   #   #####   ####  #    # #    # #####  
    //                                                        
    // ----------------------------------------------------------------------
    _handleKeyDownCapture(tabContentElement, ev) {
        if (this._keyBindingManager === null || this._keyBindingManager.getKeyBindingContexts() === null) {
            return;
        }
        const bindings = this._keyBindingManager.getKeyBindingContexts().context(KEYBINDINGS_MAIN_UI);
        if (bindings === null) {
            return;
        }
        const command = bindings.mapEventToCommand(ev);
        if (this.executeCommand(command, { tabElement: tabContentElement })) {
            ev.stopPropagation();
            ev.preventDefault();
        }
    }
    //-----------------------------------------------------------------------
    //
    //    #####                                               ######                                          
    //   #     #  ####  #    # #    #   ##   #    # #####     #     #   ##   #      ###### ##### ##### ###### 
    //   #       #    # ##  ## ##  ##  #  #  ##   # #    #    #     #  #  #  #      #        #     #   #      
    //   #       #    # # ## # # ## # #    # # #  # #    #    ######  #    # #      #####    #     #   #####  
    //   #       #    # #    # #    # ###### #  # # #    #    #       ###### #      #        #     #   #      
    //   #     # #    # #    # #    # #    # #   ## #    #    #       #    # #      #        #     #   #      
    //    #####   ####  #    # #    # #    # #    # #####     #       #    # ###### ######   #     #   ###### 
    //
    //-----------------------------------------------------------------------
    getCommandPaletteEntries(commandableStack) {
        const thisIndex = commandableStack.indexOf(this);
        const tabContentElement = commandableStack[thisIndex - 1];
        if (tabContentElement instanceof Element) {
            return this._commandPaletteEntriesWithTarget(tabContentElement, this._tabWidgetFromElement(tabContentElement));
        }
        else {
            this._log.severe("commandableStack[thisIndex-1] wasn't an Element");
            return [];
        }
    }
    _commandPaletteEntriesWithTarget(tabContentElement, tabWidget) {
        const commandExecutor = this;
        const commandArguments = { tabElement: tabContentElement };
        const commandList = [
            { id: COMMAND_NEW_TERMINAL, group: PALETTE_GROUP, iconRight: "plus", label: "New Terminal", commandExecutor, commandArguments },
            { id: COMMAND_CLOSE_TAB, group: PALETTE_GROUP, iconRight: "times", label: "Close Tab", commandExecutor, commandArguments },
            { id: COMMAND_SELECT_TAB_LEFT, group: PALETTE_GROUP, label: "Select Previous Tab", commandExecutor, commandArguments },
            { id: COMMAND_SELECT_TAB_RIGHT, group: PALETTE_GROUP, label: "Select Next Tab", commandExecutor, commandArguments },
            { id: COMMAND_SELECT_PANE_LEFT, group: PALETTE_GROUP, label: " Select pane left", commandExecutor, commandArguments },
            { id: COMMAND_SELECT_PANE_RIGHT, group: PALETTE_GROUP, label: " Select pane right", commandExecutor, commandArguments },
            { id: COMMAND_SELECT_PANE_ABOVE, group: PALETTE_GROUP, label: " Select pane above", commandExecutor, commandArguments },
            { id: COMMAND_SELECT_PANE_BELOW, group: PALETTE_GROUP, label: " Select pane below", commandExecutor, commandArguments },
            { id: COMMAND_HORIZONTAL_SPLIT, group: PALETTE_GROUP, iconRight: "extraicon-#xea08", label: "Horizontal Split", commandExecutor, commandArguments },
            { id: COMMAND_VERTICAL_SPLIT, group: PALETTE_GROUP, iconRight: "columns", label: "Vertical Split", commandExecutor, commandArguments },
            { id: COMMAND_MOVE_TAB_LEFT, group: PALETTE_GROUP, label: "Move Tab Left", commandExecutor, commandArguments },
            { id: COMMAND_MOVE_TAB_RIGHT, group: PALETTE_GROUP, label: "Move Tab Right", commandExecutor, commandArguments },
            { id: COMMAND_MOVE_TAB_UP, group: PALETTE_GROUP, label: "Move Tab Up", commandExecutor, commandArguments },
            { id: COMMAND_MOVE_TAB_DOWN, group: PALETTE_GROUP, label: "Move Tab Down", commandExecutor, commandArguments },
        ];
        // FIXME
        if (tabWidget != null && tabWidget.parentElement instanceof Splitter_1.Splitter ||
            tabContentElement instanceof EmptyPaneMenu_1.EmptyPaneMenu) {
            commandList.push({ id: COMMAND_CLOSE_PANE, group: PALETTE_GROUP, label: "Close Pane", commandExecutor, commandArguments });
        }
        this._insertCommandKeyBindings(commandList);
        return commandList;
    }
    _insertCommandKeyBindings(commandList) {
        const keyBindings = this._keyBindingManager.getKeyBindingContexts().context(KEYBINDINGS_MAIN_UI);
        if (keyBindings !== null) {
            commandList.forEach((commandEntry) => {
                const shortcut = keyBindings.mapCommandToKeyBinding(commandEntry.id);
                commandEntry.shortcut = shortcut === null ? "" : shortcut;
            });
        }
    }
    executeCommand(commandId, options) {
        if (options == null) {
            return false;
        }
        const tabElement = options["tabElement"];
        if (tabElement == null) {
            return false;
        }
        switch (commandId) {
            case COMMAND_SELECT_TAB_LEFT:
                this._shiftTab(this._tabWidgetFromElement(tabElement), -1);
                break;
            case COMMAND_SELECT_TAB_RIGHT:
                this._shiftTab(this._tabWidgetFromElement(tabElement), 1);
                break;
            case COMMAND_SELECT_PANE_LEFT:
                this._selectPaneLeft(tabElement);
                break;
            case COMMAND_SELECT_PANE_RIGHT:
                this._selectPaneRight(tabElement);
                break;
            case COMMAND_SELECT_PANE_ABOVE:
                this._selectPaneAbove(tabElement);
                break;
            case COMMAND_SELECT_PANE_BELOW:
                this._selectPaneBelow(tabElement);
                break;
            case COMMAND_NEW_TERMINAL:
                this._switchToTab(this.newTerminalTab(this._tabWidgetFromElement(tabElement)));
                break;
            case COMMAND_CLOSE_TAB:
                this.closeTab(tabElement);
                break;
            case COMMAND_HORIZONTAL_SPLIT:
                this._horizontalSplit(tabElement);
                break;
            case COMMAND_VERTICAL_SPLIT:
                this._verticalSplit(tabElement);
                break;
            case COMMAND_CLOSE_PANE:
                this._closeSplit(tabElement);
                break;
            case COMMAND_MOVE_TAB_LEFT:
                this._moveTabElementLeft(tabElement);
                break;
            case COMMAND_MOVE_TAB_RIGHT:
                this._moveTabElementRight(tabElement);
                break;
            case COMMAND_MOVE_TAB_UP:
                this._moveTabElementUp(tabElement);
                break;
            case COMMAND_MOVE_TAB_DOWN:
                this._moveTabElementDown(tabElement);
                break;
            default:
                return false;
        }
        return true;
    }
    //-----------------------------------------------------------------------
    //
    //   ######  ####### #     # 
    //   #     #    #     #   #  
    //   #     #    #      # #   
    //   ######     #       #    
    //   #          #       #    
    //   #          #       #    
    //   #          #       #    
    //
    //-----------------------------------------------------------------------
    _setupIpc() {
        WebIpc.registerDefaultHandler(11 /* PTY_OUTPUT */, this._handlePtyOutput.bind(this));
        WebIpc.registerDefaultHandler(13 /* PTY_CLOSE */, this._handlePtyClose.bind(this));
    }
    _handlePtyOutput(msg) {
        const terminal = this._ptyIdTerminalMap.get(msg.id);
        if (terminal == null) {
            this._log.warn(`Unable to find a Terminal object to match pty ID ${msg.id}`);
            return;
        }
        const status = terminal.write(msg.data);
        WebIpc.ptyOutputBufferSize(msg.id, status.bufferSize);
    }
    _handlePtyClose(msg) {
        const terminal = this._ptyIdTerminalMap.get(msg.id);
        if (terminal == null) {
            this._log.warn(`Unable to find a Terminal object to match pty ID ${msg.id}`);
            return;
        }
        this.closeTab(terminal);
    }
    //-----------------------------------------------------------------------
    _getById(id) {
        return DomUtils.getShadowRoot(this).querySelector('#' + id);
    }
}
MainWebUi.TAG_NAME = 'EXTRATERM-MAINWEBUI';
MainWebUi.EVENT_TAB_OPENED = 'mainwebui-tab-opened';
MainWebUi.EVENT_TAB_CLOSED = 'mainwebui-tab-closed';
MainWebUi.EVENT_TITLE = 'mainwebui-title';
MainWebUi.EVENT_MINIMIZE_WINDOW_REQUEST = "mainwebui-minimize-window-request";
MainWebUi.EVENT_MAXIMIZE_WINDOW_REQUEST = "mainwebui-maximize-window-request";
MainWebUi.EVENT_CLOSE_WINDOW_REQUEST = "mainwebui-close-window-request";
exports.MainWebUi = MainWebUi;
function elementSupportsFocus(content) {
    return content instanceof Terminal_1.EtTerminal || content instanceof EmptyPaneMenu_1.EmptyPaneMenu || content instanceof ViewerTab_1.EtViewerTab;
}
//# sourceMappingURL=MainWebUi.js.map