"use strict";
/*
 * Copyright 2014-2016 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const fs = require("fs");
const path = require("path");
const SourceDir = require("./SourceDir");
const PATCH_MODULE_VERSION = "53"; // This version number also appears in build_package.js
if (process.versions.modules === PATCH_MODULE_VERSION) {
    // Patch in our special node-sass binary for the V8 module version used by Electron.
    process.env.SASS_BINARY_PATH = path.join(SourceDir.path, `node-sass-binary/${process.platform}-${process.arch}-${PATCH_MODULE_VERSION}/binding.node`);
}
const NodeSass = require("node-sass");
const _ = require("lodash");
const Logger_1 = require("./Logger");
const ThemeTypes = require("./Theme");
const THEME_CONFIG = "theme.json";
const DEBUG_SASS = false;
const DEBUG_SASS_FINE = false;
const DEBUG_SCAN = false;
//-------------------------------------------------------------------------
//
//  #######                             #     #                                           ###                      
//     #    #    # ###### #    # ###### ##   ##   ##   #    #   ##    ####  ###### #####   #  #    # #####  #      
//     #    #    # #      ##  ## #      # # # #  #  #  ##   #  #  #  #    # #      #    #  #  ##  ## #    # #      
//     #    ###### #####  # ## # #####  #  #  # #    # # #  # #    # #      #####  #    #  #  # ## # #    # #      
//     #    #    # #      #    # #      #     # ###### #  # # ###### #  ### #      #####   #  #    # #####  #      
//     #    #    # #      #    # #      #     # #    # #   ## #    # #    # #      #   #   #  #    # #      #      
//     #    #    # ###### #    # ###### #     # #    # #    # #    #  ####  ###### #    # ### #    # #      ######
//
//-------------------------------------------------------------------------
class ThemeManagerImpl {
    constructor(directories) {
        this._log = null;
        this._directories = null;
        this._themes = null;
        this._log = Logger_1.getLogger("ThemeManagerImpl", this);
        this._directories = directories;
        const allThemes = new Map();
        this._directories.forEach((directory) => {
            const themes = this._scanThemeDirectory(directory);
            themes.forEach((value, key) => {
                allThemes.set(key, value);
            });
        });
        this._themes = allThemes;
    }
    getTheme(themeId) {
        return this._themes.get(themeId) || null;
    }
    getAllThemes() {
        const result = [];
        this._themes.forEach((value) => {
            result.push(value);
        });
        return result;
    }
    renderThemes(themeStack, cssFileList, globalVariables) {
        const themeInfoList = this._themeIdListToThemeInfoList(themeStack);
        globalVariables = globalVariables || new Map();
        return this._renderThemeStackContents(themeInfoList, cssFileList, globalVariables)
            .then((renderResults) => {
            return renderResults;
        });
    }
    _themeIdListToThemeInfoList(themeIdList) {
        // Convert the list of theme IDs to a list of ThemeInfos
        const themeInfoList = themeIdList.map((themeId) => {
            const themeInfo = this.getTheme(themeId);
            if (themeInfo === null) {
                this._log.warn("The requested theme name '" + themeId + "' is unknown.");
                return null;
            }
            else {
                return themeInfo;
            }
        }).filter((themeInfo) => themeInfo !== null);
        return themeInfoList;
    }
    /**
     * Scan for themes.
     *
     * @param themesdir The directory to scan for themes.
     * @returns Map of found theme config objects.
     */
    _scanThemeDirectory(themesDir) {
        let themeMap = new Map();
        if (fs.existsSync(themesDir)) {
            const contents = fs.readdirSync(themesDir);
            contents.forEach((item) => {
                const infoPath = path.join(themesDir, item, THEME_CONFIG);
                try {
                    const infoStr = fs.readFileSync(infoPath, { encoding: "utf8" });
                    const themeInfo = JSON.parse(infoStr);
                    themeInfo.path = path.join(themesDir, item);
                    themeInfo.id = item;
                    this._fillThemeInfoDefaults(themeInfo);
                    if (this._validateThemeInfo(themeInfo)) {
                        themeMap.set(item, themeInfo);
                    }
                    if (DEBUG_SCAN) {
                        this._log.debug(themeInfo);
                    }
                }
                catch (err) {
                    this._log.warn("Warning: Unable to read file ", infoPath, err);
                }
            });
        }
        return themeMap;
    }
    _fillThemeInfoDefaults(themeInfo) {
        themeInfo.comment = themeInfo.comment === undefined ? "" : themeInfo.comment;
        themeInfo.debug = themeInfo.debug === undefined ? false : themeInfo.debug;
    }
    /**
     *
     */
    _validateThemeInfo(themeinfo) {
        return _.isString(themeinfo.name) && themeinfo.name !== "";
    }
    _renderThemeStackContents(themeStack, cssFileList, globalVariables) {
        return this._recursiveRenderThemeStackContents(cssFileList.slice(0), themeStack, globalVariables);
    }
    _recursiveRenderThemeStackContents(todoCssFile, themeStack, globalVariables) {
        if (DEBUG_SASS) {
            this._log.debug("Compiling _recursiveRenderThemeStackContents:" + todoCssFile.length);
        }
        const cssFile = todoCssFile[0];
        const dirStack = _.uniq(themeStack.map((themeInfo) => themeInfo.path));
        const sassFileName = ThemeTypes.cssFileNameBase(cssFile) + '.scss';
        // Create SASS variables for the location of each theme directory.
        const variables = new Map(globalVariables);
        themeStack.forEach((themeInfo) => {
            variables.set('--source-dir-' + themeInfo.id, themeInfo.path.replace(/\\/g, "/"));
        });
        if (DEBUG_SASS) {
            this._log.debug("Compiling " + sassFileName);
            this._log.startTime("Compiling " + sassFileName);
        }
        const loadSassPromise = this._loadSassFile(dirStack, sassFileName, variables);
        return loadSassPromise.then((cssText) => {
            if (DEBUG_SASS) {
                this._log.endTime("Compiling " + sassFileName);
                this._log.debug("Done " + sassFileName);
            }
            if (todoCssFile.length <= 1) {
                // Done. Send back some results.
                const themeContents = {
                    cssFiles: {}
                };
                themeContents.cssFiles[ThemeTypes.cssFileNameBase(cssFile)] = cssText;
                const renderResult = { success: true, themeContents: themeContents, errorMessage: null };
                return renderResult;
            }
            else {
                return this._recursiveRenderThemeStackContents(todoCssFile.slice(1), themeStack, globalVariables)
                    .then((renderResult) => {
                    renderResult.themeContents.cssFiles[ThemeTypes.cssFileNameBase(cssFile)] = cssText;
                    return renderResult;
                });
            }
        });
    }
    _loadSassFile(dirStack, sassFileName, variables) {
        let formattedVariables = "";
        if (variables !== undefined) {
            variables.forEach((value, key) => {
                const formattedValue = typeof (value) === "string" ? `"${value}"` : "" + value;
                formattedVariables += `$${key}: ${formattedValue};
`;
            });
        }
        if (DEBUG_SASS) {
            this._log.debug("Formatted SASS variables: ", formattedVariables);
        }
        return new Promise((resolve, cancel) => {
            if (DEBUG_SASS) {
                this._log.debug("Processing " + sassFileName);
            }
            try {
                const importer = (url, prev, done) => {
                    const basePath = url;
                    const contextBaseDir = path.dirname(prev);
                    const dirName = path.join(contextBaseDir, path.dirname(basePath));
                    const baseName = path.basename(basePath);
                    if (DEBUG_SASS_FINE) {
                        this._log.debug(`Import request: URL: ${url} prev: ${prev}`);
                        this._log.debug("dirName:", dirName);
                        this._log.debug("baseName:", baseName);
                        this._log.debug("contextBaseDir:", contextBaseDir);
                    }
                    const candidates = [basePath,
                        path.join(dirName, '_' + baseName + '.scss'),
                        path.join(dirName, '_' + baseName + '.sass'),
                        path.join(dirName, '_' + baseName + '.css'),
                        basePath + '.scss',
                        basePath + '.sass',
                        basePath + '.css'
                    ];
                    for (let candidate of candidates) {
                        const candidateFileName = this._findFile(dirStack, candidate);
                        if (candidateFileName) {
                            if (DEBUG_SASS_FINE) {
                                this._log.debug("Trying " + candidateFileName);
                            }
                            if (DEBUG_SASS_FINE) {
                                this._log.debug("Importer returning SASS file: " + candidateFileName);
                            }
                            try {
                                const content = fs.readFileSync(candidateFileName, { encoding: 'utf-8' });
                                done({ contents: content });
                            }
                            catch (err) {
                                done(err);
                            }
                            return;
                        }
                    }
                    done(new Error("Unable to find " + basePath));
                };
                // Start the compile using a small virtual 'boot' file.
                const scssText = formattedVariables + '\n@import "' + sassFileName + '";\n';
                if (DEBUG_SASS_FINE) {
                    this._log.debug("Root sass text: ", scssText);
                }
                NodeSass.render({ data: scssText, precision: 8, importer: importer }, (err, result) => {
                    if (err === null) {
                        if (DEBUG_SASS) {
                            this._log.debug("Succeeded done processing " + sassFileName);
                        }
                        resolve(result.css.toString('utf8'));
                    }
                    else {
                        this._log.warn("An SASS error occurred while processing " + sassFileName, err, err.message);
                        if (DEBUG_SASS) {
                            this._log.debug("Failed processing " + sassFileName);
                        }
                        cancel(new Error("An SASS error occurred while processing " + sassFileName + "\n" + err.message));
                    }
                });
            }
            catch (err) {
                this._log.warn("An error occurred while processing " + sassFileName, err);
                cancel(err);
                return;
            }
        });
    }
    _findFile(dirStack, fileName) {
        if (DEBUG_SASS_FINE) {
            this._log.debug(`findFile(): Looking for ${fileName} in dirs ${dirStack}`);
        }
        for (const dir of dirStack) {
            const candidateFileName = path.join(dir, fileName);
            if (DEBUG_SASS_FINE) {
                this._log.debug(`findFile(): Checking ${candidateFileName}`);
            }
            if (fs.existsSync(candidateFileName) && fs.statSync(candidateFileName).isFile()) {
                return candidateFileName;
            }
        }
        return null;
    }
}
/**
 * Maps a stack of theme names to a file name
 *
 * @param  themeStack the stack of theme names to encode into a file name
 * @return the file name which encodes the list of theme names
 */
function themeStackName(themeStack) {
    // We use the period char to separate parts of the theme names.
    return themeStack.map(escapeThemeFileName).join(".");
}
function escapeThemeFileName(name) {
    return name.replace(/_/g, "__").replace(/\./g, "_.");
}
function unescapeThemeFileName(name) {
    return name.replace(/_\./g, ".").replace(/__/g, "_");
}
/**
 * Decomposes a theme cache file name into a stack theme names.
 *
 * @param  fileName the base file name without extension
 * @return the stack of theme names
 */
function decomposeThemeStackName(fileName) {
    const parts = fileName.split(/\./g).reduce((accu, part) => {
        if (accu.length === 0) {
            return [part];
        }
        else {
            const last = accu[accu.length - 1];
            if (last.endsWith("_")) {
                accu[accu.length - 1] = last + "." + part;
                return accu;
            }
            else {
                return [...accu, part];
            }
        }
    }, []);
    return parts.map(unescapeThemeFileName);
}
/**
 * Finds the path of a path in a list of directories.
 *
 * @param  themeDirectories the list of directories to search
 * @param  themeName        the theme to search for
 * @return                  the full path to the theme directory or null if it could not be found
 */
function findThemePath(themeDirectories, themeName) {
    for (let themeDir of themeDirectories) {
        const fullThemePath = path.join(themeDir, themeName);
        if (fs.existsSync(fullThemePath)) {
            return fullThemePath;
        }
    }
    return null;
}
function makeThemeManager(directories) {
    return new ThemeManagerImpl(directories);
}
exports.makeThemeManager = makeThemeManager;
//# sourceMappingURL=ThemeManager.js.map