"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/**
 * term.js - an xterm emulator
 * Copyright (c) 2012-2013, Christopher Jeffrey (MIT License)
 * Copyright (c) 2014-2017, Simon Edwards <simon@simonzone.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 * Originally forked from (with the author's permission):
 *   Fabrice Bellard's javascript vt100 for jslinux:
 *   http://bellard.org/jslinux/
 *   Copyright (c) 2011 Fabrice Bellard
 *   The original design remains. The terminal itself
 *   has been extended to include xterm CSI codes, among
 *   other features.
 *
 * Forked again from Christopher Jeffrey's work by Simon Edwards in 2014 and
 * converted over to TypeScript.
 */
const TermApi_1 = require("./TermApi");
const DEBUG_RESIZE = false;
const REFRESH_START_NULL = 100000000;
const REFRESH_END_NULL = -100000000;
const MAX_BATCH_TIME = 16; // 16 ms = 60Hz
const REFRESH_DELAY = 100; // ms. How long to wait before doing a screen refresh during busy times.
/**
 * Terminal Emulation References:
 *   http://vt100.net/
 *   http://invisible-island.net/xterm/ctlseqs/ctlseqs.txt
 *   http://invisible-island.net/xterm/ctlseqs/ctlseqs.html
 *   http://invisible-island.net/vttest/
 *   http://www.inwap.com/pdp10/ansicode.txt
 *   http://linux.die.net/man/4/console_codes
 *   http://linux.die.net/man/7/urxvt
 */
/**
 * States
 */
const STATE_NORMAL = 0;
const STATE_ESCAPE = 1;
const STATE_CSI = 2;
const STATE_OSC = 3;
const STATE_CHARSET = 4;
const STATE_DCS = 5;
const STATE_IGNORE = 6;
const STATE_APPLICATION_START = 7;
const STATE_APPLICATION_END = 8;
const STATE_DEC_HASH = 9;
const MAX_PROCESS_WRITE_SIZE = 4096;
;
const RENDER_EVENT = "RENDER_EVENT";
const BELL_EVENT = "BELL_EVENT";
const DATA_EVENT = "DATA_EVENT";
const TITLE_EVENT = "TITLE_EVENT";
const APPLICATIONMODESTART_EVENT = "APPLICATIONMODESTART_EVENT";
const APPLICATIONMODEDATA_EVENT = "APPLICATIONMODEDATA_EVENT";
const APPLICATIONMODEEND_EVENT = "APPLICATIONMODEEND_EVENT";
const WRITEBUFFERSIZE_EVENT = "WRITEBUFFERSIZE_EVENT";
const MAX_WRITE_BUFFER_SIZE = 1024 * 100; // 100 KB
function packCharAttr(flags, fg, bg) {
    return (flags << 18) | (fg << 9) | bg;
}
function isCharAttrCursor(attr) {
    return attr === -1;
}
class Emulator {
    constructor(options) {
        this.cols = 80;
        this.rows = 24;
        this.lastReportedPhysicalHeight = 0;
        this.state = 0; // Escape code parsing state.
        this.mouseEvents = false;
        this.x = 0; // Cursor x position
        this.y = 0; // Cursor y position
        this.savedX = 0;
        this.savedY = 0;
        this.oldy = 0;
        this.cursorState = false; // Cursor blink state.
        this.cursorHidden = false;
        this._hasFocus = false;
        this.queue = '';
        this.scrollTop = 0;
        this.scrollBottom = 23;
        // modes
        this.applicationKeypad = false;
        this.applicationCursor = false;
        this.originMode = false;
        this.insertMode = false;
        this.wraparoundMode = false;
        this.normal = null;
        // charset
        this.charset = null;
        this.savedCharset = null;
        this.gcharset = null;
        this.glevel = 0;
        this.charsets = [null];
        this.curAttr = 0; // Current character style.
        this.savedCurAttr = packCharAttr(0, 257, 256); // Default character style;
        this.params = [];
        this.currentParam = 0;
        this.prefix = '';
        this.postfix = '';
        this.lines = [];
        this.isMac = false;
        this._writeBuffers = []; // Buffer for incoming data waiting to be processed.
        this._processWriteChunkTimer = null; // Timer ID for our write chunk timer.  
        this._refreshTimer = null; // Timer ID for triggering an on scren refresh.
        this._performanceNow = null;
        this._scrollbackLineQueue = []; // Queue of scrollback lines which need to sent via an event.
        this._refreshStart = REFRESH_START_NULL;
        this._refreshEnd = REFRESH_END_NULL;
        this.sendFocus = false;
        this._mouseButtonDown = false;
        this.utfMouse = false;
        this.decLocator = false;
        this.urxvtMouse = false;
        this.sgrMouse = false;
        this.vt300Mouse = false;
        this.vt200Mouse = false;
        this.normalMouse = false;
        this.x10Mouse = false;
        this._pressed = 32;
        this._lastMovePos = null;
        this._events = {};
        this._blinker = null;
        this.title = "";
        const defaults = {
            termName: 'xterm',
            geometry: [80, 24],
            debug: false,
            applicationModeCookie: null,
            performanceNow: () => window.performance.now()
        };
        this.termName = options.termName === undefined ? 'xterm' : options.termName;
        this.rows = options.rows === undefined ? 24 : options.rows;
        this.cols = options.columns === undefined ? 80 : options.columns;
        this.debug = options.debug === undefined ? false : options.debug;
        this.applicationModeCookie = options.applicationModeCookie === undefined ? null : options.applicationModeCookie;
        if (options.performanceNowFunc == null) {
            this._performanceNow = window.performance.now.bind(window.performance);
        }
        else {
            this._performanceNow = options.performanceNowFunc;
        }
        if (options.userAgent !== undefined) {
            this.isMac = options.userAgent.indexOf('Mac') !== -1;
        }
        this.state = 0; // Escape code parsing state.
        this._resetVariables();
        this._hasFocus = false;
        this._writeBuffers = []; // Buffer for incoming data waiting to be processed.
        this._processWriteChunkTimer = null; // Timer ID for our write chunk timer.
        this._refreshTimer = null; // Timer ID for triggering an on scren refresh.
    }
    destroy() {
        if (this._processWriteChunkTimer !== null) {
            clearTimeout(this._processWriteChunkTimer);
            this._processWriteChunkTimer = null;
        }
        if (this._refreshTimer !== null) {
            clearTimeout(this._refreshTimer);
        }
        this._events = {};
        this.handler = function () { };
        this.write = () => ({ bufferSize: 0 });
    }
    _resetVariables() {
        this.x = 0;
        this._setCursorY(0);
        this.oldy = 0;
        this.cursorState = true; // Cursor blink state.
        this.cursorHidden = false;
        this._hasFocus = false;
        //  this.convertEol;
        this.queue = '';
        this.scrollTop = 0;
        this.scrollBottom = this.rows - 1;
        // modes
        this.applicationKeypad = false;
        this.applicationCursor = false;
        this.originMode = false;
        this.insertMode = false;
        this.wraparoundMode = false;
        this.normal = null;
        // charset
        this.charset = null;
        this.gcharset = null;
        this.glevel = 0;
        this.charsets = [null];
        // mouse properties
        //  this.decLocator;
        //  this.x10Mouse;
        //  this.vt200Mouse;
        //  this.vt300Mouse;
        //  this.normalMouse;
        //  this.mouseEvents;
        //  this.sendFocus;
        //  this.utfMouse;
        //  this.sgrMouse;
        //  this.urxvtMouse;
        // misc
        //  this.element;
        //  this.children;
        //  this.savedX;
        //  this.savedY;
        //  this.savedCols;
        this.curAttr = Emulator.defAttr; // Current character style.
        this.params = [];
        this.currentParam = 0;
        this.prefix = '';
        this.postfix = '';
        this.lines = [];
        //  this.tabs;
        this.setupStops();
    }
    // back_color_erase feature for xterm.
    eraseAttr() {
        // if (this.is('screen')) return this.defAttr;
        return (Emulator.defAttr & ~0x1ff) | (this.curAttr & 0x1ff);
    }
    focus() {
        if (this.sendFocus) {
            this.send('\x1b[I');
        }
        this._hasFocus = true;
        this._dispatchEvents();
    }
    /**
     * Returns true if this terminal has the input focus.
     *
     * @return true if the terminal has the focus.
     */
    hasFocus() {
        return this._hasFocus;
    }
    blur() {
        if (!this._hasFocus) {
            return;
        }
        this.markRowRangeForRefresh(this.y, this.y);
        if (this.sendFocus) {
            this.send('\x1b[O');
        }
        this._hasFocus = false;
        this._dispatchEvents();
    }
    // XTerm mouse events
    // http://invisible-island.net/xterm/ctlseqs/ctlseqs.html#Mouse%20Tracking
    // To better understand these
    // the xterm code is very helpful:
    // Relevant files:
    //   button.c, charproc.c, misc.c
    // Relevant functions in xterm/button.c:
    //   BtnCode, EmitButtonCode, EditorButton, SendMousePosition
    mouseDown(ev) {
        if (!this.mouseEvents) {
            return false;
        }
        let button = this.mouseEventOptionsToButtons(ev);
        // no mods
        if (this.vt200Mouse) {
            button = button & ~0xc; // ctrl only
        }
        else if (!this.normalMouse) {
            button = button & ~0x1c; // no mods
        }
        this.sendMouseButtonSequence({ x: ev.column, y: ev.row }, button);
        if (this.vt200Mouse) {
            this.sendMouseButtonSequence({ x: ev.column, y: ev.row }, 3); // release button
            return true;
        }
        // bind events
        if (this.normalMouse) {
            this._lastMovePos = null;
            this._mouseButtonDown = true;
        }
        return true;
    }
    mouseMove(ev) {
        if (!this.mouseEvents) {
            return false;
        }
        if (!this._mouseButtonDown) {
            return false;
        }
        if (this._lastMovePos !== null && this._lastMovePos.x === ev.column && this._lastMovePos.y === ev.row) {
            return true;
        }
        const pos = { x: ev.column, y: ev.row };
        let button = this.mouseEventOptionsToButtons(ev);
        // no mods
        if (this.vt200Mouse) {
            button = button & ~0xc; // ctrl only
        }
        else if (!this.normalMouse) {
            button = button & ~0x1c; // no mods
        }
        // buttons marked as motions
        // are incremented by 32
        button |= 32;
        this.sendMouseSequence(button, pos);
        this._lastMovePos = pos;
        return true;
    }
    mouseUp(ev) {
        if (!this.mouseEvents) {
            return false;
        }
        if (!this._mouseButtonDown) {
            return false;
        }
        if (this.x10Mouse) {
            this._mouseButtonDown = false;
            return false; // No mouse ups for x10.
        }
        if (ev === null) {
            this._mouseButtonDown = false;
            return true;
        }
        let button = this.mouseEventOptionsToButtons(ev, true);
        // no mods
        if (this.vt200Mouse) {
            button = button & ~0xc; // ctrl only
        }
        else if (!this.normalMouse) {
            button = button & ~0x1c; // no mods
        }
        this.sendMouseButtonSequence({ x: ev.column, y: ev.row }, button);
        this._mouseButtonDown = false;
        return true;
    }
    mouseEventOptionsToButtons(ev, release = false) {
        // two low bits:
        // 0 = left
        // 1 = middle
        // 2 = right
        // 3 = release
        // wheel up/down:
        // 1, and 2 - with 64 added
        let button = 0;
        if (release) {
            button = 3;
        }
        else if (ev.leftButton) {
            button = 0;
        }
        else if (ev.middleButton) {
            button = 1;
        }
        else if (ev.rightButton) {
            button = 2;
        }
        const shift = ev.shiftKey ? 4 : 0;
        const meta = ev.metaKey ? 8 : 0;
        const ctrl = ev.ctrlKey ? 16 : 0;
        const mod = shift | meta | ctrl;
        return (mod << 2) | button;
    }
    refreshScreen() {
        this.markRowRangeForRefresh(0, this.lines.length);
        this._dispatchEvents();
    }
    /**
     * Moves all of the rows above the cursor into the physical scrollback area.
     */
    moveRowsAboveCursorToScrollback() {
        const lines = this.lines.slice(0, this.y);
        const newLines = this.lines.slice(this.y);
        lines.forEach(line => this._scrollbackLineQueue.push(line));
        this.lines = newLines;
        this.markAllRowsForRefresh();
        this._setCursorY(0);
        this.oldy = 0;
        this._scheduleRefresh(true);
    }
    flushRenderQueue() {
        this._dispatchEvents();
    }
    _getRow(row) {
        while (row >= this.lines.length) {
            this.lines.push(this.blankLine());
        }
        return this.lines[row];
    }
    // Fetch the LineCell at row 'row' if it exists, else return null.
    _tryGetRow(row) {
        return row >= this.lines.length ? null : this.lines[row];
    }
    getDimensions() {
        return {
            rows: this.rows,
            cols: this.cols,
            materializedRows: this.lines.length,
            cursorX: this.x,
            cursorY: this.y
        };
    }
    getLineText(y) {
        if (y < 0 || y >= this.lines.length) {
            return null;
        }
        const row = this.lines[y];
        return String.fromCodePoint(...row.chars);
    }
    // encode button and
    // position to characters
    encodeMouseData(buffer, ch) {
        if (!this.utfMouse) {
            if (ch === 255) {
                buffer.push(0);
                return;
            }
            if (ch > 127) {
                ch = 127;
            }
            buffer.push(ch);
        }
        else {
            if (ch === 2047) {
                buffer.push(0);
            }
            if (ch < 127) {
                buffer.push(ch);
            }
            else {
                if (ch > 2047) {
                    ch = 2047;
                }
                buffer.push(0xC0 | (ch >> 6));
                buffer.push(0x80 | (ch & 0x3F));
            }
        }
    }
    // send a mouse event:
    // regular/utf8: ^[[M Cb Cx Cy
    // urxvt: ^[[ Cb ; Cx ; Cy M
    // sgr: ^[[ Cb ; Cx ; Cy M/m
    // vt300: ^[[ 24(1/3/5)~ [ Cx , Cy ] \r
    // locator: CSI P e ; P b ; P r ; P c ; P p & w
    sendMouseSequence(button, pos0based) {
        let data;
        const pos = { x: pos0based.x + 1, y: pos0based.y + 1 };
        if (this.vt300Mouse) {
            this.log("sendEvent(): vt300Mouse");
            // NOTE: Unstable.
            // http://www.vt100.net/docs/vt3xx-gp/chapter15.html
            button &= 3;
            const x = pos.x;
            const y = pos.y;
            let data = '\x1b[24';
            if (button === 0) {
                data += '1';
            }
            else if (button === 1) {
                data += '3';
            }
            else if (button === 2) {
                data += '5';
            }
            else if (button === 3) {
                return;
            }
            else {
                data += '0';
            }
            data += '~[' + x + ',' + y + ']\r';
            this.send(data);
            return;
        }
        if (this.decLocator) {
            // NOTE: Unstable.
            this.log("sendEvent with decLocator is not implemented!");
            // const x = pos.x;
            // const y = pos.y;
            // const translatedButton = {0:2, 1:4, 2:6, 3:3}[button & 3];
            // self.send('\x1b[' + translatedButton + ';' + (translatedButton === 3 ? 4 : 0) + ';' + y + ';' + x + ';' +
            //   (pos.page || 0) + '&w');
            return;
        }
        if (this.urxvtMouse) {
            this.log("sendEvent(): urxvtMouse");
            const x = pos.x + 1;
            const y = pos.y + 1;
            this.send('\x1b[' + (button + 32) + ';' + x + ';' + y + 'M');
            return;
        }
        if (this.sgrMouse) {
            this.log("sendEvent(): sgrMouse");
            const x = pos.x;
            const y = pos.y;
            this.send('\x1b[<' + ((button & 3) === 3 ? button & ~3 : button) + ';' + x +
                ';' + y + ((button & 3) === 3 ? 'm' : 'M'));
            return;
        }
        this.log("sendEvent(): default");
        const encodedData = [];
        this.encodeMouseData(encodedData, button + 32);
        // xterm sends raw bytes and
        // starts at 32 (SP) for each.    
        this.encodeMouseData(encodedData, pos.x + 32);
        this.encodeMouseData(encodedData, pos.y + 32);
        this.send('\x1b[M' + String.fromCharCode.apply(String, encodedData));
    }
    // mouseup, mousedown, mousewheel
    // left click: ^[[M 3<^[[M#3<
    // mousewheel up: ^[[M`3>
    sendMouseButtonSequence(pos, button) {
        this.sendMouseSequence(button, pos);
        this._pressed = (button === 3) ? 32 : button;
    }
    /**
     * Rendering Engine
     */
    /**
     * Schedule a screen refresh and update.
     *
     * @param {boolean} immediate True if the refresh should occur as soon as possible. False if a slight delay is permitted.
     */
    _scheduleRefresh(immediate) {
        if (this._refreshTimer === null) {
            this._refreshTimer = setTimeout(() => {
                this._refreshTimer = null;
                this._refreshFrame();
            }, immediate ? 0 : REFRESH_DELAY);
        }
    }
    /**
     * Refresh and update the screen.
     *
     * Usually call via a timer.
     */
    _refreshFrame() {
        this._dispatchEvents();
        this._refreshStart = REFRESH_START_NULL;
        this._refreshEnd = REFRESH_END_NULL;
    }
    // In the screen buffer, each character
    // is stored as a an array with a character
    // and a 32-bit integer.
    // First value: a utf-16 character.
    // Second value:
    // Next 9 bits: background color (0-511).
    // Next 9 bits: foreground color (0-511).
    // Next 14 bits: a mask for misc. flags:
    //   1=bold, 2=underline, 4=blink, 8=inverse, 16=invisible
    /**
     * Marks a range of rows to be refreshed on the screen.
     *
     * @param {number} start start row to refresh
     * @param {number} end   end row (INCLUSIVE!) to refresh
     */
    markRowRangeForRefresh(start, end) {
        this._refreshStart = Math.min(start, this._refreshStart);
        this._refreshEnd = Math.max(end + 1, this._refreshEnd);
    }
    lineAtRow(row) {
        if (row < 0 || row >= this.rows) {
            return null;
        }
        let line = this._getRow(row);
        // Place the cursor in the row.
        if (row === this.y &&
            this.cursorState &&
            !this.cursorHidden &&
            this.x < this.cols) {
            const x = this.x;
            const newLine = { chars: new Uint32Array(line.chars), attrs: new Uint32Array(line.attrs) };
            newLine.attrs[x] = 0xffffffff;
            return newLine;
        }
        return line;
    }
    scroll() {
        // Drop the oldest line into the scrollback buffer.
        if (this.scrollTop === 0) {
            this._scrollbackLineQueue.push(this.lines[0]);
        }
        // last line
        const lastline = this.rows - 1;
        // subtract the bottom scroll region
        const insertRow = lastline - this.rows + 1 + this.scrollBottom;
        this.lines.splice(this.scrollTop, 1);
        // add our new line
        this.lines.splice(insertRow, 0, this.blankLine());
        this.markRowForRefresh(this.scrollTop);
        this.markRowForRefresh(this.scrollBottom);
    }
    // private scrollDisp(disp) {
    //   this.ydisp += disp;
    //   if (this.ydisp > this.ybase) {
    //     this.ydisp = this.ybase;
    //   } else if (this.ydisp < 0) {
    //     this.ydisp = 0;
    //   }
    //   this.markRowRangeForRefresh(0, this.rows - 1);
    // }
    write(data) {
        this._writeBuffers.push(data);
        this._scheduleProcessWriteChunk();
        return this._writeBufferStatus();
    }
    _writeBufferStatus() {
        const size = this._writeBuffers.map((buf) => buf.length).reduce((accu, x) => accu + x, 0);
        return { bufferSize: MAX_WRITE_BUFFER_SIZE - size };
    }
    _emitWriteBufferSizeEvent() {
        this._emit(WRITEBUFFERSIZE_EVENT, this, this._writeBufferStatus());
    }
    /**
     * Schedule the write chunk process to run the next time the event loop is entered.
     */
    _scheduleProcessWriteChunk() {
        if (this._processWriteChunkTimer === null) {
            this._processWriteChunkTimer = setTimeout(() => {
                this._processWriteChunkTimer = null;
                this._processWriteChunkRealTime();
            }, 0);
        }
    }
    /**
     * Process the next chunk of data to written into a the line array.
     */
    _processWriteChunkRealTime() {
        const starttime = this._performanceNow();
        //console.log("++++++++ _processWriteChunk() start time: " + starttime);
        // Schedule a call back just in case. setTimeout(.., 0) still carries a ~4ms delay. 
        this._scheduleProcessWriteChunk();
        while (true) {
            if (this._processOneWriteChunk() === false) {
                clearTimeout(this._processWriteChunkTimer);
                this._processWriteChunkTimer = null;
                this._scheduleRefresh(true);
                this._emitWriteBufferSizeEvent();
                break;
            }
            const nowtime = this._performanceNow();
            if ((nowtime - starttime) > MAX_BATCH_TIME) {
                this._scheduleRefresh(false);
                this._emitWriteBufferSizeEvent();
                break;
            }
        }
        //  console.log("---------- _processWriteChunk() end time: " + this._performanceNow());
    }
    /**
     * Process one chunk of written data.
     *
     * @returns {boolean} True if there are extra chunks available which need processing.
     */
    _processOneWriteChunk() {
        if (this._writeBuffers.length === 0) {
            return false; // Nothing to do.
        }
        let chunk = this._writeBuffers[0];
        if (chunk.length <= MAX_PROCESS_WRITE_SIZE) {
            this._writeBuffers.splice(0, 1);
        }
        else {
            this._writeBuffers[0] = chunk.slice(MAX_PROCESS_WRITE_SIZE);
            chunk = chunk.slice(0, MAX_PROCESS_WRITE_SIZE);
        }
        this._processWriteData(chunk);
        return this._writeBuffers.length !== 0;
    }
    _flushWriteBuffer() {
        while (this._processOneWriteChunk()) {
            // Keep on going until it is all done.
        }
        this._emitWriteBufferSizeEvent();
    }
    /**
     * Process a block of characters and control sequences and render them to the screen.
     *
     * @param data the string of characters and control sequences to process.
     */
    _processWriteData(data) {
        //console.log("write() data.length: " + data.length);
        //var starttime = window.performance.now();
        //var endtime;
        //console.log("write() start time: " + starttime);
        this.oldy = this.y;
        const len = data.length;
        for (let i = 0; i < len; i++) {
            let ch = data[i];
            let codePoint = 0;
            // Unicode UTF-16 surrogate handling.
            if ((ch.charCodeAt(0) & 0xFC00) === 0xD800) {
                const highSurrogate = ch.charCodeAt(0);
                i++;
                if (i < len) {
                    const lowSurrogate = data[i].charCodeAt(0);
                    codePoint = ((highSurrogate & 0x03FF) << 10) | (lowSurrogate & 0x03FF) + 0x10000;
                }
                ch = ' '; // fake it just enough to hit the right code below.
            }
            else {
                codePoint = ch.codePointAt(0);
            }
            switch (this.state) {
                case STATE_NORMAL:
                    switch (ch) {
                        // '\0'
                        // case '\0':
                        // case '\200':l
                        //   break;
                        // '\a'
                        case '\x07':
                            this.bell();
                            break;
                        // '\n', '\v', '\f'
                        case '\n':
                        case '\x0b':
                        case '\x0c':
                            this.newLine();
                            break;
                        // '\r'
                        case '\r':
                            this.carriageReturn();
                            break;
                        // '\b'
                        case '\x08':
                            if (this.x > 0) {
                                this.x--;
                            }
                            break;
                        // '\t'
                        case '\t':
                            this.x = this.nextStop();
                            break;
                        // shift out
                        case '\x0e':
                            this.setgLevel(1);
                            break;
                        // shift in
                        case '\x0f':
                            this.setgLevel(0);
                            break;
                        // '\e'
                        case '\x1b':
                            this.state = STATE_ESCAPE;
                            break;
                        default:
                            // ' '
                            if (ch >= ' ') {
                                if (this.charset && this.charset[ch]) {
                                    ch = this.charset[ch];
                                    codePoint = ch.codePointAt(0);
                                }
                                if (this.x >= this.cols) {
                                    this.x = 0;
                                    this.markRowForRefresh(this.y);
                                    if (this.y + 1 > this.scrollBottom) {
                                        this.scroll();
                                    }
                                    else {
                                        this._setCursorY(this.y + 1);
                                    }
                                }
                                const { chars, attrs } = this._getRow(this.y);
                                if (this.insertMode) {
                                    // Push the characters out of the way to make space.
                                    chars.copyWithin(this.x + 1, this.x);
                                    chars[this.x] = ' '.codePointAt(0);
                                    attrs.copyWithin(this.x + 1, this.x);
                                    attrs[this.x] = this.curAttr;
                                    if (isWide(ch)) {
                                        chars.copyWithin(this.x + 1, this.x);
                                        chars[this.x] = ' '.codePointAt(0);
                                        attrs.copyWithin(this.x + 1, this.x);
                                        attrs[this.x] = this.curAttr;
                                    }
                                }
                                chars[this.x] = codePoint;
                                attrs[this.x] = this.curAttr;
                                this.x++;
                                this.markRowForRefresh(this.y);
                                if (isWide(ch)) {
                                    const j = this.y;
                                    const line = this._getRow(j);
                                    if (this.cols < 2 || this.x >= this.cols) {
                                        line.chars[this.x - 1] = ' '.codePointAt(0);
                                        line.attrs[this.x - 1] = this.curAttr;
                                        break;
                                    }
                                    line.chars[this.x] = ' '.codePointAt(0);
                                    line.attrs[this.x] = this.curAttr;
                                    this.x++;
                                }
                            }
                            break;
                    }
                    break;
                case STATE_ESCAPE:
                    i = this._processDataEscape(ch, i);
                    break;
                case STATE_CHARSET:
                    i = this._processDataCharset(ch, i);
                    break;
                case STATE_OSC:
                    i = this._processDataOSC(ch, i);
                    break;
                case STATE_CSI:
                    i = this._processDataCSI(ch, i);
                    break;
                case STATE_DCS:
                    i = this._processDataDCS(ch, i);
                    break;
                case STATE_IGNORE:
                    i = this._processDataIgnore(ch, i);
                    break;
                case STATE_APPLICATION_START:
                    this._processDataApplicationStart(ch);
                    break;
                case STATE_APPLICATION_END:
                    i = this._processDataApplicationEnd(ch, data, i);
                    break;
                case STATE_DEC_HASH:
                    this._processDataDecHash(ch);
                    break;
            }
            if (this.y !== this.oldy) {
                this.markRowForRefresh(this.oldy);
                this.markRowForRefresh(this.y);
                this.oldy = this.y;
            }
        }
        this.markRowForRefresh(this.y);
        //  endtime = window.performance.now();
        //console.log("write() end time: " + endtime);
        //  console.log("duration: " + (endtime - starttime) + "ms");
    }
    _processDataCSI(ch, i) {
        // '?', '>', '!'
        if (ch === '?' || ch === '>' || ch === '!') {
            this.prefix = ch;
            return i;
        }
        // 0 - 9
        if (ch >= '0' && ch <= '9') {
            this.currentParam = this.currentParam * 10 + ch.charCodeAt(0) - 48;
            return i;
        }
        // '$', '"', ' ', '\''
        if (ch === '$' || ch === '"' || ch === ' ' || ch === '\'') {
            this.postfix = ch;
            return i;
        }
        this.params.push(this.currentParam);
        this.currentParam = 0;
        // ';'
        if (ch === ';') {
            return i;
        }
        this.state = STATE_NORMAL;
        switch (ch) {
            // CSI Ps A
            // Cursor Up Ps Times (default = 1) (CUU).
            case 'A':
                this.cursorUp(this.params);
                break;
            // CSI Ps B
            // Cursor Down Ps Times (default = 1) (CUD).
            case 'B':
                this.cursorDown(this.params);
                break;
            // CSI Ps C
            // Cursor Forward Ps Times (default = 1) (CUF).
            case 'C':
                this.cursorForward(this.params);
                break;
            // CSI Ps D
            // Cursor Backward Ps Times (default = 1) (CUB).
            case 'D':
                this.cursorBackward(this.params);
                break;
            // CSI Ps ; Ps H
            // Cursor Position [row;column] (default = [1,1]) (CUP).
            case 'H':
                this.cursorPos(this.params);
                break;
            // CSI Ps J  Erase in Display (ED).
            case 'J':
                this.eraseInDisplay(this.params);
                break;
            // CSI Ps K  Erase in Line (EL).
            case 'K':
                this.eraseInLine(this.params);
                break;
            // CSI Pm m  Character Attributes (SGR).
            case 'm':
                if (!this.prefix) {
                    this.charAttributes(this.params);
                }
                break;
            // CSI Ps n  Device Status Report (DSR).
            case 'n':
                if (!this.prefix) {
                    this.deviceStatus(this.params);
                }
                break;
            /**
             * Additions
             */
            // CSI Ps @
            // Insert Ps (Blank) Character(s) (default = 1) (ICH).
            case '@':
                this.insertChars(this.params);
                break;
            // CSI Ps E
            // Cursor Next Line Ps Times (default = 1) (CNL).
            case 'E':
                this.cursorNextLine(this.params);
                break;
            // CSI Ps F
            // Cursor Preceding Line Ps Times (default = 1) (CNL).
            case 'F':
                this.cursorPrecedingLine(this.params);
                break;
            // CSI Ps G
            // Cursor Character Absolute  [column] (default = [row,1]) (CHA).
            case 'G':
                this.cursorCharAbsolute(this.params);
                break;
            // CSI Ps L
            // Insert Ps Line(s) (default = 1) (IL).
            case 'L':
                this.insertLines(this.params);
                break;
            // CSI Ps M
            // Delete Ps Line(s) (default = 1) (DL).
            case 'M':
                this.deleteLines(this.params);
                break;
            // CSI Ps P
            // Delete Ps Character(s) (default = 1) (DCH).
            case 'P':
                this.deleteChars(this.params);
                break;
            // CSI Ps X
            // Erase Ps Character(s) (default = 1) (ECH).
            case 'X':
                this.eraseChars(this.params);
                break;
            // CSI Pm `  Character Position Absolute
            //   [column] (default = [row,1]) (HPA).
            case '`':
                this.charPosAbsolute(this.params);
                break;
            // 141 61 a * HPR -
            // Horizontal Position Relative
            case 'a':
                this.HPositionRelative(this.params);
                break;
            // CSI P s c
            // Send Device Attributes (Primary DA).
            // CSI > P s c
            // Send Device Attributes (Secondary DA)
            case 'c':
                this.sendDeviceAttributes(this.params);
                break;
            // CSI Pm d
            // Line Position Absolute  [row] (default = [1,column]) (VPA).
            case 'd':
                this.linePosAbsolute(this.params);
                break;
            // 145 65 e * VPR - Vertical Position Relative
            case 'e':
                this.VPositionRelative(this.params);
                break;
            // CSI Ps ; Ps f
            //   Horizontal and Vertical Position [row;column] (default =
            //   [1,1]) (HVP).
            case 'f':
                this.HVPosition(this.params);
                break;
            // CSI Pm h  Set Mode (SM).
            // CSI ? Pm h - mouse escape codes, cursor escape codes
            case 'h':
                this.setMode(this.params);
                break;
            // CSI Pm l  Reset Mode (RM).
            // CSI ? Pm l
            case 'l':
                this.resetMode(this.params);
                break;
            // CSI Ps ; Ps r
            //   Set Scrolling Region [top;bottom] (default = full size of win-
            //   dow) (DECSTBM).
            // CSI ? Pm r
            case 'r':
                this.setScrollRegion(this.params);
                break;
            // CSI s     Save cursor (ANSI.SYS).
            // CSI ? Pm s
            case 's':
                if (this.prefix === '?') {
                    this.savePrivateValues(this.params);
                }
                else if (this.prefix === '') {
                    this.saveCursor();
                }
                break;
            // CSI u
            //   Restore cursor (ANSI.SYS).
            case 'u':
                if (this.prefix === '') {
                    this.restoreCursor();
                }
                break;
            /**
             * Lesser Used
             */
            // CSI Ps I
            // Cursor Forward Tabulation Ps tab stops (default = 1) (CHT).
            case 'I':
                this.cursorForwardTab(this.params);
                break;
            // CSI Ps S  Scroll up Ps lines (default = 1) (SU).
            case 'S':
                this.scrollUp(this.params);
                break;
            // CSI Ps T  Scroll down Ps lines (default = 1) (SD).
            // CSI Ps ; Ps ; Ps ; Ps ; Ps T
            // CSI > Ps; Ps T
            case 'T':
                // if (this.prefix === '>') {
                //   this.resetTitleModes(this.params);
                //   break;
                // }
                // if (this.params.length > 2) {
                //   this.initMouseTracking(this.params);
                //   break;
                // }
                if (this.params.length < 2 && !this.prefix) {
                    this.scrollDown(this.params);
                }
                break;
            // CSI Ps Z
            // Cursor Backward Tabulation Ps tab stops (default = 1) (CBT).
            case 'Z':
                this.cursorBackwardTab(this.params);
                break;
            // CSI Ps b  Repeat the preceding graphic character Ps times (REP).
            case 'b':
                this.repeatPrecedingCharacter(this.params);
                break;
            // CSI Ps g  Tab Clear (TBC).
            case 'g':
                this.tabClear(this.params);
                break;
            // CSI Pm i  Media Copy (MC).
            // CSI ? Pm i
            // case 'i':
            //   this.mediaCopy(this.params);
            //   break;
            // CSI Pm m  Character Attributes (SGR).
            // CSI > Ps; Ps m
            // case 'm': // duplicate
            //   if (this.prefix === '>') {
            //     this.setResources(this.params);
            //   } else {
            //     this.charAttributes(this.params);
            //   }
            //   break;
            // CSI Ps n  Device Status Report (DSR).
            // CSI > Ps n
            // case 'n': // duplicate
            //   if (this.prefix === '>') {
            //     this.disableModifiers(this.params);
            //   } else {
            //     this.deviceStatus(this.params);
            //   }
            //   break;
            // CSI > Ps p  Set pointer mode.
            // CSI ! p   Soft terminal reset (DECSTR).
            // CSI Ps$ p
            //   Request ANSI mode (DECRQM).
            // CSI ? Ps$ p
            //   Request DEC private mode (DECRQM).
            // CSI Ps ; Ps " p
            case 'p':
                switch (this.prefix) {
                    // case '>':
                    //   this.setPointerMode(this.params);
                    //   break;
                    case '!':
                        this.softReset(this.params);
                        break;
                }
                break;
            // CSI Ps q  Load LEDs (DECLL).
            // CSI Ps SP q
            // CSI Ps " q
            // case 'q':
            //   if (this.postfix === ' ') {
            //     this.setCursorStyle(this.params);
            //     break;
            //   }
            //   if (this.postfix === '"') {
            //     this.setCharProtectionAttr(this.params);
            //     break;
            //   }
            //   this.loadLEDs(this.params);
            //   break;
            // CSI Ps ; Ps r
            //   Set Scrolling Region [top;bottom] (default = full size of win-
            //   dow) (DECSTBM).
            // CSI ? Pm r
            // CSI Pt; Pl; Pb; Pr; Ps$ r
            // case 'r': // duplicate
            //   if (this.prefix === '?') {
            //     this.restorePrivateValues(this.params);
            //   } else if (this.postfix === '$') {
            //     this.setAttrInRectangle(this.params);
            //   } else {
            //     this.setScrollRegion(this.params);
            //   }
            //   break;
            // CSI Ps ; Ps ; Ps t
            // CSI Pt; Pl; Pb; Pr; Ps$ t
            // CSI > Ps; Ps t
            // CSI Ps SP t
            // case 't':
            //   if (this.postfix === '$') {
            //     this.reverseAttrInRectangle(this.params);
            //   } else if (this.postfix === ' ') {
            //     this.setWarningBellVolume(this.params);
            //   } else {
            //     if (this.prefix === '>') {
            //       this.setTitleModeFeature(this.params);
            //     } else {
            //       this.manipulateWindow(this.params);
            //     }
            //   }
            //   break;
            // CSI u     Restore cursor (ANSI.SYS).
            // CSI Ps SP u
            // case 'u': // duplicate
            //   if (this.postfix === ' ') {
            //     this.setMarginBellVolume(this.params);
            //   } else {
            //     this.restoreCursor(this.params);
            //   }
            //   break;
            // CSI Pt; Pl; Pb; Pr; Pp; Pt; Pl; Pp$ v
            // case 'v':
            //   if (this.postfix === '$') {
            //     this.copyRectagle(this.params);
            //   }
            //   break;
            // CSI Pt ; Pl ; Pb ; Pr ' w
            // case 'w':
            //   if (this.postfix === '\'') {
            //     this.enableFilterRectangle(this.params);
            //   }
            //   break;
            // CSI Ps x  Request Terminal Parameters (DECREQTPARM).
            // CSI Ps x  Select Attribute Change Extent (DECSACE).
            // CSI Pc; Pt; Pl; Pb; Pr$ x
            // case 'x':
            //   if (this.postfix === '$') {
            //     this.fillRectangle(this.params);
            //   } else {
            //     this.requestParameters(this.params);
            //     //this.__(this.params);
            //   }
            //   break;
            // CSI Ps ; Pu ' z
            // CSI Pt; Pl; Pb; Pr$ z
            // case 'z':
            //   if (this.postfix === '\'') {
            //     this.enableLocatorReporting(this.params);
            //   } else if (this.postfix === '$') {
            //     this.eraseRectangle(this.params);
            //   }
            //   break;
            // CSI Pm ' {
            // CSI Pt; Pl; Pb; Pr$ {
            // case '{':
            //   if (this.postfix === '\'') {
            //     this.setLocatorEvents(this.params);
            //   } else if (this.postfix === '$') {
            //     this.selectiveEraseRectangle(this.params);
            //   }
            //   break;
            // CSI Ps ' |
            // case '|':
            //   if (this.postfix === '\'') {
            //     this.requestLocatorPosition(this.params);
            //   }
            //   break;
            // CSI P m SP }
            // Insert P s Column(s) (default = 1) (DECIC), VT420 and up.
            // case '}':
            //   if (this.postfix === ' ') {
            //     this.insertColumns(this.params);
            //   }
            //   break;
            // CSI P m SP ~
            // Delete P s Column(s) (default = 1) (DECDC), VT420 and up
            // case '~':
            //   if (this.postfix === ' ') {
            //     this.deleteColumns(this.params);
            //   }
            //   break;
            default:
                this.error('Unknown CSI code: %s (%i).', ch, "" + ch.charCodeAt(0));
                break;
        }
        this.prefix = '';
        this.postfix = '';
        return i;
    }
    _processDataEscape(ch, i) {
        switch (ch) {
            // ESC [ Control Sequence Introducer ( CSI is 0x9b).
            case '[':
                this.params = [];
                this.currentParam = 0;
                this.state = STATE_CSI;
                break;
            // ESC ] Operating System Command ( OSC is 0x9d).
            case ']':
                this.params = [];
                this.currentParam = 0;
                this.state = STATE_OSC;
                break;
            // ESC & Application mode
            case '&':
                this.params = [];
                this.currentParam = "";
                this.state = STATE_APPLICATION_START;
                break;
            // ESC P Device Control String ( DCS is 0x90).
            case 'P':
                this.params = [];
                this.currentParam = 0;
                this.state = STATE_DCS;
                break;
            // ESC _ Application Program Command ( APC is 0x9f).
            case '_':
                this.state = STATE_IGNORE;
                break;
            // ESC ^ Privacy Message ( PM is 0x9e).
            case '^':
                this.state = STATE_IGNORE;
                break;
            // ESC c Full Reset (RIS).
            case 'c':
                this._fullReset();
                break;
            // ESC E Next Line ( NEL is 0x85).
            case 'E':
                this.x = 0;
                this.index();
                break;
            // ESC D Index ( IND is 0x84).
            case 'D':
                this.index();
                break;
            // ESC M Reverse Index ( RI is 0x8d).
            case 'M':
                this.reverseIndex();
                break;
            // ESC % Select default/utf-8 character set.
            // @ = default, G = utf-8
            case '%':
                //this.charset = null;
                this.setgLevel(0);
                this.setgCharset(0, Emulator.charsets.US);
                this.state = STATE_NORMAL;
                i++;
                break;
            // ESC (,),*,+,-,. Designate G0-G2 Character Set.
            case '(': // <-- this seems to get all the attention
            case ')':
            case '*':
            case '+':
            case '-':
            case '.':
                switch (ch) {
                    case '(':
                        this.gcharset = 0;
                        break;
                    case ')':
                        this.gcharset = 1;
                        break;
                    case '*':
                        this.gcharset = 2;
                        break;
                    case '+':
                        this.gcharset = 3;
                        break;
                    case '-':
                        this.gcharset = 1;
                        break;
                    case '.':
                        this.gcharset = 2;
                        break;
                }
                this.state = STATE_CHARSET;
                break;
            // Designate G3 Character Set (VT300).
            // A = ISO Latin-1 Supplemental.
            // Not implemented.
            case '/':
                this.gcharset = 3;
                this.state = STATE_CHARSET;
                i--;
                break;
            // ESC N
            // Single Shift Select of G2 Character Set
            // ( SS2 is 0x8e). This affects next character only.
            case 'N':
                break;
            // ESC O
            // Single Shift Select of G3 Character Set
            // ( SS3 is 0x8f). This affects next character only.
            case 'O':
                break;
            // ESC n
            // Invoke the G2 Character Set as GL (LS2).
            case 'n':
                this.setgLevel(2);
                break;
            // ESC o
            // Invoke the G3 Character Set as GL (LS3).
            case 'o':
                this.setgLevel(3);
                break;
            // ESC |
            // Invoke the G3 Character Set as GR (LS3R).
            case '|':
                this.setgLevel(3);
                break;
            // ESC }
            // Invoke the G2 Character Set as GR (LS2R).
            case '}':
                this.setgLevel(2);
                break;
            // ESC ~
            // Invoke the G1 Character Set as GR (LS1R).
            case '~':
                this.setgLevel(1);
                break;
            // ESC 7 Save Cursor (DECSC).
            case '7':
                this.saveCursor();
                this.state = STATE_NORMAL;
                break;
            // ESC 8 Restore Cursor (DECRC).
            case '8':
                this.restoreCursor();
                this.state = STATE_NORMAL;
                break;
            // ESC # 3 DEC line height/width
            case '#':
                this.state = STATE_DEC_HASH;
                break;
            // ESC H Tab Set (HTS is 0x88).
            case 'H':
                this.tabSet();
                break;
            // ESC = Application Keypad (DECPAM).
            case '=':
                this.log('Serial port requested application keypad.');
                this.applicationKeypad = true;
                this.state = STATE_NORMAL;
                break;
            // ESC > Normal Keypad (DECPNM).
            case '>':
                this.log('Switching back to normal keypad.');
                this.applicationKeypad = false;
                this.state = STATE_NORMAL;
                break;
            default:
                this.state = STATE_NORMAL;
                this.error('Unknown ESC control: %s.', ch);
                break;
        }
        return i;
    }
    _processDataCharset(ch, i) {
        let cs;
        switch (ch) {
            case '0':// DEC Special Character and Line Drawing Set.
                cs = Emulator.charsets.SCLD;
                break;
            case 'A':// UK
                cs = Emulator.charsets.UK;
                break;
            case 'B':// United States (USASCII).
                cs = Emulator.charsets.US;
                break;
            case '4':// Dutch
                cs = Emulator.charsets.Dutch;
                break;
            case 'C': // Finnish
            case '5':
                cs = Emulator.charsets.Finnish;
                break;
            case 'R':// French
                cs = Emulator.charsets.French;
                break;
            case 'Q':// FrenchCanadian
                cs = Emulator.charsets.FrenchCanadian;
                break;
            case 'K':// German
                cs = Emulator.charsets.German;
                break;
            case 'Y':// Italian
                cs = Emulator.charsets.Italian;
                break;
            case 'E': // NorwegianDanish
            case '6':
                cs = Emulator.charsets.NorwegianDanish;
                break;
            case 'Z':// Spanish
                cs = Emulator.charsets.Spanish;
                break;
            case 'H': // Swedish
            case '7':
                cs = Emulator.charsets.Swedish;
                break;
            case '=':// Swiss
                cs = Emulator.charsets.Swiss;
                break;
            case '/':// ISOLatin (actually /A)
                cs = Emulator.charsets.ISOLatin;
                i++;
                break;
            default:// Default
                cs = Emulator.charsets.US;
                break;
        }
        this.setgCharset(this.gcharset, cs);
        this.gcharset = null;
        this.state = STATE_NORMAL;
        return i;
    }
    _processDataOSC(ch, i) {
        // OSC Ps ; Pt ST
        // OSC Ps ; Pt BEL
        //   Set Text Parameters.
        if (ch === '\x1b' || ch === '\x07') {
            if (ch === '\x1b') {
                i++;
            }
            this.params.push(this.currentParam);
            switch (this.params[0]) {
                case 0:
                case 1:
                case 2:
                    if (this.params[1]) {
                        this.title = this.params[1];
                        this.handleTitle(this.title);
                    }
                    break;
                case 3:
                    // set X property
                    break;
                case 4:
                case 5:
                    // change dynamic colors
                    break;
                case 10:
                case 11:
                case 12:
                case 13:
                case 14:
                case 15:
                case 16:
                case 17:
                case 18:
                case 19:
                    // change dynamic ui colors
                    break;
                case 46:
                    // change log file
                    break;
                case 50:
                    // dynamic font
                    break;
                case 51:
                    // emacs shell
                    break;
                case 52:
                    // manipulate selection data
                    break;
                case 104:
                case 105:
                case 110:
                case 111:
                case 112:
                case 113:
                case 114:
                case 115:
                case 116:
                case 117:
                case 118:
                    // reset colors
                    break;
            }
            this.params = [];
            this.currentParam = 0;
            this.state = STATE_NORMAL;
        }
        else {
            if (!this.params.length) {
                if (ch >= '0' && ch <= '9') {
                    this.currentParam =
                        this.currentParam * 10 + ch.charCodeAt(0) - 48;
                }
                else if (ch === ';') {
                    this.params.push(this.currentParam);
                    this.currentParam = '';
                }
            }
            else {
                this.currentParam += ch;
            }
        }
        return i;
    }
    _processDataDCS(ch, i) {
        if (ch === '\x1b' || ch === '\x07') {
            if (ch === '\x1b')
                i++;
            switch (this.prefix) {
                // User-Defined Keys (DECUDK).
                case '':
                    break;
                // Request Status String (DECRQSS).
                // test: echo -e '\eP$q"p\e\\'
                case '$q':
                    var pt = this.currentParam;
                    var valid = false;
                    let replyPt = "";
                    switch (pt) {
                        // DECSCA
                        case '"q':
                            replyPt = '0"q';
                            break;
                        // DECSCL
                        case '"p':
                            replyPt = '61"p';
                            break;
                        // DECSTBM
                        case 'r':
                            replyPt = '' +
                                (this.scrollTop + 1) +
                                ';' +
                                (this.scrollBottom + 1) +
                                'r';
                            break;
                        // SGR
                        case 'm':
                            replyPt = '0m';
                            break;
                        default:
                            this.error('Unknown DCS Pt: %s.', "" + pt);
                            replyPt = '';
                            break;
                    }
                    this.send('\x1bP' + (valid ? 1 : 0) + '$r' + replyPt + '\x1b\\');
                    break;
                // Set Termcap/Terminfo Data (xterm, experimental).
                case '+p':
                    break;
                // Request Termcap/Terminfo String (xterm, experimental)
                // Regular xterm does not even respond to this sequence.
                // This can cause a small glitch in vim.
                // test: echo -ne '\eP+q6b64\e\\'
                case '+q':
                    pt = this.currentParam;
                    valid = false;
                    this.send('\x1bP' + (valid ? 1 : 0) + '+r' + pt + '\x1b\\');
                    break;
                default:
                    this.error('Unknown DCS prefix: %s.', this.prefix);
                    break;
            }
            this.currentParam = 0;
            this.prefix = '';
            this.state = STATE_NORMAL;
        }
        else if (!this.currentParam) {
            if (!this.prefix && ch !== '$' && ch !== '+') {
                this.currentParam = ch;
            }
            else if (this.prefix.length === 2) {
                this.currentParam = ch;
            }
            else {
                this.prefix += ch;
            }
        }
        else {
            this.currentParam += ch;
        }
        return i;
    }
    _processDataIgnore(ch, i) {
        // For PM and APC.
        if (ch === '\x1b' || ch === '\x07') {
            if (ch === '\x1b') {
                i++;
            }
            this.state = STATE_NORMAL;
        }
        return i;
    }
    _processDataApplicationStart(ch) {
        if ((ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z') || (ch >= '0' && ch <= '9') || ch === '-'
            || ch === '/') {
            // Add to the current parameter.
            this.currentParam += ch; // FIXME don't absorb infinite data here.
        }
        else if (ch === ';') {
            // Parameter separator.
            this.params.push(this.currentParam);
            this.currentParam = '';
        }
        else if (ch === '\x07') {
            // End of parameters.
            this.params.push(this.currentParam);
            if (this.params[0] === this.applicationModeCookie) {
                this.state = STATE_APPLICATION_END;
                this._dispatchEvents();
                this._emit(APPLICATIONMODESTART_EVENT, this, this.params);
            }
            else {
                this.log("Invalid application mode cookie.");
                this.state = STATE_NORMAL;
            }
        }
        else {
            // Invalid application start.
            this.state = STATE_NORMAL;
            this.log("Invalid application mode start command.");
        }
    }
    _processDataApplicationEnd(ch, data, i) {
        // Efficiently look for an end-mode character.
        const nextzero = data.indexOf('\x00', i);
        if (nextzero === -1) {
            // Send all of the data on right now.
            this._emit(APPLICATIONMODEDATA_EVENT, data.slice(i));
            i = data.length - 1;
        }
        else if (nextzero === i) {
            // We are already at the end-mode character.
            this._dispatchEvents();
            this._emit(APPLICATIONMODEEND_EVENT, this);
            this.state = STATE_NORMAL;
        }
        else {
            // Incoming end-mode character. Send the last piece of data.
            this._dispatchEvents();
            this._emit(APPLICATIONMODEDATA_EVENT, data.slice(i, nextzero));
            i = nextzero - 1;
        }
        return i;
    }
    // ESC # variations
    _processDataDecHash(ch) {
        switch (ch) {
            // ESC # 8
            // Screen Alignment Display (DECALN)
            case '8':
                this.fillScreen('E');
                break;
            default:
                break;
        }
        this.state = STATE_NORMAL;
    }
    writeln(data) {
        this.write(data + '\r\n');
    }
    ;
    keyDown(ev) {
        let key = null;
        // Modifiers keys are often encoded using this scheme.
        const mod = (ev.shiftKey ? 1 : 0) + (ev.altKey ? 2 : 0) + (ev.ctrlKey ? 4 : 0) + 1;
        const modStr = mod === 1 ? "" : ";" + mod;
        switch (ev.key) {
            // backspace
            case 'Backspace':
                if ((!this.isMac && ev.altKey) || (this.isMac && ev.metaKey)) {
                    key = '\x1b\x7f'; // ^[^?
                    break;
                }
                key = '\x7f'; // ^?
                break;
            // tab
            case 'Tab':
                if (ev.shiftKey) {
                    key = '\x1b[Z';
                }
                else {
                    key = '\t';
                }
                break;
            // return/enter
            case 'Enter':
                key = '\r';
                break;
            // escape
            case 'Escape':
                key = '\x1b';
                break;
            // left-arrow
            case 'ArrowLeft':
                if (this.applicationCursor) {
                    key = '\x1bOD'; // SS3 as ^[O for 7-bit
                    //key = '\x8fD'; // SS3 as 0x8f for 8-bit
                    break;
                }
                if (mod !== 1) {
                    key = "\x1b[1" + modStr + "D";
                }
                else {
                    key = '\x1b[D';
                }
                break;
            // right-arrow
            case 'ArrowRight':
                if (this.applicationCursor) {
                    key = '\x1bOC';
                    break;
                }
                if (mod !== 1) {
                    key = "\x1b[1" + modStr + "C";
                }
                else {
                    key = '\x1b[C';
                }
                break;
            // up-arrow
            case 'ArrowUp':
                if (this.applicationCursor) {
                    key = '\x1bOA';
                    break;
                }
                if (mod !== 1) {
                    key = "\x1b[1" + modStr + "A";
                }
                else {
                    key = '\x1b[A';
                }
                break;
            // down-arrow
            case 'ArrowDown':
                if (this.applicationCursor) {
                    key = '\x1bOB';
                    break;
                }
                if (mod !== 1) {
                    key = "\x1b[1" + modStr + "B";
                }
                else {
                    key = '\x1b[B';
                }
                break;
            // home
            case 'Home':
                if (this.applicationKeypad) {
                    key = '\x1bOH';
                    break;
                }
                if (mod !== 1) {
                    key = '\x1b[1' + modStr + 'H';
                }
                else {
                    key = '\x1b[H';
                }
                break;
            // end
            case 'End':
                if (this.applicationKeypad) {
                    key = '\x1bOF';
                    break;
                }
                if (mod !== 1) {
                    key = '\x1b[1' + modStr + 'F';
                }
                else {
                    key = '\x1b[F';
                }
                break;
            case 'PageUp':
            case 'PageDown':
            case 'Delete':
            case 'Insert':
            case 'F1':
            case 'F2':
            case 'F3':
            case 'F4':
            case 'F5':
            case 'F6':
            case 'F7':
            case 'F8':
            case 'F9':
            case 'F10':
            case 'F11':
            case 'F12':
                if (mod === 1) {
                    switch (ev.key) {
                        case 'F1':
                            key = '\x1bOP';
                            break;
                        case 'F2':
                            key = '\x1bOQ';
                            break;
                        case 'F3':
                            key = '\x1bOR';
                            break;
                        case 'F4':
                            key = '\x1bOS';
                            break;
                        default: break;
                    }
                    if (key !== null) {
                        break;
                    }
                }
                switch (ev.key) {
                    case 'PageUp':
                        key = '\x1b[5' + modStr + '~';
                        break;
                    case 'PageDown':
                        key = '\x1b[6' + modStr + '~';
                        break;
                    case 'Delete':
                        key = '\x1b[3' + modStr + '~';
                        break;
                    case 'Insert':
                        key = '\x1b[2' + modStr + '~';
                        break;
                    case 'F1':
                        key = '\x1b[1' + modStr + 'P';
                        break;
                    case 'F2':
                        key = '\x1b[1' + modStr + 'Q';
                        break;
                    case 'F3':
                        key = '\x1b[1' + modStr + 'R';
                        break;
                    case 'F4':
                        key = '\x1b[1' + modStr + 'S';
                        break;
                    case 'F5':
                        key = '\x1b[15' + modStr + '~';
                        break;
                    case 'F6':
                        key = '\x1b[17' + modStr + '~';
                        break;
                    case 'F7':
                        key = '\x1b[18' + modStr + '~';
                        break;
                    case 'F8':
                        key = '\x1b[19' + modStr + '~';
                        break;
                    case 'F9':
                        key = '\x1b[20' + modStr + '~';
                        break;
                    case 'F10':
                        key = '\x1b[21' + modStr + '~';
                        break;
                    case 'F11':
                        key = '\x1b[23' + modStr + '~';
                        break;
                    case 'F12':
                        key = '\x1b[24' + modStr + '~';
                        break;
                    default: break;
                }
                break;
            default:
                // Control codes
                if (ev.key.length === 1) {
                    if (ev.ctrlKey) {
                        if (ev.key >= '@' && ev.key <= '_') {
                            key = String.fromCodePoint(ev.key.codePointAt(0) - '@'.codePointAt(0));
                        }
                        else if (ev.key >= 'a' && ev.key <= 'z') {
                            key = String.fromCodePoint(ev.key.codePointAt(0) - 'a'.codePointAt(0) + 1);
                        }
                        else if (ev.key === ' ') {
                            // Ctrl space
                            key = '\x00';
                        }
                    }
                    else if ((!this.isMac && ev.altKey) || (this.isMac && ev.metaKey)) {
                        if (ev.key.length === 1) {
                            key = '\x1b' + ev.key;
                        }
                    }
                }
                break;
        }
        if (key === null) {
            return false;
        }
        this.handler(key);
        cancelEvent(ev);
        return true;
    }
    setgLevel(g) {
        this.glevel = g;
        this.charset = this.charsets[g];
    }
    setgCharset(g, charset) {
        this.charsets[g] = charset;
        if (this.glevel === g) {
            this.charset = charset;
        }
    }
    keyPress(ev) {
        const key = ev.key;
        if (!key || key.length !== 1 || ev.ctrlKey || ev.altKey || ev.metaKey) {
            return false;
        }
        this.handler(key);
        cancelEvent(ev);
        return true;
    }
    send(data) {
        if (!this.queue) {
            setTimeout(() => {
                this.handler(this.queue);
                this.queue = '';
            }, 1);
        }
        this.queue += data;
    }
    bell() {
        this._emit(BELL_EVENT, this);
    }
    newLine() {
        // TODO: Implement eat_newline_glitch.
        // if (this.realX >= this.cols) break;
        // this.realX = 0;
        if (this.y + 1 > this.scrollBottom) {
            this.scroll();
        }
        else {
            this._setCursorY(this.y + 1);
        }
    }
    carriageReturn() {
        this.x = 0;
    }
    log(...args) {
        if (!this.debug) {
            return;
        }
        console.log.apply(console, ["[TERM] ", ...args]);
        console.log.apply(console, ["[TERM] " + args]);
    }
    error(...args) {
        if (!this.debug)
            return;
        console.error.apply(console, args);
        console.trace();
    }
    size() {
        return { rows: this.rows, columns: this.cols };
    }
    resize(newSize) {
        const newcols = Math.max(newSize.columns, 1);
        const newrows = Math.max(newSize.rows, 1);
        // resize cols
        if (this.cols < newcols) {
            // Add chars to the lines to match the new (bigger) cols value.
            const chCodePoint = ' '.codePointAt(0);
            for (let i = this.lines.length - 1; i >= 0; i--) {
                const { chars, attrs } = this.lines[i];
                const biggerChars = new Uint32Array(newcols);
                biggerChars.set(chars);
                const biggerAttrs = new Uint32Array(newcols);
                biggerAttrs.set(attrs);
                for (let j = chars.length; j < newcols; j++) {
                    biggerChars[j] = chCodePoint;
                    biggerAttrs[j] = Emulator.defAttr;
                }
                this.lines[i] = { chars: biggerChars, attrs: biggerAttrs };
            }
        }
        else if (this.cols > newcols) {
            // Remove chars from the lines to match the new (smaller) cols value.
            for (let i = this.lines.length - 1; i >= 0; i--) {
                const line = this.lines[i];
                this.lines[i] = { chars: line.chars.slice(0, newcols), attrs: line.attrs.slice(0, newcols) };
            }
        }
        this.setupStops(newcols);
        this.cols = newcols;
        // resize rows
        if (this.rows < newrows) {
        }
        else if (this.rows > newrows) {
            // Remove rows to match the new smaller rows value.
            while (this.lines.length > newrows) {
                this._scrollbackLineQueue.push(this.lines.shift());
            }
        }
        this.rows = newrows;
        // make sure the cursor stays on screen
        if (this.y >= newrows) {
            this._setCursorY(newrows - 1);
        }
        if (this.x >= newcols) {
            this.x = newcols - 1;
        }
        this.scrollTop = 0;
        this.scrollBottom = newrows - 1;
        // it's a real nightmare trying
        // to resize the original
        // screen buffer. just set it
        // to null for now.
        this.normal = null;
        this.lastReportedPhysicalHeight = this.lines.length;
        this.markRowRangeForRefresh(0, this.lines.length - 1);
        this._dispatchEvents();
    }
    _dispatchEvents() {
        let checkedRefreshEnd = Math.min(this._refreshEnd, this.lines.length);
        let checkedRefreshStart = this._refreshStart;
        if (checkedRefreshEnd === REFRESH_END_NULL || checkedRefreshStart === checkedRefreshEnd) {
            // Don't signal to refresh anything. This can happen when there are no realized rows yet.
            checkedRefreshEnd = -1;
            checkedRefreshStart = -1;
        }
        const event = {
            rows: this.rows,
            columns: this.cols,
            realizedRows: this.lines.length,
            refreshStartRow: checkedRefreshStart,
            refreshEndRow: checkedRefreshEnd,
            scrollbackLines: this._scrollbackLineQueue
        };
        this._refreshStart = REFRESH_START_NULL;
        this._refreshEnd = REFRESH_END_NULL;
        this._scrollbackLineQueue = [];
        this._emit(RENDER_EVENT, this, event);
    }
    markRowForRefresh(y) {
        this.markRowRangeForRefresh(y, y);
    }
    _setCursorY(newY) {
        this._getRow(newY);
        this.y = newY;
        this.markRowRangeForRefresh(newY, newY);
    }
    markAllRowsForRefresh() {
        this.markRowRangeForRefresh(0, this.rows - 1);
    }
    setupStops(maxCols) {
        const cols = maxCols === undefined ? this.cols : maxCols;
        this.tabs = {};
        for (let i = 0; i < cols; i += 8) {
            this.tabs[i] = true;
        }
    }
    prevStop(x) {
        if (x === undefined)
            x = this.x;
        while (!this.tabs[--x] && x > 0)
            ;
        return x >= this.cols ? this.cols - 1 : (x < 0 ? 0 : x);
    }
    nextStop(x) {
        if (x === undefined)
            x = this.x;
        while (!this.tabs[++x] && x < this.cols)
            ;
        return x >= this.cols ? this.cols - 1 : (x < 0 ? 0 : x);
    }
    eraseRight(x, y) {
        this.fillRight(x, y);
    }
    fillRight(x, y, ch = ' ') {
        const line = this._tryGetRow(y);
        if (line === null) {
            return;
        }
        const { chars, attrs } = line;
        const attr = this.eraseAttr();
        const chCodePoint = ch.codePointAt(0);
        for (; x < this.cols; x++) {
            chars[x] = chCodePoint;
            attrs[x] = attr;
        }
        this.markRowForRefresh(y);
    }
    fillScreen(fillChar = ' ') {
        let j = this.rows;
        while (j--) {
            this.fillRight(0, j, fillChar);
        }
    }
    eraseLeft(x, y) {
        const line = this._getRow(y);
        const { chars, attrs } = line;
        const attr = this.eraseAttr();
        const space = ' '.codePointAt(0);
        x++;
        while (x !== 0) {
            x--;
            chars[x] = space;
            attrs[x] = attr;
        }
        this.markRowForRefresh(y);
    }
    eraseLine(y) {
        this.eraseRight(0, y);
    }
    blankLine(cur) {
        const chars = new Uint32Array(this.cols);
        const attrs = new Uint32Array(this.cols);
        const line = { chars, attrs };
        chars.fill(' '.codePointAt(0));
        attrs.fill(cur ? this.eraseAttr() : Emulator.defAttr);
        return line;
    }
    is(term) {
        const name = this.termName;
        return (name + '').indexOf(term) === 0;
    }
    handler(data) {
        this._emit(DATA_EVENT, this, data);
    }
    handleTitle(title) {
        this._emit(TITLE_EVENT, this, title);
    }
    /**
     * ESC
     */
    // ESC D Index (IND is 0x84).
    index() {
        if (this.y + 1 > this.scrollBottom) {
            this.scroll();
        }
        else {
            this._setCursorY(this.y + 1);
        }
        this.state = STATE_NORMAL;
    }
    // ESC M Reverse Index (RI is 0x8d).
    reverseIndex() {
        if (this.y - 1 < this.scrollTop) {
            // possibly move the code below to term.reverseScroll();
            // test: echo -ne '\e[1;1H\e[44m\eM\e[0m'
            // blankLine(true) is xterm/linux behavior
            this.lines.splice(this.y, 0, this.blankLine(true));
            const j = this.rows - 1 - this.scrollBottom;
            this.lines.splice(this.rows - 1 - j + 1, 1);
            this.markRowForRefresh(this.scrollTop);
            this.markRowForRefresh(this.scrollBottom);
        }
        else {
            this._setCursorY(this.y - 1);
        }
        this.state = STATE_NORMAL;
    }
    reset() {
        this._fullReset();
        this.x = 0;
        this._setCursorY(0);
        this._dispatchEvents();
    }
    // ESC c Full Reset (RIS).
    _fullReset() {
        this._resetVariables();
    }
    // ESC H Tab Set (HTS is 0x88).
    tabSet() {
        this.tabs[this.x] = true;
        this.state = STATE_NORMAL;
    }
    /**
     * CSI
     */
    // CSI Ps A
    // Cursor Up Ps Times (default = 1) (CUU).
    cursorUp(params) {
        let param = params[0];
        if (param < 1) {
            param = 1;
        }
        if (this.y - param < this.scrollTop) {
            this._setCursorY(this.scrollTop);
        }
        else {
            this._setCursorY(this.y - param);
        }
    }
    // CSI Ps B
    // Cursor Down Ps Times (default = 1) (CUD).
    cursorDown(params) {
        let param = params[0];
        if (param < 1) {
            param = 1;
        }
        const bottom = this.scrollBottom !== 0 ? this.scrollBottom : this.rows - 1;
        if (this.y + param > bottom) {
            this._setCursorY(bottom);
        }
        else {
            this._setCursorY(this.y + param);
        }
    }
    // CSI Ps C
    // Cursor Forward Ps Times (default = 1) (CUF).
    cursorForward(params) {
        let param = params[0];
        if (param < 1) {
            param = 1;
        }
        this.x += param;
        if (this.x >= this.cols) {
            this.x = this.cols - 1;
        }
    }
    // CSI Ps D
    // Cursor Backward Ps Times (default = 1) (CUB).
    cursorBackward(params) {
        var param = params[0];
        if (param < 1)
            param = 1;
        this.x -= param;
        if (this.x < 0)
            this.x = 0;
    }
    ;
    // CSI Ps ; Ps H
    // Cursor Position [row;column] (default = [1,1]) (CUP).
    cursorPos(params) {
        let y = params[0] - 1 + (this.originMode ? this.scrollTop : 0);
        let x = (params.length >= 2) ? params[1] - 1 : 0;
        if (y < 0) {
            y = 0;
        }
        else if (y >= this.rows) {
            y = this.rows - 1;
        }
        if (x < 0) {
            x = 0;
        }
        else if (x >= this.cols) {
            x = this.cols - 1;
        }
        this.x = x;
        this._setCursorY(y);
    }
    // CSI Ps J  Erase in Display (ED).
    //     Ps = 0  -> Erase Below (default).
    //     Ps = 1  -> Erase Above.
    //     Ps = 2  -> Erase All.
    //     Ps = 3  -> Erase Saved Lines (xterm).
    // CSI ? Ps J
    //   Erase in Display (DECSED).
    //     Ps = 0  -> Selective Erase Below (default).
    //     Ps = 1  -> Selective Erase Above.
    //     Ps = 2  -> Selective Erase All.
    eraseInDisplay(params) {
        let j;
        switch (params[0]) {
            case 0:
                this.eraseRight(this.x, this.y);
                j = this.y + 1;
                for (; j < this.rows; j++) {
                    this.eraseLine(j);
                }
                break;
            case 1:
                this.eraseLeft(this.x, this.y);
                j = this.y;
                while (j--) {
                    this.eraseLine(j);
                }
                break;
            case 2:
                j = this.rows;
                while (j--) {
                    this.eraseLine(j);
                }
                break;
            case 3:
                // no saved lines
                break;
        }
    }
    // CSI Ps K  Erase in Line (EL).
    //     Ps = 0  -> Erase to Right (default).
    //     Ps = 1  -> Erase to Left.
    //     Ps = 2  -> Erase All.
    // CSI ? Ps K
    //   Erase in Line (DECSEL).
    //     Ps = 0  -> Selective Erase to Right (default).
    //     Ps = 1  -> Selective Erase to Left.
    //     Ps = 2  -> Selective Erase All.
    eraseInLine(params) {
        switch (params[0]) {
            case 0:
                this.eraseRight(this.x, this.y);
                break;
            case 1:
                this.eraseLeft(this.x, this.y);
                break;
            case 2:
                this.eraseLine(this.y);
                break;
        }
    }
    // CSI Pm m  Character Attributes (SGR).
    //     Ps = 0  -> Normal (default).
    //     Ps = 1  -> Bold.
    //     Ps = 2  -> Faint, decreased intensity (ISO 6429).
    //     Ps = 3  -> Italicized (ISO 6429).
    //     Ps = 4  -> Underlined.
    //     Ps = 5  -> Blink (appears as Bold).
    //     Ps = 7  -> Inverse.
    //     Ps = 8  -> Invisible, i.e., hidden (VT300).
    //     Ps = 9  -> Crossed-out characters (ISO 6429).
    //     Ps = 2 2  -> Normal (neither bold nor faint).
    //     Ps = 2 4  -> Not underlined.
    //     Ps = 2 5  -> Steady (not blinking).
    //     Ps = 2 7  -> Positive (not inverse).
    //     Ps = 2 8  -> Visible, i.e., not hidden (VT300).
    //     Ps = 3 0  -> Set foreground color to Black.
    //     Ps = 3 1  -> Set foreground color to Red.
    //     Ps = 3 2  -> Set foreground color to Green.
    //     Ps = 3 3  -> Set foreground color to Yellow.
    //     Ps = 3 4  -> Set foreground color to Blue.
    //     Ps = 3 5  -> Set foreground color to Magenta.
    //     Ps = 3 6  -> Set foreground color to Cyan.
    //     Ps = 3 7  -> Set foreground color to White.
    //     Ps = 3 9  -> Set foreground color to default (original).
    //     Ps = 4 0  -> Set background color to Black.
    //     Ps = 4 1  -> Set background color to Red.
    //     Ps = 4 2  -> Set background color to Green.
    //     Ps = 4 3  -> Set background color to Yellow.
    //     Ps = 4 4  -> Set background color to Blue.
    //     Ps = 4 5  -> Set background color to Magenta.
    //     Ps = 4 6  -> Set background color to Cyan.
    //     Ps = 4 7  -> Set background color to White.
    //     Ps = 4 9  -> Set background color to default (original).
    //   If 16-color support is compiled, the following apply.  Assume
    //   that xterm's resources are set so that the ISO color codes are
    //   the first 8 of a set of 16.  Then the aixterm colors are the
    //   bright versions of the ISO colors:
    //     Ps = 9 0  -> Set foreground color to Black.
    //     Ps = 9 1  -> Set foreground color to Red.
    //     Ps = 9 2  -> Set foreground color to Green.
    //     Ps = 9 3  -> Set foreground color to Yellow.
    //     Ps = 9 4  -> Set foreground color to Blue.
    //     Ps = 9 5  -> Set foreground color to Magenta.
    //     Ps = 9 6  -> Set foreground color to Cyan.
    //     Ps = 9 7  -> Set foreground color to White.
    //     Ps = 1 0 0  -> Set background color to Black.
    //     Ps = 1 0 1  -> Set background color to Red.
    //     Ps = 1 0 2  -> Set background color to Green.
    //     Ps = 1 0 3  -> Set background color to Yellow.
    //     Ps = 1 0 4  -> Set background color to Blue.
    //     Ps = 1 0 5  -> Set background color to Magenta.
    //     Ps = 1 0 6  -> Set background color to Cyan.
    //     Ps = 1 0 7  -> Set background color to White.
    //   If xterm is compiled with the 16-color support disabled, it
    //   supports the following, from rxvt:
    //     Ps = 1 0 0  -> Set foreground and background color to
    //     default.
    //   If 88- or 256-color support is compiled, the following apply.
    //     Ps = 3 8  ; 5  ; Ps -> Set foreground color to the second
    //     Ps.
    //     Ps = 4 8  ; 5  ; Ps -> Set background color to the second
    //     Ps.
    charAttributes(params) {
        // Optimize a single SGR0.
        if (params.length === 1 && params[0] === 0) {
            this.curAttr = Emulator.defAttr;
            return;
        }
        const len = params.length;
        let flags = TermApi_1.flagsFromCharAttr(this.curAttr);
        let fg = TermApi_1.foregroundFromCharAttr(this.curAttr);
        let bg = TermApi_1.backgroundFromCharAttr(this.curAttr);
        for (let i = 0; i < len; i++) {
            let p = params[i];
            if (p >= 30 && p <= 37) {
                // fg color 8
                fg = p - 30;
            }
            else if (p >= 40 && p <= 47) {
                // bg color 8
                bg = p - 40;
            }
            else if (p >= 90 && p <= 97) {
                // fg color 16
                p += 8;
                fg = p - 90;
            }
            else if (p >= 100 && p <= 107) {
                // bg color 16
                p += 8;
                bg = p - 100;
            }
            else if (p === 0) {
                // default
                flags = TermApi_1.flagsFromCharAttr(Emulator.defAttr);
                fg = TermApi_1.foregroundFromCharAttr(Emulator.defAttr);
                bg = TermApi_1.backgroundFromCharAttr(Emulator.defAttr);
                // flags = 0;
                // fg = 0x1ff;
                // bg = 0x1ff;
            }
            else if (p === 1) {
                // bold text
                flags |= TermApi_1.BOLD_ATTR_FLAG;
            }
            else if (p === 2) {
                // Faint, decreased intensity (ISO 6429).
                flags |= TermApi_1.FAINT_ATTR_FLAG;
            }
            else if (p === 3) {
                // Italic
                flags |= TermApi_1.ITALIC_ATTR_FLAG;
            }
            else if (p === 4) {
                // underlined text
                flags |= TermApi_1.UNDERLINE_ATTR_FLAG;
            }
            else if (p === 5) {
                // blink
                flags |= TermApi_1.BLINK_ATTR_FLAG;
            }
            else if (p === 7) {
                // inverse and positive
                // test with: echo -e '\e[31m\e[42mhello\e[7mworld\e[27mhi\e[m'
                flags |= TermApi_1.INVERSE_ATTR_FLAG;
            }
            else if (p === 8) {
                // invisible
                flags |= TermApi_1.INVISIBLE_ATTR_FLAG;
            }
            else if (p === 9) {
                // Crossed-out characters (ISO 6429).
                flags |= TermApi_1.STRIKE_THROUGH_ATTR_FLAG;
            }
            else if (p >= 10 && p <= 20) {
                // Font setting. Ignore
            }
            else if (p === 22) {
                // not bold and not faint.
                flags &= ~TermApi_1.BOLD_ATTR_FLAG;
                flags &= ~TermApi_1.FAINT_ATTR_FLAG;
            }
            else if (p === 23) {
                // not italic
                flags &= ~TermApi_1.ITALIC_ATTR_FLAG;
            }
            else if (p === 24) {
                // not underlined
                flags &= ~TermApi_1.UNDERLINE_ATTR_FLAG;
            }
            else if (p === 25) {
                // not blink
                flags &= ~TermApi_1.BLINK_ATTR_FLAG;
            }
            else if (p === 27) {
                // not inverse
                flags &= ~TermApi_1.INVERSE_ATTR_FLAG;
            }
            else if (p === 28) {
                // not invisible
                flags &= ~TermApi_1.INVISIBLE_ATTR_FLAG;
            }
            else if (p === 29) {
                // not strike through
                flags &= ~TermApi_1.STRIKE_THROUGH_ATTR_FLAG;
            }
            else if (p === 39) {
                // reset fg
                fg = TermApi_1.foregroundFromCharAttr(Emulator.defAttr);
            }
            else if (p === 49) {
                // reset bg
                bg = TermApi_1.backgroundFromCharAttr(Emulator.defAttr);
            }
            else if (p === 38) {
                // fg color 256
                if (params[i + 1] === 2) {
                    i += 2;
                    fg = matchColor(params[i] & 0xff, params[i + 1] & 0xff, params[i + 2] & 0xff);
                    if (fg === -1) {
                        fg = 0x1ff;
                    }
                    i += 2;
                }
                else if (params[i + 1] === 5) {
                    i += 2;
                    p = params[i] & 0xff;
                    fg = p;
                }
            }
            else if (p === 48) {
                // bg color 256
                if (params[i + 1] === 2) {
                    i += 2;
                    bg = matchColor(params[i] & 0xff, params[i + 1] & 0xff, params[i + 2] & 0xff);
                    if (bg === -1) {
                        bg = 0x1ff;
                    }
                    i += 2;
                }
                else if (params[i + 1] === 5) {
                    i += 2;
                    p = params[i] & 0xff;
                    bg = p;
                }
            }
            else if (p === 100) {
                // reset fg/bg
                fg = TermApi_1.foregroundFromCharAttr(Emulator.defAttr);
                bg = TermApi_1.backgroundFromCharAttr(Emulator.defAttr);
            }
            else {
                this.error('Unknown SGR attribute: %s.', "" + p);
            }
        }
        this.curAttr = packCharAttr(flags, fg, bg);
    }
    // CSI Ps n  Device Status Report (DSR).
    //     Ps = 5  -> Status Report.  Result (``OK'') is
    //   CSI 0 n
    //     Ps = 6  -> Report Cursor Position (CPR) [row;column].
    //   Result is
    //   CSI r ; c R
    // CSI ? Ps n
    //   Device Status Report (DSR, DEC-specific).
    //     Ps = 6  -> Report Cursor Position (CPR) [row;column] as CSI
    //     ? r ; c R (assumes page is zero).
    //     Ps = 1 5  -> Report Printer status as CSI ? 1 0  n  (ready).
    //     or CSI ? 1 1  n  (not ready).
    //     Ps = 2 5  -> Report UDK status as CSI ? 2 0  n  (unlocked)
    //     or CSI ? 2 1  n  (locked).
    //     Ps = 2 6  -> Report Keyboard status as
    //   CSI ? 2 7  ;  1  ;  0  ;  0  n  (North American).
    //   The last two parameters apply to VT400 & up, and denote key-
    //   board ready and LK01 respectively.
    //     Ps = 5 3  -> Report Locator status as
    //   CSI ? 5 3  n  Locator available, if compiled-in, or
    //   CSI ? 5 0  n  No Locator, if not.
    deviceStatus(params) {
        if (!this.prefix) {
            switch (params[0]) {
                case 5:
                    // status report
                    this.send('\x1b[0n');
                    break;
                case 6:
                    // cursor position
                    this.send('\x1b[' + (this.y + 1) + ';' + (this.x + 1) + 'R');
                    break;
            }
        }
        else if (this.prefix === '?') {
            // modern xterm doesnt seem to
            // respond to any of these except ?6, 6, and 5
            switch (params[0]) {
                case 6:
                    // cursor position
                    this.send('\x1b[?' + (this.y + 1) + ';' + (this.x + 1) + 'R');
                    break;
                case 15:
                    // no printer
                    // this.send('\x1b[?11n');
                    break;
                case 25:
                    // dont support user defined keys
                    // this.send('\x1b[?21n');
                    break;
                case 26:
                    // north american keyboard
                    // this.send('\x1b[?27;1;0;0n');
                    break;
                case 53:
                    // no dec locator/mouse
                    // this.send('\x1b[?50n');
                    break;
            }
        }
    }
    /**
     * Additions
     */
    // CSI Ps @
    // Insert Ps (Blank) Character(s) (default = 1) (ICH).
    insertChars(params) {
        let param = params[0];
        if (param < 1) {
            param = 1;
        }
        const row = this.y;
        let j = this.x;
        const attr = this.eraseAttr();
        const chCodePoint = ' '.codePointAt(0);
        const { chars, attrs } = this._getRow(row);
        while (param-- && j < this.cols) {
            chars.copyWithin(j + 1, j);
            chars[j] = chCodePoint;
            attrs.copyWithin(j + 1, j);
            attrs[j] = attr;
            j++;
        }
    }
    // CSI Ps E
    // Cursor Next Line Ps Times (default = 1) (CNL).
    // same as CSI Ps B ?
    cursorNextLine(params) {
        let param = params[0];
        if (param < 1) {
            param = 1;
        }
        if (this.y + param >= this.rows) {
            this._setCursorY(this.rows - 1);
        }
        else {
            this._setCursorY(this.y + param);
        }
        this.x = 0;
    }
    // CSI Ps F
    // Cursor Preceding Line Ps Times (default = 1) (CNL).
    // reuse CSI Ps A ?
    cursorPrecedingLine(params) {
        let param = params[0];
        if (param < 1) {
            param = 1;
        }
        if (this.y - param < 0) {
            this._setCursorY(0);
        }
        else {
            this._setCursorY(this.y - param);
        }
        this.x = 0;
    }
    // CSI Ps G
    // Cursor Character Absolute  [column] (default = [row,1]) (CHA).
    cursorCharAbsolute(params) {
        let param = params[0];
        if (param < 1) {
            param = 1;
        }
        this.x = param - 1;
    }
    // CSI Ps L
    // Insert Ps Line(s) (default = 1) (IL).
    insertLines(params) {
        let param = params[0];
        if (param < 1) {
            param = 1;
        }
        const row = this.y;
        const j = this.scrollBottom + 1;
        while (param--) {
            // test: echo -e '\e[44m\e[1L\e[0m'
            // blankLine(true) - xterm/linux behavior
            this._getRow(row);
            this.lines.splice(row, 0, this.blankLine(true));
            this.lines.splice(j, 1);
        }
        // this.maxRange();
        this.markRowForRefresh(this.y);
        this.markRowForRefresh(this.scrollBottom);
    }
    // CSI Ps M
    // Delete Ps Line(s) (default = 1) (DL).
    deleteLines(params) {
        let param = params[0];
        if (param < 1) {
            param = 1;
        }
        const row = this.y;
        const j = this.scrollBottom;
        while (param--) {
            // test: echo -e '\e[44m\e[1M\e[0m'
            // blankLine(true) - xterm/linux behavior
            this._getRow(j + 1);
            this.lines.splice(j + 1, 0, this.blankLine(true));
            this.lines.splice(row, 1);
        }
        // this.maxRange();
        this.markRowForRefresh(this.y);
        this.markRowForRefresh(this.scrollBottom);
    }
    // CSI Ps P
    // Delete Ps Character(s) (default = 1) (DCH).
    deleteChars(params) {
        let param = params[0];
        if (param < 1) {
            param = 1;
        }
        const row = this.y;
        const attr = this.eraseAttr();
        const chCodePoint = ' '.codePointAt(0);
        const { chars, attrs } = this.lines[row];
        while (param--) {
            chars.copyWithin(this.x, this.x + 1);
            chars[chars.length - 1] = chCodePoint;
            attrs.copyWithin(this.x, this.x + 1);
            attrs[attrs.length - 1] = attr;
        }
    }
    // CSI Ps X
    // Erase Ps Character(s) (default = 1) (ECH).
    eraseChars(params) {
        let param = params[0];
        if (param < 1) {
            param = 1;
        }
        const row = this.y;
        let j = this.x;
        const attr = this.eraseAttr();
        const spaceCodePoint = ' '.codePointAt(0);
        const { chars, attrs } = this._getRow(row);
        while (param-- && j < this.cols) {
            chars[j] = spaceCodePoint;
            attrs[j] = attr;
            j++;
        }
    }
    // CSI Pm `  Character Position Absolute
    //   [column] (default = [row,1]) (HPA).
    charPosAbsolute(params) {
        let param = params[0];
        if (param < 1) {
            param = 1;
        }
        this.x = param - 1;
        if (this.x >= this.cols) {
            this.x = this.cols - 1;
        }
    }
    // 141 61 a * HPR -
    // Horizontal Position Relative
    // reuse CSI Ps C ?
    HPositionRelative(params) {
        let param = params[0];
        if (param < 1) {
            param = 1;
        }
        this.x += param;
        if (this.x >= this.cols) {
            this.x = this.cols - 1;
        }
    }
    // CSI Ps c  Send Device Attributes (Primary DA).
    //     Ps = 0  or omitted -> request attributes from terminal.  The
    //     response depends on the decTerminalID resource setting.
    //     -> CSI ? 1 ; 2 c  (``VT100 with Advanced Video Option'')
    //     -> CSI ? 1 ; 0 c  (``VT101 with No Options'')
    //     -> CSI ? 6 c  (``VT102'')
    //     -> CSI ? 6 0 ; 1 ; 2 ; 6 ; 8 ; 9 ; 1 5 ; c  (``VT220'')
    //   The VT100-style response parameters do not mean anything by
    //   themselves.  VT220 parameters do, telling the host what fea-
    //   tures the terminal supports:
    //     Ps = 1  -> 132-columns.
    //     Ps = 2  -> Printer.
    //     Ps = 6  -> Selective erase.
    //     Ps = 8  -> User-defined keys.
    //     Ps = 9  -> National replacement character sets.
    //     Ps = 1 5  -> Technical characters.
    //     Ps = 2 2  -> ANSI color, e.g., VT525.
    //     Ps = 2 9  -> ANSI text locator (i.e., DEC Locator mode).
    // CSI > Ps c
    //   Send Device Attributes (Secondary DA).
    //     Ps = 0  or omitted -> request the terminal's identification
    //     code.  The response depends on the decTerminalID resource set-
    //     ting.  It should apply only to VT220 and up, but xterm extends
    //     this to VT100.
    //     -> CSI  > Pp ; Pv ; Pc c
    //   where Pp denotes the terminal type
    //     Pp = 0  -> ``VT100''.
    //     Pp = 1  -> ``VT220''.
    //   and Pv is the firmware version (for xterm, this was originally
    //   the XFree86 patch number, starting with 95).  In a DEC termi-
    //   nal, Pc indicates the ROM cartridge registration number and is
    //   always zero.
    // More information:
    //   xterm/charproc.c - line 2012, for more information.
    //   vim responds with ^[[?0c or ^[[?1c after the terminal's response (?)
    sendDeviceAttributes(params) {
        if (params[0] > 0) {
            return;
        }
        if (!this.prefix) {
            if (this.is('xterm') || this.is('rxvt-unicode') || this.is('screen')) {
                this.send('\x1b[?1;2c');
            }
            else if (this.is('linux')) {
                this.send('\x1b[?6c');
            }
        }
        else if (this.prefix === '>') {
            // xterm and urxvt
            // seem to spit this
            // out around ~370 times (?).
            if (this.is('xterm')) {
                this.send('\x1b[>0;276;0c');
            }
            else if (this.is('rxvt-unicode')) {
                this.send('\x1b[>85;95;0c');
            }
            else if (this.is('linux')) {
                // not supported by linux console.
                // linux console echoes parameters.
                this.send(params[0] + 'c');
            }
            else if (this.is('screen')) {
                this.send('\x1b[>83;40003;0c');
            }
        }
    }
    // CSI Pm d
    // Line Position Absolute  [row] (default = [1,column]) (VPA).
    linePosAbsolute(params) {
        let param = params[0];
        if (param < 1) {
            param = 1;
        }
        let y = param - 1 + (this.originMode ? this.scrollTop : 0);
        if (y >= this.rows) {
            y = this.rows - 1;
        }
        this._setCursorY(y);
    }
    // 145 65 e * VPR - Vertical Position Relative
    // reuse CSI Ps B ?
    VPositionRelative(params) {
        let param = params[0];
        if (param < 1) {
            param = 1;
        }
        if (this.y + param >= this.rows) {
            this._setCursorY(this.rows - 1);
        }
        else {
            this._setCursorY(this.y + param);
        }
    }
    // CSI Ps ; Ps f
    //   Horizontal and Vertical Position [row;column] (default =
    //   [1,1]) (HVP).
    HVPosition(params) {
        this.cursorPos(params);
    }
    // CSI Pm h  Set Mode (SM).
    //     Ps = 2  -> Keyboard Action Mode (AM).
    //     Ps = 4  -> Insert Mode (IRM).
    //     Ps = 1 2  -> Send/receive (SRM).
    //     Ps = 2 0  -> Automatic Newline (LNM).
    // CSI ? Pm h
    //   DEC Private Mode Set (DECSET).
    //     Ps = 1  -> Application Cursor Keys (DECCKM).
    //     Ps = 2  -> Designate USASCII for character sets G0-G3
    //     (DECANM), and set VT100 mode.
    //     Ps = 3  -> 132 Column Mode (DECCOLM).
    //     Ps = 4  -> Smooth (Slow) Scroll (DECSCLM).
    //     Ps = 5  -> Reverse Video (DECSCNM).
    //     Ps = 6  -> Origin Mode (DECOM).
    //     Ps = 7  -> Wraparound Mode (DECAWM).
    //     Ps = 8  -> Auto-repeat Keys (DECARM).
    //     Ps = 9  -> Send Mouse X & Y on button press.  See the sec-
    //     tion Mouse Tracking.
    //     Ps = 1 0  -> Show toolbar (rxvt).
    //     Ps = 1 2  -> Start Blinking Cursor (att610).
    //     Ps = 1 8  -> Print form feed (DECPFF).
    //     Ps = 1 9  -> Set print extent to full screen (DECPEX).
    //     Ps = 2 5  -> Show Cursor (DECTCEM).
    //     Ps = 3 0  -> Show scrollbar (rxvt).
    //     Ps = 3 5  -> Enable font-shifting functions (rxvt).
    //     Ps = 3 8  -> Enter Tektronix Mode (DECTEK).
    //     Ps = 4 0  -> Allow 80 -> 132 Mode.
    //     Ps = 4 1  -> more(1) fix (see curses resource).
    //     Ps = 4 2  -> Enable Nation Replacement Character sets (DECN-
    //     RCM).
    //     Ps = 4 4  -> Turn On Margin Bell.
    //     Ps = 4 5  -> Reverse-wraparound Mode.
    //     Ps = 4 6  -> Start Logging.  This is normally disabled by a
    //     compile-time option.
    //     Ps = 4 7  -> Use Alternate Screen Buffer.  (This may be dis-
    //     abled by the titeInhibit resource).
    //     Ps = 6 6  -> Application keypad (DECNKM).
    //     Ps = 6 7  -> Backarrow key sends backspace (DECBKM).
    //     Ps = 1 0 0 0  -> Send Mouse X & Y on button press and
    //     release.  See the section Mouse Tracking.
    //     Ps = 1 0 0 1  -> Use Hilite Mouse Tracking.
    //     Ps = 1 0 0 2  -> Use Cell Motion Mouse Tracking.
    //     Ps = 1 0 0 3  -> Use All Motion Mouse Tracking.
    //     Ps = 1 0 0 4  -> Send FocusIn/FocusOut events.
    //     Ps = 1 0 0 5  -> Enable Extended Mouse Mode.
    //     Ps = 1 0 1 0  -> Scroll to bottom on tty output (rxvt).
    //     Ps = 1 0 1 1  -> Scroll to bottom on key press (rxvt).
    //     Ps = 1 0 3 4  -> Interpret "meta" key, sets eighth bit.
    //     (enables the eightBitInput resource).
    //     Ps = 1 0 3 5  -> Enable special modifiers for Alt and Num-
    //     Lock keys.  (This enables the numLock resource).
    //     Ps = 1 0 3 6  -> Send ESC   when Meta modifies a key.  (This
    //     enables the metaSendsEscape resource).
    //     Ps = 1 0 3 7  -> Send DEL from the editing-keypad Delete
    //     key.
    //     Ps = 1 0 3 9  -> Send ESC  when Alt modifies a key.  (This
    //     enables the altSendsEscape resource).
    //     Ps = 1 0 4 0  -> Keep selection even if not highlighted.
    //     (This enables the keepSelection resource).
    //     Ps = 1 0 4 1  -> Use the CLIPBOARD selection.  (This enables
    //     the selectToClipboard resource).
    //     Ps = 1 0 4 2  -> Enable Urgency window manager hint when
    //     Control-G is received.  (This enables the bellIsUrgent
    //     resource).
    //     Ps = 1 0 4 3  -> Enable raising of the window when Control-G
    //     is received.  (enables the popOnBell resource).
    //     Ps = 1 0 4 7  -> Use Alternate Screen Buffer.  (This may be
    //     disabled by the titeInhibit resource).
    //     Ps = 1 0 4 8  -> Save cursor as in DECSC.  (This may be dis-
    //     abled by the titeInhibit resource).
    //     Ps = 1 0 4 9  -> Save cursor as in DECSC and use Alternate
    //     Screen Buffer, clearing it first.  (This may be disabled by
    //     the titeInhibit resource).  This combines the effects of the 1
    //     0 4 7  and 1 0 4 8  modes.  Use this with terminfo-based
    //     applications rather than the 4 7  mode.
    //     Ps = 1 0 5 0  -> Set terminfo/termcap function-key mode.
    //     Ps = 1 0 5 1  -> Set Sun function-key mode.
    //     Ps = 1 0 5 2  -> Set HP function-key mode.
    //     Ps = 1 0 5 3  -> Set SCO function-key mode.
    //     Ps = 1 0 6 0  -> Set legacy keyboard emulation (X11R6).
    //     Ps = 1 0 6 1  -> Set VT220 keyboard emulation.
    //     Ps = 2 0 0 4  -> Set bracketed paste mode.
    // Modes:
    //   http://vt100.net/docs/vt220-rm/chapter4.html
    setMode(params) {
        if (typeof params === 'object') {
            const l = params.length;
            let i = 0;
            for (; i < l; i++) {
                this.setMode(params[i]);
            }
            return;
        }
        if (this.prefix === '') {
            switch (params) {
                case 4:
                    this.insertMode = true;
                    break;
                case 20:
                    //this.convertEol = true;
                    break;
            }
        }
        else if (this.prefix === '?') {
            switch (params) {
                case 1:
                    this.applicationCursor = true;
                    break;
                case 2:
                    this.setgCharset(0, Emulator.charsets.US);
                    this.setgCharset(1, Emulator.charsets.US);
                    this.setgCharset(2, Emulator.charsets.US);
                    this.setgCharset(3, Emulator.charsets.US);
                    // set VT100 mode here
                    break;
                case 3:// 132 col mode
                    this.savedCols = this.cols;
                    this.resize({ rows: this.rows, columns: 132 });
                    break;
                case 6:
                    this.originMode = true;
                    this.x = 0;
                    this._setCursorY(this.scrollTop);
                    break;
                case 7:
                    this.wraparoundMode = true;
                    break;
                case 12:
                    // this.cursorBlink = true;
                    break;
                case 66:
                    this.log('Serial port requested application keypad.');
                    this.applicationKeypad = true;
                    break;
                case 9: // X10 Mouse
                // no release, no motion, no wheel, no modifiers.
                case 1000: // vt200 mouse
                // no motion.
                // no modifiers, except control on the wheel.
                case 1002: // button event mouse
                case 1003:// any event mouse
                    // any event - sends motion events,
                    // even if there is no button held down.
                    this.x10Mouse = params === 9;
                    this.vt200Mouse = params === 1000;
                    this.normalMouse = params > 1000;
                    this.mouseEvents = true;
                    break;
                case 1004:// send focusin/focusout events
                    // focusin: ^[[I
                    // focusout: ^[[O
                    this.sendFocus = true;
                    break;
                case 1005:// utf8 ext mode mouse
                    this.utfMouse = true;
                    // for wide terminals
                    // simply encodes large values as utf8 characters
                    break;
                case 1006:// sgr ext mode mouse
                    this.sgrMouse = true;
                    // for wide terminals
                    // does not add 32 to fields
                    // press: ^[[<b;x;yM
                    // release: ^[[<b;x;ym
                    break;
                case 1015:// urxvt ext mode mouse
                    this.urxvtMouse = true;
                    // for wide terminals
                    // numbers for fields
                    // press: ^[[b;x;yM
                    // motion: ^[[b;x;yT
                    break;
                case 25:// show cursor
                    this.cursorHidden = false;
                    break;
                case 1049: // alt screen buffer cursor
                //this.saveCursor();
                // FALL-THROUGH
                case 47: // alt screen buffer
                case 1047:// alt screen buffer
                    if (!this.normal) {
                        const normal = {
                            cols: this.cols,
                            rows: this.rows,
                            lines: this.lines,
                            x: this.x,
                            y: this.y,
                            scrollTop: this.scrollTop,
                            scrollBottom: this.scrollBottom,
                            tabs: this.tabs,
                        };
                        // Preserve these variables during the reset().
                        const previousCharset = this.charset;
                        const previousGlevel = this.glevel;
                        const previousCharsets = this.charsets;
                        this._fullReset();
                        this.charset = previousCharset;
                        this.glevel = previousGlevel;
                        this.charsets = previousCharsets;
                        // Materialize the same number of rows as the normal lines array to ensure that
                        // the new screen buffer is rendered over all of the previous screen.
                        while (this.lines.length < normal.lines.length) {
                            this._getRow(this.lines.length);
                        }
                        this.normal = normal;
                    }
                    break;
            }
        }
    }
    // CSI Pm l  Reset Mode (RM).
    //     Ps = 2  -> Keyboard Action Mode (AM).
    //     Ps = 4  -> Replace Mode (IRM).
    //     Ps = 1 2  -> Send/receive (SRM).
    //     Ps = 2 0  -> Normal Linefeed (LNM).
    // CSI ? Pm l
    //   DEC Private Mode Reset (DECRST).
    //     Ps = 1  -> Normal Cursor Keys (DECCKM).
    //     Ps = 2  -> Designate VT52 mode (DECANM).
    //     Ps = 3  -> 80 Column Mode (DECCOLM).
    //     Ps = 4  -> Jump (Fast) Scroll (DECSCLM).
    //     Ps = 5  -> Normal Video (DECSCNM).
    //     Ps = 6  -> Normal Cursor Mode (DECOM).
    //     Ps = 7  -> No Wraparound Mode (DECAWM).
    //     Ps = 8  -> No Auto-repeat Keys (DECARM).
    //     Ps = 9  -> Don't send Mouse X & Y on button press.
    //     Ps = 1 0  -> Hide toolbar (rxvt).
    //     Ps = 1 2  -> Stop Blinking Cursor (att610).
    //     Ps = 1 8  -> Don't print form feed (DECPFF).
    //     Ps = 1 9  -> Limit print to scrolling region (DECPEX).
    //     Ps = 2 5  -> Hide Cursor (DECTCEM).
    //     Ps = 3 0  -> Don't show scrollbar (rxvt).
    //     Ps = 3 5  -> Disable font-shifting functions (rxvt).
    //     Ps = 4 0  -> Disallow 80 -> 132 Mode.
    //     Ps = 4 1  -> No more(1) fix (see curses resource).
    //     Ps = 4 2  -> Disable Nation Replacement Character sets (DEC-
    //     NRCM).
    //     Ps = 4 4  -> Turn Off Margin Bell.
    //     Ps = 4 5  -> No Reverse-wraparound Mode.
    //     Ps = 4 6  -> Stop Logging.  (This is normally disabled by a
    //     compile-time option).
    //     Ps = 4 7  -> Use Normal Screen Buffer.
    //     Ps = 6 6  -> Numeric keypad (DECNKM).
    //     Ps = 6 7  -> Backarrow key sends delete (DECBKM).
    //     Ps = 1 0 0 0  -> Don't send Mouse X & Y on button press and
    //     release.  See the section Mouse Tracking.
    //     Ps = 1 0 0 1  -> Don't use Hilite Mouse Tracking.
    //     Ps = 1 0 0 2  -> Don't use Cell Motion Mouse Tracking.
    //     Ps = 1 0 0 3  -> Don't use All Motion Mouse Tracking.
    //     Ps = 1 0 0 4  -> Don't send FocusIn/FocusOut events.
    //     Ps = 1 0 0 5  -> Disable Extended Mouse Mode.
    //     Ps = 1 0 1 0  -> Don't scroll to bottom on tty output
    //     (rxvt).
    //     Ps = 1 0 1 1  -> Don't scroll to bottom on key press (rxvt).
    //     Ps = 1 0 3 4  -> Don't interpret "meta" key.  (This disables
    //     the eightBitInput resource).
    //     Ps = 1 0 3 5  -> Disable special modifiers for Alt and Num-
    //     Lock keys.  (This disables the numLock resource).
    //     Ps = 1 0 3 6  -> Don't send ESC  when Meta modifies a key.
    //     (This disables the metaSendsEscape resource).
    //     Ps = 1 0 3 7  -> Send VT220 Remove from the editing-keypad
    //     Delete key.
    //     Ps = 1 0 3 9  -> Don't send ESC  when Alt modifies a key.
    //     (This disables the altSendsEscape resource).
    //     Ps = 1 0 4 0  -> Do not keep selection when not highlighted.
    //     (This disables the keepSelection resource).
    //     Ps = 1 0 4 1  -> Use the PRIMARY selection.  (This disables
    //     the selectToClipboard resource).
    //     Ps = 1 0 4 2  -> Disable Urgency window manager hint when
    //     Control-G is received.  (This disables the bellIsUrgent
    //     resource).
    //     Ps = 1 0 4 3  -> Disable raising of the window when Control-
    //     G is received.  (This disables the popOnBell resource).
    //     Ps = 1 0 4 7  -> Use Normal Screen Buffer, clearing screen
    //     first if in the Alternate Screen.  (This may be disabled by
    //     the titeInhibit resource).
    //     Ps = 1 0 4 8  -> Restore cursor as in DECRC.  (This may be
    //     disabled by the titeInhibit resource).
    //     Ps = 1 0 4 9  -> Use Normal Screen Buffer and restore cursor
    //     as in DECRC.  (This may be disabled by the titeInhibit
    //     resource).  This combines the effects of the 1 0 4 7  and 1 0
    //     4 8  modes.  Use this with terminfo-based applications rather
    //     than the 4 7  mode.
    //     Ps = 1 0 5 0  -> Reset terminfo/termcap function-key mode.
    //     Ps = 1 0 5 1  -> Reset Sun function-key mode.
    //     Ps = 1 0 5 2  -> Reset HP function-key mode.
    //     Ps = 1 0 5 3  -> Reset SCO function-key mode.
    //     Ps = 1 0 6 0  -> Reset legacy keyboard emulation (X11R6).
    //     Ps = 1 0 6 1  -> Reset keyboard emulation to Sun/PC style.
    //     Ps = 2 0 0 4  -> Reset bracketed paste mode.
    resetMode(params) {
        if (typeof params === 'object') {
            const l = params.length;
            for (let i = 0; i < l; i++) {
                this.resetMode(params[i]);
            }
            return;
        }
        if (!this.prefix) {
            switch (params) {
                case 4:
                    this.insertMode = false;
                    break;
                case 20:
                    //this.convertEol = false;
                    break;
            }
        }
        else if (this.prefix === '?') {
            switch (params) {
                case 1:
                    this.applicationCursor = false;
                    break;
                // 80 Column Mode (DECCOLM).
                case 3:
                    this.fillScreen();
                    this.x = 0;
                    this._setCursorY(0);
                    if (this.cols === 132 && this.savedCols) {
                        this.resize({ rows: this.rows, columns: this.savedCols });
                    }
                    break;
                case 6:
                    this.originMode = false;
                    this.x = 0;
                    this._setCursorY(0);
                    break;
                case 7:
                    this.wraparoundMode = false;
                    break;
                case 12:
                    // this.cursorBlink = false;
                    break;
                case 66:
                    this.log('Switching back to normal keypad.');
                    this.applicationKeypad = false;
                    break;
                case 9: // X10 Mouse
                case 1000: // vt200 mouse
                case 1002: // button event mouse
                case 1003:// any event mouse
                    this.x10Mouse = false;
                    this.vt200Mouse = false;
                    this.normalMouse = false;
                    this.mouseEvents = false;
                    break;
                case 1004:// send focusin/focusout events
                    this.sendFocus = false;
                    break;
                case 1005:// utf8 ext mode mouse
                    this.utfMouse = false;
                    break;
                case 1006:// sgr ext mode mouse
                    this.sgrMouse = false;
                    break;
                case 1015:// urxvt ext mode mouse
                    this.urxvtMouse = false;
                    break;
                case 25:// hide cursor
                    this.cursorHidden = true;
                    break;
                case 1049: // alt screen buffer cursor
                // FALL-THROUGH
                case 47: // normal screen buffer
                case 1047:// normal screen buffer - clearing it first
                    if (this.normal) {
                        const currentcols = this.cols;
                        const currentrows = this.rows;
                        this.lines = this.normal.lines;
                        this.cols = this.normal.cols;
                        this.rows = this.normal.rows;
                        this.x = this.normal.x;
                        this._setCursorY(this.normal.y);
                        this.scrollTop = this.normal.scrollTop;
                        this.scrollBottom = this.normal.scrollBottom;
                        this.tabs = this.normal.tabs;
                        this.normal = null;
                        // if (params === 1049) {
                        //   this.x = this.savedX;
                        //   this._setCursorY(this.savedY);
                        // }
                        this.resize({ rows: currentrows, columns: currentcols });
                        this.markRowRangeForRefresh(0, this.rows - 1);
                    }
                    break;
            }
        }
    }
    // CSI Ps ; Ps r
    //   Set Scrolling Region [top;bottom] (default = full size of win-
    //   dow) (DECSTBM).
    // CSI ? Pm r
    setScrollRegion(params) {
        if (this.prefix === '') {
            const top = (params[0] || 1) - 1;
            const bottom = (params[1] || this.rows) - 1;
            if (!(top >= 0 && bottom < this.rows && top < bottom)) {
                return;
            }
            this.scrollTop = top;
            this.scrollBottom = bottom;
            this.x = 0;
            this._setCursorY(this.originMode ? this.scrollTop : 0);
        }
    }
    // CSI s
    //   Save cursor (ANSI.SYS).
    saveCursor() {
        this.savedX = this.x;
        this.savedY = this.y;
        this.savedCharset = this.charset;
        this.savedCurAttr = this.curAttr;
    }
    // CSI u
    //   Restore cursor (ANSI.SYS).
    restoreCursor() {
        this.x = this.savedX;
        this._setCursorY(this.savedY);
        this.charset = this.savedCharset;
        this.curAttr = this.savedCurAttr;
    }
    /**
     * Lesser Used
     */
    // CSI Ps I
    //   Cursor Forward Tabulation Ps tab stops (default = 1) (CHT).
    cursorForwardTab(params) {
        let param = params[0] || 1;
        while (param--) {
            this.x = this.nextStop();
        }
    }
    // CSI Ps S  Scroll up Ps lines (default = 1) (SU).
    scrollUp(params) {
        let param = params[0] || 1;
        while (param--) {
            this.lines.splice(this.scrollTop, 1);
            this.lines.splice(this.scrollBottom, 0, this.blankLine());
        }
        // this.maxRange();
        this.markRowForRefresh(this.scrollTop);
        this.markRowForRefresh(this.scrollBottom);
    }
    // CSI Ps T  Scroll down Ps lines (default = 1) (SD).
    scrollDown(params) {
        let param = params[0] || 1;
        while (param--) {
            this.lines.splice(this.scrollBottom, 1);
            this.lines.splice(this.scrollTop, 0, this.blankLine());
        }
        // this.maxRange();
        this.markRowForRefresh(this.scrollTop);
        this.markRowForRefresh(this.scrollBottom);
    }
    // CSI Ps ; Ps ; Ps ; Ps ; Ps T
    //   Initiate highlight mouse tracking.  Parameters are
    //   [func;startx;starty;firstrow;lastrow].  See the section Mouse
    //   Tracking.
    initMouseTracking(params) {
        // Relevant: DECSET 1001
    }
    // CSI > Ps; Ps T
    //   Reset one or more features of the title modes to the default
    //   value.  Normally, "reset" disables the feature.  It is possi-
    //   ble to disable the ability to reset features by compiling a
    //   different default for the title modes into xterm.
    //     Ps = 0  -> Do not set window/icon labels using hexadecimal.
    //     Ps = 1  -> Do not query window/icon labels using hexadeci-
    //     mal.
    //     Ps = 2  -> Do not set window/icon labels using UTF-8.
    //     Ps = 3  -> Do not query window/icon labels using UTF-8.
    //   (See discussion of "Title Modes").
    esetTitleModes(params) {
    }
    // CSI Ps Z  Cursor Backward Tabulation Ps tab stops (default = 1) (CBT).
    cursorBackwardTab(params) {
        let param = params[0] || 1;
        while (param--) {
            this.x = this.prevStop();
        }
    }
    // CSI Ps b  Repeat the preceding graphic character Ps times (REP).
    repeatPrecedingCharacter(params) {
        let param = params[0] || 1;
        const { chars, attrs } = this._getRow(this.y);
        const attr = this.x == 0 ? Emulator.defAttr : attrs[this.x - 1];
        const chCodePoint = this.x === 0 ? ' '.codePointAt(0) : chars[this.x - 1];
        while (param--) {
            chars[this.x] = chCodePoint;
            attrs[this.x] = attr;
            this.x++;
        }
    }
    // CSI Ps g  Tab Clear (TBC).
    //     Ps = 0  -> Clear Current Column (default).
    //     Ps = 3  -> Clear All.
    // Potentially:
    //   Ps = 2  -> Clear Stops on Line.
    //   http://vt100.net/annarbor/aaa-ug/section6.html
    tabClear(params) {
        let param = params[0];
        if (param <= 0) {
            delete this.tabs[this.x];
        }
        else if (param === 3) {
            this.tabs = {};
        }
    }
    // CSI Pm i  Media Copy (MC).
    //     Ps = 0  -> Print screen (default).
    //     Ps = 4  -> Turn off printer controller mode.
    //     Ps = 5  -> Turn on printer controller mode.
    // CSI ? Pm i
    //   Media Copy (MC, DEC-specific).
    //     Ps = 1  -> Print line containing cursor.
    //     Ps = 4  -> Turn off autoprint mode.
    //     Ps = 5  -> Turn on autoprint mode.
    //     Ps = 1  0  -> Print composed display, ignores DECPEX.
    //     Ps = 1  1  -> Print all pages.
    mediaCopy(params) {
    }
    // CSI > Ps; Ps m
    //   Set or reset resource-values used by xterm to decide whether
    //   to construct escape sequences holding information about the
    //   modifiers pressed with a given key.  The first parameter iden-
    //   tifies the resource to set/reset.  The second parameter is the
    //   value to assign to the resource.  If the second parameter is
    //   omitted, the resource is reset to its initial value.
    //     Ps = 1  -> modifyCursorKeys.
    //     Ps = 2  -> modifyFunctionKeys.
    //     Ps = 4  -> modifyOtherKeys.
    //   If no parameters are given, all resources are reset to their
    //   initial values.
    setResources(params) {
    }
    // CSI > Ps n
    //   Disable modifiers which may be enabled via the CSI > Ps; Ps m
    //   sequence.  This corresponds to a resource value of "-1", which
    //   cannot be set with the other sequence.  The parameter identi-
    //   fies the resource to be disabled:
    //     Ps = 1  -> modifyCursorKeys.
    //     Ps = 2  -> modifyFunctionKeys.
    //     Ps = 4  -> modifyOtherKeys.
    //   If the parameter is omitted, modifyFunctionKeys is disabled.
    //   When modifyFunctionKeys is disabled, xterm uses the modifier
    //   keys to make an extended sequence of functions rather than
    //   adding a parameter to each function key to denote the modi-
    //   fiers.
    disableModifiers(params) {
    }
    // CSI > Ps p
    //   Set resource value pointerMode.  This is used by xterm to
    //   decide whether to hide the pointer cursor as the user types.
    //   Valid values for the parameter:
    //     Ps = 0  -> never hide the pointer.
    //     Ps = 1  -> hide if the mouse tracking mode is not enabled.
    //     Ps = 2  -> always hide the pointer.  If no parameter is
    //     given, xterm uses the default, which is 1 .
    setPointerMode(params) {
    }
    // CSI ! p   Soft terminal reset (DECSTR).
    // http://vt100.net/docs/vt220-rm/table4-10.html
    softReset(params) {
        this.cursorHidden = false;
        this.insertMode = false;
        this.originMode = false;
        this.wraparoundMode = false; // autowrap
        this.applicationKeypad = false; // ?
        this.applicationCursor = false;
        this.scrollTop = 0;
        this.scrollBottom = this.rows - 1;
        this.curAttr = Emulator.defAttr;
        this.x = 0;
        this._setCursorY(0);
        this.charset = null;
        this.glevel = 0; // ??
        this.charsets = [null]; // ??
    }
    // CSI Ps$ p
    //   Request ANSI mode (DECRQM).  For VT300 and up, reply is
    //     CSI Ps; Pm$ y
    //   where Ps is the mode number as in RM, and Pm is the mode
    //   value:
    //     0 - not recognized
    //     1 - set
    //     2 - reset
    //     3 - permanently set
    //     4 - permanently reset
    requestAnsiMode(params) {
    }
    // CSI ? Ps$ p
    //   Request DEC private mode (DECRQM).  For VT300 and up, reply is
    //     CSI ? Ps; Pm$ p
    //   where Ps is the mode number as in DECSET, Pm is the mode value
    //   as in the ANSI DECRQM.
    requestPrivateMode(params) {
    }
    // CSI Ps ; Ps " p
    //   Set conformance level (DECSCL).  Valid values for the first
    //   parameter:
    //     Ps = 6 1  -> VT100.
    //     Ps = 6 2  -> VT200.
    //     Ps = 6 3  -> VT300.
    //   Valid values for the second parameter:
    //     Ps = 0  -> 8-bit controls.
    //     Ps = 1  -> 7-bit controls (always set for VT100).
    //     Ps = 2  -> 8-bit controls.
    setConformanceLevel(params) {
    }
    // CSI Ps q  Load LEDs (DECLL).
    //     Ps = 0  -> Clear all LEDS (default).
    //     Ps = 1  -> Light Num Lock.
    //     Ps = 2  -> Light Caps Lock.
    //     Ps = 3  -> Light Scroll Lock.
    //     Ps = 2  1  -> Extinguish Num Lock.
    //     Ps = 2  2  -> Extinguish Caps Lock.
    //     Ps = 2  3  -> Extinguish Scroll Lock.
    loadLEDs(params) {
    }
    // CSI Ps SP q
    //   Set cursor style (DECSCUSR, VT520).
    //     Ps = 0  -> blinking block.
    //     Ps = 1  -> blinking block (default).
    //     Ps = 2  -> steady block.
    //     Ps = 3  -> blinking underline.
    //     Ps = 4  -> steady underline.
    setCursorStyle(params) {
    }
    // CSI Ps " q
    //   Select character protection attribute (DECSCA).  Valid values
    //   for the parameter:
    //     Ps = 0  -> DECSED and DECSEL can erase (default).
    //     Ps = 1  -> DECSED and DECSEL cannot erase.
    //     Ps = 2  -> DECSED and DECSEL can erase.
    setCharProtectionAttr(params) {
    }
    // CSI ? Pm r
    //   Restore DEC Private Mode Values.  The value of Ps previously
    //   saved is restored.  Ps values are the same as for DECSET.
    restorePrivateValues(params) {
    }
    // CSI Pt; Pl; Pb; Pr; Ps$ r
    //   Change Attributes in Rectangular Area (DECCARA), VT400 and up.
    //     Pt; Pl; Pb; Pr denotes the rectangle.
    //     Ps denotes the SGR attributes to change: 0, 1, 4, 5, 7.
    // NOTE: xterm doesn't enable this code by default.
    setAttrInRectangle(params) {
        let t = params[0];
        const l = params[1];
        const b = params[2];
        const r = params[3];
        const attr = params[4];
        for (; t < b + 1; t++) {
            const line = this._getRow(t);
            const attrs = line.attrs;
            for (let i = l; i < r; i++) {
                attrs[i] = attr;
            }
        }
        // this.maxRange();
        this.markRowForRefresh(params[0]);
        this.markRowForRefresh(params[2]);
    }
    // CSI ? Pm s
    //   Save DEC Private Mode Values.  Ps values are the same as for
    //   DECSET.
    savePrivateValues(params) {
    }
    // CSI Ps ; Ps ; Ps t
    //   Window manipulation (from dtterm, as well as extensions).
    //   These controls may be disabled using the allowWindowOps
    //   resource.  Valid values for the first (and any additional
    //   parameters) are:
    //     Ps = 1  -> De-iconify window.
    //     Ps = 2  -> Iconify window.
    //     Ps = 3  ;  x ;  y -> Move window to [x, y].
    //     Ps = 4  ;  height ;  width -> Resize the xterm window to
    //     height and width in pixels.
    //     Ps = 5  -> Raise the xterm window to the front of the stack-
    //     ing order.
    //     Ps = 6  -> Lower the xterm window to the bottom of the
    //     stacking order.
    //     Ps = 7  -> Refresh the xterm window.
    //     Ps = 8  ;  height ;  width -> Resize the text area to
    //     [height;width] in characters.
    //     Ps = 9  ;  0  -> Restore maximized window.
    //     Ps = 9  ;  1  -> Maximize window (i.e., resize to screen
    //     size).
    //     Ps = 1 0  ;  0  -> Undo full-screen mode.
    //     Ps = 1 0  ;  1  -> Change to full-screen.
    //     Ps = 1 1  -> Report xterm window state.  If the xterm window
    //     is open (non-iconified), it returns CSI 1 t .  If the xterm
    //     window is iconified, it returns CSI 2 t .
    //     Ps = 1 3  -> Report xterm window position.  Result is CSI 3
    //     ; x ; y t
    //     Ps = 1 4  -> Report xterm window in pixels.  Result is CSI
    //     4  ;  height ;  width t
    //     Ps = 1 8  -> Report the size of the text area in characters.
    //     Result is CSI  8  ;  height ;  width t
    //     Ps = 1 9  -> Report the size of the screen in characters.
    //     Result is CSI  9  ;  height ;  width t
    //     Ps = 2 0  -> Report xterm window's icon label.  Result is
    //     OSC  L  label ST
    //     Ps = 2 1  -> Report xterm window's title.  Result is OSC  l
    //     label ST
    //     Ps = 2 2  ;  0  -> Save xterm icon and window title on
    //     stack.
    //     Ps = 2 2  ;  1  -> Save xterm icon title on stack.
    //     Ps = 2 2  ;  2  -> Save xterm window title on stack.
    //     Ps = 2 3  ;  0  -> Restore xterm icon and window title from
    //     stack.
    //     Ps = 2 3  ;  1  -> Restore xterm icon title from stack.
    //     Ps = 2 3  ;  2  -> Restore xterm window title from stack.
    //     Ps >= 2 4  -> Resize to Ps lines (DECSLPP).
    manipulateWindow(params) {
    }
    // CSI Pt; Pl; Pb; Pr; Ps$ t
    //   Reverse Attributes in Rectangular Area (DECRARA), VT400 and
    //   up.
    //     Pt; Pl; Pb; Pr denotes the rectangle.
    //     Ps denotes the attributes to reverse, i.e.,  1, 4, 5, 7.
    // NOTE: xterm doesn't enable this code by default.
    reverseAttrInRectangle(params) {
    }
    // CSI > Ps; Ps t
    //   Set one or more features of the title modes.  Each parameter
    //   enables a single feature.
    //     Ps = 0  -> Set window/icon labels using hexadecimal.
    //     Ps = 1  -> Query window/icon labels using hexadecimal.
    //     Ps = 2  -> Set window/icon labels using UTF-8.
    //     Ps = 3  -> Query window/icon labels using UTF-8.  (See dis-
    //     cussion of "Title Modes")
    setTitleModeFeature(params) {
    }
    // CSI Ps SP t
    //   Set warning-bell volume (DECSWBV, VT520).
    //     Ps = 0  or 1  -> off.
    //     Ps = 2 , 3  or 4  -> low.
    //     Ps = 5 , 6 , 7 , or 8  -> high.
    setWarningBellVolume(params) {
    }
    // CSI Ps SP u
    //   Set margin-bell volume (DECSMBV, VT520).
    //     Ps = 1  -> off.
    //     Ps = 2 , 3  or 4  -> low.
    //     Ps = 0 , 5 , 6 , 7 , or 8  -> high.
    setMarginBellVolume(params) {
    }
    // CSI Pt; Pl; Pb; Pr; Pp; Pt; Pl; Pp$ v
    //   Copy Rectangular Area (DECCRA, VT400 and up).
    //     Pt; Pl; Pb; Pr denotes the rectangle.
    //     Pp denotes the source page.
    //     Pt; Pl denotes the target location.
    //     Pp denotes the target page.
    // NOTE: xterm doesn't enable this code by default.
    copyRectangle(params) {
    }
    // CSI Pt ; Pl ; Pb ; Pr ' w
    //   Enable Filter Rectangle (DECEFR), VT420 and up.
    //   Parameters are [top;left;bottom;right].
    //   Defines the coordinates of a filter rectangle and activates
    //   it.  Anytime the locator is detected outside of the filter
    //   rectangle, an outside rectangle event is generated and the
    //   rectangle is disabled.  Filter rectangles are always treated
    //   as "one-shot" events.  Any parameters that are omitted default
    //   to the current locator position.  If all parameters are omit-
    //   ted, any locator motion will be reported.  DECELR always can-
    //   cels any prevous rectangle definition.
    enableFilterRectangle(params) {
    }
    // CSI Ps x  Request Terminal Parameters (DECREQTPARM).
    //   if Ps is a "0" (default) or "1", and xterm is emulating VT100,
    //   the control sequence elicits a response of the same form whose
    //   parameters describe the terminal:
    //     Ps -> the given Ps incremented by 2.
    //     Pn = 1  <- no parity.
    //     Pn = 1  <- eight bits.
    //     Pn = 1  <- 2  8  transmit 38.4k baud.
    //     Pn = 1  <- 2  8  receive 38.4k baud.
    //     Pn = 1  <- clock multiplier.
    //     Pn = 0  <- STP flags.
    requestParameters(params) {
    }
    // CSI Ps x  Select Attribute Change Extent (DECSACE).
    //     Ps = 0  -> from start to end position, wrapped.
    //     Ps = 1  -> from start to end position, wrapped.
    //     Ps = 2  -> rectangle (exact).
    selectChangeExtent(params) {
    }
    // CSI Pc; Pt; Pl; Pb; Pr$ x
    //   Fill Rectangular Area (DECFRA), VT420 and up.
    //     Pc is the character to use.
    //     Pt; Pl; Pb; Pr denotes the rectangle.
    // NOTE: xterm doesn't enable this code by default.
    fillRectangle(params) {
        const ch = params[0];
        let t = params[1];
        const l = params[2];
        const b = params[3];
        const r = params[4];
        for (; t < b + 1; t++) {
            const chars = this._getRow(t).chars;
            for (let i = l; i < r; i++) {
                chars[i] = ch;
            }
        }
        // this.maxRange();
        this.markRowForRefresh(params[1]);
        this.markRowForRefresh(params[3]);
    }
    // CSI Ps ; Pu ' z
    //   Enable Locator Reporting (DECELR).
    //   Valid values for the first parameter:
    //     Ps = 0  -> Locator disabled (default).
    //     Ps = 1  -> Locator enabled.
    //     Ps = 2  -> Locator enabled for one report, then disabled.
    //   The second parameter specifies the coordinate unit for locator
    //   reports.
    //   Valid values for the second parameter:
    //     Pu = 0  <- or omitted -> default to character cells.
    //     Pu = 1  <- device physical pixels.
    //     Pu = 2  <- character cells.
    enableLocatorReporting(params) {
        //  var val = params[0] > 0;
        //this.mouseEvents = val;
        //this.decLocator = val;
    }
    // CSI Pt; Pl; Pb; Pr$ z
    //   Erase Rectangular Area (DECERA), VT400 and up.
    //     Pt; Pl; Pb; Pr denotes the rectangle.
    // NOTE: xterm doesn't enable this code by default.
    eraseRectangle(params) {
        let t = params[0];
        const l = params[1];
        const b = params[2];
        const r = params[3];
        const attr = this.eraseAttr();
        const chCodePoint = ' '.codePointAt(0);
        for (; t < b + 1; t++) {
            const { chars, attrs } = this._getRow(t);
            for (let i = l; i < r; i++) {
                chars[i] = chCodePoint;
                attrs[i] = attr;
            }
        }
        // this.maxRange();
        this.markRowForRefresh(params[0]);
        this.markRowForRefresh(params[2]);
    }
    // CSI Pm ' {
    //   Select Locator Events (DECSLE).
    //   Valid values for the first (and any additional parameters)
    //   are:
    //     Ps = 0  -> only respond to explicit host requests (DECRQLP).
    //                (This is default).  It also cancels any filter
    //   rectangle.
    //     Ps = 1  -> report button down transitions.
    //     Ps = 2  -> do not report button down transitions.
    //     Ps = 3  -> report button up transitions.
    //     Ps = 4  -> do not report button up transitions.
    setLocatorEvents(params) {
    }
    // CSI Pt; Pl; Pb; Pr$ {
    //   Selective Erase Rectangular Area (DECSERA), VT400 and up.
    //     Pt; Pl; Pb; Pr denotes the rectangle.
    selectiveEraseRectangle(params) {
    }
    // CSI Ps ' |
    //   Request Locator Position (DECRQLP).
    //   Valid values for the parameter are:
    //     Ps = 0 , 1 or omitted -> transmit a single DECLRP locator
    //     report.
    //   If Locator Reporting has been enabled by a DECELR, xterm will
    //   respond with a DECLRP Locator Report.  This report is also
    //   generated on button up and down events if they have been
    //   enabled with a DECSLE, or when the locator is detected outside
    //   of a filter rectangle, if filter rectangles have been enabled
    //   with a DECEFR.
    //     -> CSI Pe ; Pb ; Pr ; Pc ; Pp &  w
    //   Parameters are [event;button;row;column;page].
    //   Valid values for the event:
    //     Pe = 0  -> locator unavailable - no other parameters sent.
    //     Pe = 1  -> request - xterm received a DECRQLP.
    //     Pe = 2  -> left button down.
    //     Pe = 3  -> left button up.
    //     Pe = 4  -> middle button down.
    //     Pe = 5  -> middle button up.
    //     Pe = 6  -> right button down.
    //     Pe = 7  -> right button up.
    //     Pe = 8  -> M4 button down.
    //     Pe = 9  -> M4 button up.
    //     Pe = 1 0  -> locator outside filter rectangle.
    //   ``button'' parameter is a bitmask indicating which buttons are
    //     pressed:
    //     Pb = 0  <- no buttons down.
    //     Pb & 1  <- right button down.
    //     Pb & 2  <- middle button down.
    //     Pb & 4  <- left button down.
    //     Pb & 8  <- M4 button down.
    //   ``row'' and ``column'' parameters are the coordinates of the
    //     locator position in the xterm window, encoded as ASCII deci-
    //     mal.
    //   The ``page'' parameter is not used by xterm, and will be omit-
    //   ted.
    requestLocatorPosition(params) {
    }
    // CSI P m SP }
    // Insert P s Column(s) (default = 1) (DECIC), VT420 and up.
    // NOTE: xterm doesn't enable this code by default.
    insertColumns(params) {
        let param = params[0];
        const l = this.rows;
        const chCodePoint = ' '.codePointAt(0);
        const attr = this.eraseAttr();
        while (param--) {
            for (let i = 0; i < l; i++) {
                const { chars, attrs } = this._getRow(i);
                chars.copyWithin(this.x + 2, this.x + 1);
                chars[this.x + 1] = chCodePoint;
                attrs.copyWithin(this.x + 2, this.x + 1);
                attrs[this.x + 1] = attr;
            }
        }
        this.markAllRowsForRefresh();
    }
    // CSI P m SP ~
    // Delete P s Column(s) (default = 1) (DECDC), VT420 and up
    // NOTE: xterm doesn't enable this code by default.
    deleteColumns(params) {
        let param = params[0];
        const l = this.rows;
        const chCodePoint = ' '.codePointAt(0);
        const attr = this.eraseAttr();
        while (param--) {
            for (let i = 0; i < l; i++) {
                const { chars, attrs } = this._getRow(i);
                chars.copyWithin(this.x, this.x + 1);
                chars[chars.length - 1] = chCodePoint;
                attrs.copyWithin(this.x, this.x + 1);
                attrs[attrs.length - 1] = attr;
            }
        }
        this.markAllRowsForRefresh();
    }
    /*************************************************************************/
    /**
     * EventEmitter
     */
    // Events
    addRenderEventListener(eventHandler) {
        this.addListener(RENDER_EVENT, eventHandler);
    }
    removeRenderEventListener(eventHandler) {
        this.removeListener(RENDER_EVENT, eventHandler);
    }
    addBellEventListener(eventHandler) {
        this.addListener(BELL_EVENT, eventHandler);
    }
    addDataEventListener(eventHandler) {
        this.addListener(DATA_EVENT, eventHandler);
    }
    addTitleChangeEventListener(eventHandler) {
        this.addListener(TITLE_EVENT, eventHandler);
    }
    addApplicationModeStartEventListener(eventHandler) {
        this.addListener(APPLICATIONMODESTART_EVENT, eventHandler);
    }
    addApplicationModeDataEventListener(eventHandler) {
        this.addListener(APPLICATIONMODEDATA_EVENT, eventHandler);
    }
    addApplicationModeEndEventListener(eventHandler) {
        this.addListener(APPLICATIONMODEEND_EVENT, eventHandler);
    }
    addWriteBufferSizeEventListener(eventHandler) {
        this.addListener(WRITEBUFFERSIZE_EVENT, eventHandler);
    }
    addListener(type, listener) {
        this._events[type] = this._events[type] || [];
        this._events[type].push(listener);
    }
    on(type, listener) {
        this.addListener(type, listener);
    }
    removeListener(type, listener) {
        if (!this._events[type])
            return;
        var obj = this._events[type];
        var i = obj.length;
        while (i--) {
            if (obj[i] === listener /*  || obj[i].listener === listener */) {
                obj.splice(i, 1);
                return;
            }
        }
    }
    _emit(type, ...args) {
        if (!this._events[type])
            return;
        var obj = this._events[type];
        var l = obj.length;
        var i = 0;
        for (; i < l; i++) {
            obj[i].apply(this, args);
        }
    }
    dumpLines() {
        for (let y = 0; y < this.lines.length; y++) {
            this.log("" + y + ": " + this.getLineText(y));
        }
    }
}
Emulator.defAttr = packCharAttr(0, 257, 256); // Default character style
/**
 * Colors
 *
 * This palette is only used when matching 24bit colours to the user's CSS based colours.
 * (We assume that this default palette is close enough to the actual one being used.)
 */
// Colors 0-15
Emulator.xtermColors = [
    // dark:
    '#000000',
    '#cd0000',
    '#00cd00',
    '#cdcd00',
    '#0000ee',
    '#cd00cd',
    '#00cdcd',
    '#e5e5e5',
    // bright:
    '#7f7f7f',
    '#ff0000',
    '#00ff00',
    '#ffff00',
    '#5c5cff',
    '#ff00ff',
    '#00ffff',
    '#ffffff' // white
];
// Colors 0-15 + 16-255
// Much thanks to TooTallNate for writing this.
Emulator.colors = (function () {
    var colors = Emulator.xtermColors.slice();
    var r = [0x00, 0x5f, 0x87, 0xaf, 0xd7, 0xff];
    var i;
    // 16-231
    i = 0;
    for (; i < 216; i++) {
        out(r[(i / 36) % 6 | 0], r[(i / 6) % 6 | 0], r[i % 6]);
    }
    // 232-255 (grey)
    i = 0;
    for (; i < 24; i++) {
        const v = 8 + i * 10;
        out(v, v, v);
    }
    function out(r, g, b) {
        colors.push('#' + hex(r) + hex(g) + hex(b));
    }
    function hex(c) {
        c = c.toString(16);
        return c.length < 2 ? '0' + c : c;
    }
    // Default BG/FG
    colors[256] = '#000000';
    colors[257] = '#f0f0f0';
    return colors;
})();
Emulator.vcolors = (function () {
    var out = [];
    var colors = Emulator.colors;
    var i = 0;
    var color;
    for (; i < 256; i++) {
        color = parseInt(colors[i].substring(1), 16);
        out.push([
            (color >> 16) & 0xff,
            (color >> 8) & 0xff,
            color & 0xff
        ]);
    }
    return out;
})();
/**
 * Character Sets
 */
Emulator.charsets = {
    // DEC Special Character and Line Drawing Set.
    // http://vt100.net/docs/vt102-ug/table5-13.html
    // A lot of curses apps use this if they see TERM=xterm.
    // testing: echo -e '\e(0a\e(B'
    // The xterm output sometimes seems to conflict with the
    // reference above. xterm seems in line with the reference
    // when running vttest however.
    // The table below now uses xterm's output from vttest.
    SCLD: {
        '`': '\u25c6',
        'a': '\u2592',
        'b': '\u0009',
        'c': '\u000c',
        'd': '\u000d',
        'e': '\u000a',
        'f': '\u00b0',
        'g': '\u00b1',
        'h': '\u2424',
        'i': '\u000b',
        'j': '\u2518',
        'k': '\u2510',
        'l': '\u250c',
        'm': '\u2514',
        'n': '\u253c',
        'o': '\u23ba',
        'p': '\u23bb',
        'q': '\u2500',
        'r': '\u23bc',
        's': '\u23bd',
        't': '\u251c',
        'u': '\u2524',
        'v': '\u2534',
        'w': '\u252c',
        'x': '\u2502',
        'y': '\u2264',
        'z': '\u2265',
        '{': '\u03c0',
        '|': '\u2260',
        '}': '\u00a3',
        '~': '\u00b7' // '·'
    },
    "UK": null,
    "US": null,
    "Dutch": null,
    "Finnish": null,
    "French": null,
    "FrenchCanadian": null,
    "German": null,
    "Italian": null,
    "NorwegianDanish": null,
    "Spanish": null,
    "Swedish": null,
    "Swiss": null,
    "ISOLatin": null // /A
};
exports.Emulator = Emulator;
/**
 * Helpers
 */
function cancelEvent(ev) {
    if (ev.preventDefault)
        ev.preventDefault();
    ev.returnValue = false;
    if (ev.stopPropagation)
        ev.stopPropagation();
    ev.cancelBubble = true;
    return false;
}
function isWide(ch) {
    if (ch <= '\uff00')
        return false;
    return (ch >= '\uff01' && ch <= '\uffbe') ||
        (ch >= '\uffc2' && ch <= '\uffc7') ||
        (ch >= '\uffca' && ch <= '\uffcf') ||
        (ch >= '\uffd2' && ch <= '\uffd7') ||
        (ch >= '\uffda' && ch <= '\uffdc') ||
        (ch >= '\uffe0' && ch <= '\uffe6') ||
        (ch >= '\uffe8' && ch <= '\uffee');
}
const matchColorCache = {};
function matchColor(r1, g1, b1) {
    var hash = (r1 << 16) | (g1 << 8) | b1;
    if (matchColorCache[hash] !== undefined) {
        return matchColorCache[hash];
    }
    var ldiff = Infinity;
    var li = -1;
    var i = 0;
    var c;
    var r2;
    var g2;
    var b2;
    var diff;
    for (; i < Emulator.vcolors.length; i++) {
        c = Emulator.vcolors[i];
        r2 = c[0];
        g2 = c[1];
        b2 = c[2];
        diff = matchColorDistance(r1, g1, b1, r2, g2, b2);
        if (diff === 0) {
            li = i;
            break;
        }
        if (diff < ldiff) {
            ldiff = diff;
            li = i;
        }
    }
    matchColorCache[hash] = li;
    return li;
}
// http://stackoverflow.com/questions/1633828
function matchColorDistance(r1, g1, b1, r2, g2, b2) {
    return Math.pow(30 * (r1 - r2), 2) +
        Math.pow(59 * (g1 - g2), 2) +
        Math.pow(11 * (b1 - b2), 2);
}
//# sourceMappingURL=Term.js.map