"use strict";
/*
 * Copyright 2014-2016 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const ThemeableElementBase_1 = require("../ThemeableElementBase");
const ThemeTypes = require("../Theme");
const MenuItem_1 = require("./MenuItem");
const CheckboxMenuItem_1 = require("./CheckboxMenuItem");
const DomUtils = require("../DomUtils");
const Util = require("./Util");
const Logger_1 = require("../Logger");
MenuItem_1.MenuItem.init();
const ID = "EtContextMenuTemplate";
const ID_COVER = "ID_COVER";
const ID_CONTAINER = "ID_CONTAINER";
const CLASS_COVER_CLOSED = "CLASS_COVER_CLOSED";
const CLASS_COVER_OPEN = "CLASS_COVER_OPEN";
const CLASS_CONTAINER_CLOSED = "CLASS_CONTAINER_CLOSED";
const CLASS_CONTAINER_OPEN = "CLASS_CONTAINER_OPEN";
let registered = false;
/**
 * A context menu.
 */
class ContextMenu extends ThemeableElementBase_1.ThemeableElementBase {
    //-----------------------------------------------------------------------
    //
    //   #                                                         
    //   #       # ###### ######  ####  #   #  ####  #      ###### 
    //   #       # #      #      #    #  # #  #    # #      #      
    //   #       # #####  #####  #        #   #      #      #####  
    //   #       # #      #      #        #   #      #      #      
    //   #       # #      #      #    #   #   #    # #      #      
    //   ####### # #      ######  ####    #    ####  ###### ###### 
    //
    //-----------------------------------------------------------------------
    constructor() {
        super();
        this._log = Logger_1.getLogger(ContextMenu.TAG_NAME, this);
        const shadow = this.attachShadow({ mode: 'open', delegatesFocus: false });
        const clone = this.createClone();
        shadow.appendChild(clone);
        this.updateThemeCss();
        const cover = this.__getById(ID_COVER);
        cover.addEventListener('mousedown', (ev) => {
            ev.stopPropagation();
            ev.preventDefault();
            if (ev.button === 0) {
                this.close();
            }
        });
        cover.addEventListener('contextmenu', (ev) => {
            ev.stopPropagation();
            ev.preventDefault();
            this.close();
        }, true);
        const container = this.__getById(ID_CONTAINER);
        container.addEventListener('mousedown', (ev) => {
            ev.stopPropagation();
            ev.preventDefault();
        });
        container.addEventListener('mousemove', (ev) => {
            if (ev.srcElement.nodeName === MenuItem_1.MenuItem.TAG_NAME || ev.srcElement.nodeName === CheckboxMenuItem_1.CheckboxMenuItem.TAG_NAME) {
                this.selectMenuItem(this.childNodes, ev.srcElement);
            }
            else {
                this.selectMenuItem(this.childNodes, null);
            }
        });
        container.addEventListener('mouseleave', (ev) => {
            this.selectMenuItem(this.childNodes, null);
        });
        container.addEventListener('click', (ev) => {
            if (ev.srcElement instanceof MenuItem_1.MenuItem) {
                const item = ev.srcElement;
                this.activateItem(item);
            }
        });
        container.addEventListener('keydown', (ev) => { this.handleKeyDown(ev); });
        container.addEventListener('keypress', (ev) => { this.handleKeyPress(ev); });
    }
    /**
     * Initialize the ContextMenu class and resources.
     *
     * When ContextMenu is imported into a render process, this static method
     * must be called before an instances may be created. This can be safely
     * called multiple times.
     */
    static init() {
        if (registered === false) {
            window.customElements.define(ContextMenu.TAG_NAME.toLowerCase(), ContextMenu);
            registered = true;
        }
    }
    _themeCssFiles() {
        return [ThemeTypes.CssFile.GUI_CONTEXTMENU];
    }
    createClone() {
        let template = window.document.getElementById(ID);
        if (template === null) {
            template = window.document.createElement('template');
            template.id = ID;
            template.innerHTML = `<style id="${ThemeableElementBase_1.ThemeableElementBase.ID_THEME}"></style>
          <div id='${ID_COVER}' class='${CLASS_COVER_CLOSED}'></div>
          <div id='${ID_CONTAINER}' class='${CLASS_CONTAINER_CLOSED}' tabindex='0'><slot></slot></div>`;
            window.document.body.appendChild(template);
        }
        return window.document.importNode(template.content, true);
    }
    __getById(id) {
        return DomUtils.getShadowRoot(this).querySelector('#' + id);
    }
    fetchMenuItems(kids) {
        const len = kids.length;
        const result = [];
        for (let i = 0; i < len; i++) {
            const item = kids[i];
            if (item instanceof MenuItem_1.MenuItem) {
                result.push(item);
            }
        }
        return result;
    }
    selectMenuItem(kids, selectitem) {
        const len = kids.length;
        for (let i = 0; i < len; i++) {
            const item = kids[i];
            if (item instanceof MenuItem_1.MenuItem) {
                item.setAttribute('selected', selectitem === item ? "true" : "false");
            }
        }
    }
    handleKeyDown(ev) {
        if (ev.key === "Escape") {
            this.close();
            ev.preventDefault();
            ev.stopPropagation();
            return;
        }
        if (ev.key === "ArrowUp" || ev.key === "ArrowDown" || ev.key === "Enter") {
            const menuitems = this.fetchMenuItems(this.childNodes);
            if (menuitems.length === 0) {
                return;
            }
            const keyboardselected = menuitems.filter((item) => Util.htmlValueToBool(item.getAttribute("selected")));
            if (ev.key === "ArrowUp") {
                if (keyboardselected.length === 0) {
                    this.selectMenuItem(this.childNodes, menuitems[menuitems.length - 1]);
                }
                else {
                    let i = menuitems.indexOf(keyboardselected[0]);
                    i = i === 0 ? menuitems.length - 1 : i - 1;
                    this.selectMenuItem(this.childNodes, menuitems[i]);
                }
            }
            else if (ev.key === "ArrowDown") {
                if (keyboardselected.length === 0) {
                    this.selectMenuItem(this.childNodes, menuitems[0]);
                }
                else {
                    let i = menuitems.indexOf(keyboardselected[0]) + 1;
                    if (i === menuitems.length) {
                        i = 0;
                    }
                    this.selectMenuItem(this.childNodes, menuitems[i]);
                }
            }
            else {
                // Enter
                ev.stopPropagation();
                return;
            }
        }
        ev.preventDefault();
        ev.stopPropagation();
    }
    activateItem(item) {
        item._clicked();
        const name = item.getAttribute('name');
        const checked = item.getAttribute('checked');
        this.close();
        const event = new CustomEvent('selected', { detail: { name: name, checked: checked } });
        this.dispatchEvent(event);
    }
    handleKeyPress(ev) {
        ev.preventDefault();
        ev.stopPropagation();
        if (ev.key === "Enter") {
            const menuitems = this.fetchMenuItems(this.childNodes);
            if (menuitems.length === 0) {
                return;
            }
            const keyboardselected = menuitems.filter((item) => Util.htmlValueToBool(item.getAttribute("selected")));
            if (keyboardselected.length !== 0) {
                this.activateItem(keyboardselected[0]);
            }
        }
    }
    open(x, y) {
        // Nuke any style like 'display: none' which can be use to prevent flicker.
        this.setAttribute('style', '');
        const container = this.__getById(ID_CONTAINER);
        container.classList.remove(CLASS_CONTAINER_CLOSED);
        container.classList.add(CLASS_CONTAINER_OPEN);
        const rect = container.getBoundingClientRect();
        var sx = x;
        if (sx + rect.width > window.innerWidth) {
            sx = window.innerWidth - rect.width;
        }
        var sy = y;
        if (sy + rect.height > window.innerHeight) {
            sy = window.innerHeight - rect.height;
        }
        container.style.left = "" + sx + "px";
        container.style.top = "" + sy + "px";
        this._openCover();
        this.selectMenuItem(this.childNodes, null);
        container.focus();
    }
    _openCover() {
        const cover = DomUtils.getShadowId(this, ID_COVER);
        cover.className = CLASS_COVER_OPEN;
    }
    debugScroll(msg) {
        const text = msg !== undefined ? msg : "";
        const termdiv = window.document.querySelector('div.terminal');
        console.log(text + " -- termdiv.scrollTop: " + termdiv.scrollTop);
        const active = window.document.activeElement;
        console.log("active element: " + active);
        if (active !== null) {
            console.log("active element nodeName: " + active.nodeName);
            console.log("active element class: " + active.getAttribute('class'));
        }
    }
    openAround(targetElement) {
        // Nuke any style like 'display: none' which can be use to prevent flicker.
        this.setAttribute('style', '');
        const targetElementRect = targetElement.getBoundingClientRect();
        const container = this.__getById(ID_CONTAINER);
        container.classList.remove(CLASS_CONTAINER_CLOSED);
        container.classList.add(CLASS_CONTAINER_OPEN);
        const containerRect = container.getBoundingClientRect();
        let containerX = targetElementRect.left;
        if (containerX + containerRect.width > window.innerWidth) {
            containerX = window.innerWidth - containerRect.width;
        }
        let containerY = targetElementRect.bottom;
        if (containerY + containerRect.height > window.innerHeight) {
            containerY = targetElementRect.top - containerRect.height;
        }
        container.style.left = "" + containerX + "px";
        container.style.top = "" + containerY + "px";
        this._openCover();
        this.selectMenuItem(this.childNodes, null);
        container.focus();
    }
    close() {
        let event = new CustomEvent('before-close', { detail: null });
        this.dispatchEvent(event);
        const cover = this.__getById(ID_COVER);
        cover.className = CLASS_COVER_CLOSED;
        const container = this.__getById(ID_CONTAINER);
        container.classList.remove(CLASS_CONTAINER_OPEN);
        container.classList.add(CLASS_CONTAINER_CLOSED);
        event = new CustomEvent('close', { detail: null });
        this.dispatchEvent(event);
    }
}
/**
 * The HTML tag name of this element.
 */
ContextMenu.TAG_NAME = "ET-CONTEXTMENU";
exports.ContextMenu = ContextMenu;
//# sourceMappingURL=ContextMenu.js.map