"use strict";
/*
 * Copyright 2017 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const ThemeableElementBase_1 = require("../ThemeableElementBase");
const ThemeTypes = require("../Theme");
const DomUtils = require("../DomUtils");
const PopDownDialog_1 = require("./PopDownDialog");
const Logger_1 = require("../Logger");
const ID = "EtListPickerTemplate";
const ID_DIALOG = "ID_DIALOG";
const ID_FILTER = "ID_FILTER";
const ID_RESULTS = "ID_RESULTS";
let registered = false;
/**
 * A List Picker.
 */
class ListPicker extends ThemeableElementBase_1.ThemeableElementBase {
    //-----------------------------------------------------------------------
    //
    //   #                                                         
    //   #       # ###### ######  ####  #   #  ####  #      ###### 
    //   #       # #      #      #    #  # #  #    # #      #      
    //   #       # #####  #####  #        #   #      #      #####  
    //   #       # #      #      #        #   #      #      #      
    //   #       # #      #      #    #   #   #    # #      #      
    //   ####### # #      ######  ####    #    ####  ###### ###### 
    //
    //-----------------------------------------------------------------------
    constructor() {
        super();
        this._initProperties(); // Initialise our properties. The constructor was not called.
        const shadow = this.attachShadow({ mode: 'open', delegatesFocus: true });
        const clone = this.createClone();
        shadow.appendChild(clone);
        this.updateThemeCss();
        const filterInput = DomUtils.getShadowId(this, ID_FILTER);
        filterInput.addEventListener('input', (ev) => {
            this._updateEntries();
        });
        filterInput.addEventListener('keydown', (ev) => { this.handleKeyDown(ev); });
        const resultsDiv = DomUtils.getShadowId(this, ID_RESULTS);
        resultsDiv.addEventListener('click', (ev) => {
            for (let node of ev.path) {
                if (node instanceof HTMLElement) {
                    const dataId = node.attributes.getNamedItem(ListPicker.ATTR_DATA_ID);
                    if (dataId !== undefined && dataId !== null) {
                        this._okId(dataId.value);
                    }
                }
            }
        });
    }
    /**
     * Initialize the PopDownListPicker class and resources.
     *
     * When PopDownListPicker is imported into a render process, this static method
     * must be called before an instances may be created. This is can be safely
     * called multiple times.
     */
    static init() {
        PopDownDialog_1.PopDownDialog.init();
        if (registered === false) {
            window.customElements.define(ListPicker.TAG_NAME.toLowerCase(), ListPicker);
            registered = true;
        }
    }
    _initProperties() {
        this._log = Logger_1.getLogger(ListPicker.TAG_NAME, this);
        this._entries = [];
        this._selectedId = null;
        this._filterEntries = (entries, filterText) => entries;
        this._formatEntries = (filteredEntries, selectedId, filterInputValue) => filteredEntries.map(entry => `<div ${ListPicker.ATTR_DATA_ID}='${entry.id}'>${entry.id}</div>`).join("");
        this._laterHandle = null;
        this._extraCssFiles = [];
    }
    focus() {
        const filter = DomUtils.getShadowId(this, ID_FILTER);
        if (filter != null) {
            filter.focus();
        }
    }
    getFilter() {
        const filterInput = DomUtils.getShadowId(this, ID_FILTER);
        return filterInput.value;
    }
    setFilter(text) {
        const filterInput = DomUtils.getShadowId(this, ID_FILTER);
        filterInput.value = text;
        this._updateEntries();
    }
    getSelected() {
        return this._selectedId;
    }
    setSelected(selectedId) {
        this._selectedId = selectedId;
        this._updateEntries();
        this._scrollToSelected();
    }
    setEntries(entries) {
        this._entries = entries;
        this._selectedId = null;
        const filterInput = DomUtils.getShadowId(this, ID_FILTER);
        if (filterInput !== null) {
            filterInput.value = "";
        }
        this._updateEntries();
    }
    getEntries() {
        return this._entries;
    }
    setFilterAndRankEntriesFunc(func) {
        this._filterEntries = func;
    }
    setFormatEntriesFunc(func) {
        this._formatEntries = func;
    }
    /**
     * Specify extra Css files to load into this element.
     *
     * @param extraCssFiles extra Css files which should be loaded along side the default set.
     */
    addExtraCss(extraCssFiles) {
        this._extraCssFiles = [...this._extraCssFiles, ...extraCssFiles];
        this.updateThemeCss();
    }
    /**
     *
     */
    createClone() {
        let template = window.document.getElementById(ID);
        if (template === null) {
            template = window.document.createElement('template');
            template.id = ID;
            template.innerHTML = `<style id="${ThemeableElementBase_1.ThemeableElementBase.ID_THEME}"></style>
        <div id="${ID_DIALOG}">
          <div class="form-group"><input type="text" id="${ID_FILTER}" class="form-control input-sm" /></div>
          <div id="${ID_RESULTS}"></div>
        </div>
        `;
            window.document.body.appendChild(template);
        }
        return window.document.importNode(template.content, true);
    }
    _themeCssFiles() {
        return [ThemeTypes.CssFile.GUI_CONTROLS, ThemeTypes.CssFile.FONT_AWESOME,
            ThemeTypes.CssFile.GUI_LIST_PICKER, ...this._extraCssFiles]; // FIXME
    }
    _updateEntries() {
        const filterInputValue = DomUtils.getShadowId(this, ID_FILTER).value;
        const filteredEntries = this._filterEntries(this._entries, filterInputValue);
        if (filteredEntries.length === 0) {
            this._selectedId = null;
        }
        else {
            const newSelectedIndex = filteredEntries.findIndex((entry) => entry.id === this._selectedId);
            this._selectedId = filteredEntries[Math.max(0, newSelectedIndex)].id;
        }
        const html = this._formatEntries(filteredEntries, this._selectedId, filterInputValue);
        DomUtils.getShadowId(this, ID_RESULTS).innerHTML = html;
    }
    _scrollToSelected() {
        const resultsDiv = DomUtils.getShadowId(this, ID_RESULTS);
        const selectedElement = resultsDiv.querySelector("." + ListPicker.CLASS_RESULT_SELECTED);
        const selectedRelativeTop = selectedElement.offsetTop - resultsDiv.offsetTop;
        resultsDiv.scrollTop = selectedRelativeTop;
    }
    //-----------------------------------------------------------------------
    /**
     *
     */
    handleKeyDown(ev) {
        if (ev.key === "Escape") {
            this._okId(null);
            ev.preventDefault();
            ev.stopPropagation();
            return;
        }
        const isPageKey = ev.key === "PageUp" || ev.key === "PageDown";
        const isUp = ev.key === "PageUp" || ev.key === "ArrowUp" || ev.key === "Home";
        if (isPageKey || isUp || ev.key === "ArrowDown" || ev.key === "End" || ev.key === "Enter") {
            ev.preventDefault();
            ev.stopPropagation();
            const filterInput = DomUtils.getShadowId(this, ID_FILTER);
            const filteredEntries = this._filterEntries(this._entries, filterInput.value);
            if (filteredEntries.length === 0) {
                return;
            }
            const selectedIndex = filteredEntries.findIndex((entry) => entry.id === this._selectedId);
            if (ev.key === "Enter") {
                if (this._selectedId !== null) {
                    this._okId(this._selectedId);
                }
            }
            else {
                const resultsDiv = DomUtils.getShadowId(this, ID_RESULTS);
                // Determine the step size.
                let stepSize = 1;
                if (isPageKey) {
                    const selectedElement = resultsDiv.querySelector("." + ListPicker.CLASS_RESULT_SELECTED);
                    const selectedElementDimensions = selectedElement.getBoundingClientRect();
                    stepSize = Math.floor(resultsDiv.clientHeight / selectedElementDimensions.height);
                }
                if (isUp) {
                    if (ev.key === "Home") {
                        this._selectedId = filteredEntries[0].id;
                    }
                    else {
                        this._selectedId = filteredEntries[Math.max(0, selectedIndex - stepSize)].id;
                    }
                }
                else {
                    if (ev.key === "End") {
                        this._selectedId = filteredEntries[filteredEntries.length - 1].id;
                    }
                    else {
                        this._selectedId = filteredEntries[Math.min(filteredEntries.length - 1, selectedIndex + stepSize)].id;
                    }
                }
                const top = resultsDiv.scrollTop;
                this._updateEntries();
                resultsDiv.scrollTop = top;
                const selectedElement = resultsDiv.querySelector("." + ListPicker.CLASS_RESULT_SELECTED);
                const selectedRelativeTop = selectedElement.offsetTop - resultsDiv.offsetTop;
                if (top > selectedRelativeTop) {
                    resultsDiv.scrollTop = selectedRelativeTop;
                }
                else {
                    const selectedElementDimensions = selectedElement.getBoundingClientRect();
                    if (selectedRelativeTop + selectedElementDimensions.height > top + resultsDiv.clientHeight) {
                        resultsDiv.scrollTop = selectedRelativeTop + selectedElementDimensions.height - resultsDiv.clientHeight;
                    }
                }
            }
        }
    }
    /**
     *
     */
    open(x, y, width, height) {
        const resultsDiv = DomUtils.getShadowId(this, ID_RESULTS);
        resultsDiv.style.maxHeight = `${height / 2}px`;
        const filterInput = DomUtils.getShadowId(this, ID_FILTER);
        filterInput.value = "";
        this._updateEntries();
        filterInput.focus();
        this._scrollToSelected();
    }
    _okId(selectedId) {
        if (this._laterHandle === null) {
            this._laterHandle = DomUtils.doLater(() => {
                this._laterHandle = null;
                const event = new CustomEvent("selected", { detail: { selected: selectedId } });
                this.dispatchEvent(event);
            });
        }
    }
}
/**
 * The HTML tag name of this element.
 */
ListPicker.TAG_NAME = "ET-LISTPICKER";
ListPicker.ATTR_DATA_ID = "data-id";
ListPicker.CLASS_RESULT_SELECTED = "CLASS_RESULT_SELECTED";
ListPicker.CLASS_RESULT_ENTRY = "CLASS_RESULT_ENTRY";
exports.ListPicker = ListPicker;
//# sourceMappingURL=ListPicker.js.map