"use strict";
/*
 * Copyright 2017 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const ThemeableElementBase_1 = require("../ThemeableElementBase");
const ThemeTypes = require("../Theme");
const DomUtils = require("../DomUtils");
const ID = "EtPopDownDialogTemplate";
const ID_COVER = "ID_COVER";
const ID_CONTEXT_COVER = "ID_CONTEXT_COVER";
const ID_CONTAINER = "ID_CONTAINER";
const ID_TITLE_PRIMARY = "ID_TITLE_PRIMARY";
const ID_TITLE_SECONDARY = "ID_TITLE_SECONDARY";
const ID_TITLE_CONTAINER = "ID_TITLE_CONTAINER";
const CLASS_CONTEXT_COVER_OPEN = "CLASS_CONTEXT_COVER_OPEN";
const CLASS_CONTEXT_COVER_CLOSED = "CLASS_CONTEXT_COVER_CLOSED";
const CLASS_COVER_CLOSED = "CLASS_COVER_CLOSED";
const CLASS_COVER_OPEN = "CLASS_COVER_OPEN";
const ATTR_DATA_ID = "data-id";
let registered = false;
/**
 * A Pop Down Dialog.
 */
class PopDownDialog extends ThemeableElementBase_1.ThemeableElementBase {
    //-----------------------------------------------------------------------
    //
    //   #                                                         
    //   #       # ###### ######  ####  #   #  ####  #      ###### 
    //   #       # #      #      #    #  # #  #    # #      #      
    //   #       # #####  #####  #        #   #      #      #####  
    //   #       # #      #      #        #   #      #      #      
    //   #       # #      #      #    #   #   #    # #      #      
    //   ####### # #      ######  ####    #    ####  ###### ###### 
    //
    //-----------------------------------------------------------------------
    constructor() {
        super();
        this._initProperties(); // Initialise our properties. The constructor was not called.
        const shadow = this.attachShadow({ mode: 'open', delegatesFocus: true });
        const clone = this.createClone();
        shadow.appendChild(clone);
        this.updateThemeCss();
        const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
        containerDiv.addEventListener('contextmenu', (ev) => {
            this.dispatchEvent(new CustomEvent(PopDownDialog.EVENT_CLOSE_REQUEST, { bubbles: false }));
        });
        const coverDiv = DomUtils.getShadowId(this, ID_COVER);
        coverDiv.addEventListener('mousedown', (ev) => {
            this.dispatchEvent(new CustomEvent(PopDownDialog.EVENT_CLOSE_REQUEST, { bubbles: false }));
        });
    }
    /**
     * Initialize the PopDownDialog class and resources.
     *
     * When PopDownDialog is imported into a render process, this static method
     * must be called before an instances may be created. This is can be safely
     * called multiple times.
     */
    static init() {
        if (registered === false) {
            window.customElements.define(PopDownDialog.TAG_NAME.toLowerCase(), PopDownDialog);
            registered = true;
        }
    }
    _initProperties() {
        this._laterHandle = null;
        this._titlePrimary = "";
        this._titleSecondary = "";
    }
    setTitlePrimary(text) {
        this._titlePrimary = text;
        this._updateTitle();
    }
    getTitlePrimary() {
        return this._titlePrimary;
    }
    setTitleSecondary(text) {
        this._titleSecondary = text;
        this._updateTitle();
    }
    getTitleSecondary() {
        return this._titleSecondary;
    }
    /**
     *
     */
    createClone() {
        let template = window.document.getElementById(ID);
        if (template === null) {
            template = window.document.createElement('template');
            template.id = ID;
            template.innerHTML = `<style id="${ThemeableElementBase_1.ThemeableElementBase.ID_THEME}"></style>
        <div id='${ID_COVER}' class='${CLASS_COVER_CLOSED}'></div>
        <div id='${ID_CONTEXT_COVER}' class='${CLASS_CONTEXT_COVER_CLOSED}'>
          <div id='${ID_CONTAINER}'>
            <div id="${ID_TITLE_CONTAINER}"><div id="${ID_TITLE_PRIMARY}"></div><div id="${ID_TITLE_SECONDARY}"></div></div>
            <slot></slot>
          </div>
        </div>`;
            window.document.body.appendChild(template);
        }
        return window.document.importNode(template.content, true);
    }
    _themeCssFiles() {
        return [ThemeTypes.CssFile.GUI_CONTROLS, ThemeTypes.CssFile.FONT_AWESOME, ThemeTypes.CssFile.GUI_POP_DOWN_DIALOG];
    }
    //-----------------------------------------------------------------------
    _updateTitle() {
        const titlePrimaryDiv = DomUtils.getShadowId(this, ID_TITLE_PRIMARY);
        const titleSecondaryDiv = DomUtils.getShadowId(this, ID_TITLE_SECONDARY);
        titlePrimaryDiv.innerText = this._titlePrimary;
        titleSecondaryDiv.innerText = this._titleSecondary;
    }
    /**
     *
     */
    open(x, y, width, height) {
        // Nuke any style like 'display: none' which can be use to prevent flicker.
        this.setAttribute('style', '');
        const container = DomUtils.getShadowId(this, ID_CONTEXT_COVER);
        container.classList.remove(CLASS_CONTEXT_COVER_CLOSED);
        container.classList.add(CLASS_CONTEXT_COVER_OPEN);
        container.style.left = `${x}px`;
        container.style.top = `${y}px`;
        container.style.width = `${width}px`;
        container.style.height = `${height}px`;
        const cover = DomUtils.getShadowId(this, ID_COVER);
        cover.classList.remove(CLASS_COVER_CLOSED);
        cover.classList.add(CLASS_COVER_OPEN);
    }
    /**
     *
     */
    close() {
        const cover = DomUtils.getShadowId(this, ID_COVER);
        cover.classList.remove(CLASS_COVER_OPEN);
        cover.classList.add(CLASS_COVER_CLOSED);
        const container = DomUtils.getShadowId(this, ID_CONTEXT_COVER);
        container.classList.remove(CLASS_CONTEXT_COVER_OPEN);
        container.classList.add(CLASS_CONTEXT_COVER_CLOSED);
    }
}
/**
 * The HTML tag name of this element.
 */
PopDownDialog.TAG_NAME = "ET-POPDOWNDIALOG";
PopDownDialog.EVENT_CLOSE_REQUEST = "ET-POPDOWNDIALOG-CLOSE_REQUEST";
exports.PopDownDialog = PopDownDialog;
//# sourceMappingURL=PopDownDialog.js.map