"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/*
 * Copyright 2017 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */
const ThemeableElementBase_1 = require("../ThemeableElementBase");
const ThemeTypes = require("../Theme");
const DomUtils = require("../DomUtils");
const InternalMimeTypes_1 = require("../InternalMimeTypes");
const Logger_1 = require("../Logger");
const ID = "EtSnapDropContainerTemplate";
const ID_TOP = "ID_TOP";
const ID_DRAG_COVER = "ID_DRAG_COVER";
const ID_CONTENTS = "ID_CONTENTS";
const CLASS_NOT_DRAGGING = "CLASS_NOT_DRAGGING";
const CLASS_DRAGGING = "CLASS_DRAGGING";
const CLASS_DROP_TARGET_MIDDLE = "CLASS_DROP_TARGET_MIDDLE";
const CLASS_DROP_TARGET_NORTH = "CLASS_DROP_TARGET_NORTH";
const CLASS_DROP_TARGET_SOUTH = "CLASS_DROP_TARGET_SOUTH";
const CLASS_DROP_TARGET_EAST = "CLASS_DROP_TARGET_EAST";
const CLASS_DROP_TARGET_WEST = "CLASS_DROP_TARGET_WEST";
let registered = false;
var DropLocation;
(function (DropLocation) {
    DropLocation[DropLocation["NORTH"] = 0] = "NORTH";
    DropLocation[DropLocation["SOUTH"] = 1] = "SOUTH";
    DropLocation[DropLocation["EAST"] = 2] = "EAST";
    DropLocation[DropLocation["WEST"] = 3] = "WEST";
    DropLocation[DropLocation["MIDDLE"] = 4] = "MIDDLE";
})(DropLocation = exports.DropLocation || (exports.DropLocation = {}));
const dropLocationToCss = new Map();
dropLocationToCss.set(DropLocation.NORTH, CLASS_DROP_TARGET_NORTH);
dropLocationToCss.set(DropLocation.SOUTH, CLASS_DROP_TARGET_SOUTH);
dropLocationToCss.set(DropLocation.EAST, CLASS_DROP_TARGET_EAST);
dropLocationToCss.set(DropLocation.WEST, CLASS_DROP_TARGET_WEST);
dropLocationToCss.set(DropLocation.MIDDLE, CLASS_DROP_TARGET_MIDDLE);
const SUPPORTED_MIMETYPES = [InternalMimeTypes_1.ElementMimeType.MIMETYPE, InternalMimeTypes_1.FrameMimeType.MIMETYPE];
/**
 * A container which supports splitting and snapping for dragged items.
 */
class SnapDropContainer extends ThemeableElementBase_1.ThemeableElementBase {
    //-----------------------------------------------------------------------
    //
    //   #                                                         
    //   #       # ###### ######  ####  #   #  ####  #      ###### 
    //   #       # #      #      #    #  # #  #    # #      #      
    //   #       # #####  #####  #        #   #      #      #####  
    //   #       # #      #      #        #   #      #      #      
    //   #       # #      #      #    #   #   #    # #      #      
    //   ####### # #      ######  ####    #    ####  ###### ###### 
    //
    //-----------------------------------------------------------------------
    constructor() {
        super();
        this._initProperties();
        const shadow = this.attachShadow({ mode: 'open', delegatesFocus: false });
        const clone = this.createClone();
        shadow.appendChild(clone);
        this.updateThemeCss();
        const topDiv = DomUtils.getShadowId(this, ID_TOP);
        topDiv.classList.add(CLASS_NOT_DRAGGING);
        topDiv.addEventListener("dragenter", this._handleDragEnter.bind(this), true);
        topDiv.addEventListener("dragover", this._handleDragOver.bind(this), true);
        topDiv.addEventListener("dragleave", this._handleDragLeave.bind(this), true);
        topDiv.addEventListener("drop", this._handleDrop.bind(this), true);
    }
    /**
     * Initialize the SnapDropContainer class and resources.
     *
     * When SnapDropContainer is imported into a render process, this static
     * method must be called before an instances may be created. This is can
     * be safely called multiple times.
     */
    static init() {
        if (registered === false) {
            window.customElements.define(SnapDropContainer.TAG_NAME.toLowerCase(), SnapDropContainer);
            registered = true;
        }
    }
    _initProperties() {
        this._log = Logger_1.getLogger(SnapDropContainer.TAG_NAME, this);
    }
    createClone() {
        let template = window.document.getElementById(ID);
        if (template === null) {
            template = window.document.createElement('template');
            template.id = ID;
            template.innerHTML = `<style id="${ThemeableElementBase_1.ThemeableElementBase.ID_THEME}"></style>
<div id='${ID_TOP}'>
  <div id='${ID_CONTENTS}'><slot></slot></div>
  <div id='${ID_DRAG_COVER}'></div>
</div>
`;
            window.document.body.appendChild(template);
        }
        return window.document.importNode(template.content, true);
    }
    _themeCssFiles() {
        return [ThemeTypes.CssFile.GUI_SNAP_DROP_CONTAINER];
    }
    _handleDragEnter(ev) {
        if (this._isSupportedDropMimeType(ev)) {
            ev.stopPropagation();
            this._showDragCover(ev);
        }
    }
    _handleDragOver(ev) {
        if (this._isSupportedDropMimeType(ev)) {
            ev.preventDefault();
            ev.stopPropagation();
            this._showDragCover(ev);
        }
    }
    _handleDragLeave(ev) {
        if (this._isSupportedDropMimeType(ev)) {
            ev.stopPropagation();
            if (ev.target == DomUtils.getShadowId(this, ID_DRAG_COVER)) {
                this._hideDragCover();
            }
        }
    }
    _showDragCover(ev) {
        const dragCoverDiv = DomUtils.getShadowId(this, ID_DRAG_COVER);
        DomUtils.removeAllClasses(dragCoverDiv);
        dragCoverDiv.classList.add(dropLocationToCss.get(this._cursorPositionToDropLocation(ev)));
        const topDiv = DomUtils.getShadowId(this, ID_TOP);
        topDiv.classList.add(CLASS_DRAGGING);
        topDiv.classList.remove(CLASS_NOT_DRAGGING);
    }
    // The drag over regions are north, south, east and west.
    // There is a neutral square in the middle of the drop area.
    //
    //  +-----+
    //  |\___/|
    //  | |_| |
    //  |/   \|
    //  +-----+
    _cursorPositionToDropLocation(ev) {
        const widgetRect = this.getBoundingClientRect();
        const pointXInUnitSpace = (ev.pageX - widgetRect.left) / widgetRect.width;
        const pointYInUnitSpace = (ev.pageY - widgetRect.top) / widgetRect.height;
        const inMiddle = pointXInUnitSpace > 1 / 3 && pointXInUnitSpace < 2 / 3 &&
            pointYInUnitSpace > 1 / 3 && pointYInUnitSpace < 2 / 3;
        if (inMiddle) {
            return DropLocation.MIDDLE;
        }
        const oneOneVectorSide = pointXInUnitSpace < pointYInUnitSpace;
        const reverseOneOneVectorSide = -pointXInUnitSpace + 1 < pointYInUnitSpace;
        if (oneOneVectorSide) {
            if (reverseOneOneVectorSide) {
                return DropLocation.SOUTH;
            }
            else {
                return DropLocation.WEST;
            }
        }
        else {
            if (reverseOneOneVectorSide) {
                return DropLocation.EAST;
            }
            else {
                return DropLocation.NORTH;
            }
        }
    }
    _hideDragCover() {
        const topDiv = DomUtils.getShadowId(this, ID_TOP);
        topDiv.classList.add(CLASS_NOT_DRAGGING);
        topDiv.classList.remove(CLASS_DRAGGING);
    }
    _handleDrop(ev) {
        this._hideDragCover();
        const { mimeType, data } = this._getSupportedDropMimeTypeData(ev);
        if (mimeType != null) {
            const detail = {
                dropLocation: this._cursorPositionToDropLocation(ev),
                mimeType: mimeType,
                dropData: data
            };
            const customDropEvent = new CustomEvent(SnapDropContainer.EVENT_DROPPED, { bubbles: true, detail: detail });
            this.dispatchEvent(customDropEvent);
        }
    }
    _isSupportedDropMimeType(ev) {
        for (let i = 0; i < ev.dataTransfer.items.length; i++) {
            const item = ev.dataTransfer.items[i];
            if (SUPPORTED_MIMETYPES.indexOf(item.type) !== -1) {
                return true;
            }
        }
        return false;
    }
    _getSupportedDropMimeTypeData(ev) {
        for (const mimeType of SUPPORTED_MIMETYPES) {
            const data = ev.dataTransfer.getData(mimeType);
            if (data != null && data !== "") {
                return { mimeType, data };
            }
        }
        return { mimeType: null, data: null };
    }
}
/**
 * The HTML tag name of this element.
 */
SnapDropContainer.TAG_NAME = "ET-SNAPDROPCONTAINER";
SnapDropContainer.EVENT_DROPPED = "snapdropcontainer-dropped";
exports.SnapDropContainer = SnapDropContainer;
//# sourceMappingURL=SnapDropContainer.js.map