"use strict";
/*
 * Copyright 2014-2016 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */
Object.defineProperty(exports, "__esModule", { value: true });
const ThemeableElementBase_1 = require("../ThemeableElementBase");
const ThemeTypes = require("../Theme");
const DomUtils = require("../DomUtils");
const ID = "EtStackedWidgetTemplate";
const ID_CONTAINER = 'ID_CONTAINER';
const ATTR_INDEX = 'data-et-index';
let registered = false;
/**
 * A widget which displays one of its DIV contents at a time.
 */
class StackedWidget extends ThemeableElementBase_1.ThemeableElementBase {
    //-----------------------------------------------------------------------
    //
    //   #                                                         
    //   #       # ###### ######  ####  #   #  ####  #      ###### 
    //   #       # #      #      #    #  # #  #    # #      #      
    //   #       # #####  #####  #        #   #      #      #####  
    //   #       # #      #      #        #   #      #      #      
    //   #       # #      #      #    #   #   #    # #      #      
    //   ####### # #      ######  ####    #    ####  ###### ###### 
    //
    //-----------------------------------------------------------------------
    constructor() {
        super();
        this._initProperties();
    }
    /**
     * Initialize the StackedWidget class and resources.
     *
     * When StackedWidget is imported into a render process, this static method
     * must be called before an instances may be created. This is can be safely
     * called multiple times.
     */
    static init() {
        if (registered === false) {
            window.customElements.define(StackedWidget.TAG_NAME.toLowerCase(), StackedWidget);
            registered = true;
        }
    }
    _initProperties() {
        this._currentIndex = -1;
    }
    connectedCallback() {
        super.connectedCallback();
        if (DomUtils.getShadowRoot(this) !== null) {
            return;
        }
        const shadow = this.attachShadow({ mode: 'open', delegatesFocus: true });
        const clone = this.createClone();
        shadow.appendChild(clone);
        this.updateThemeCss();
        this.createPageHolders();
        this.showIndex(0);
    }
    /**
     *
     */
    createClone() {
        let template = window.document.getElementById(ID);
        if (template === null) {
            template = window.document.createElement('template');
            template.id = ID;
            template.innerHTML = `<style id='${ThemeableElementBase_1.ThemeableElementBase.ID_THEME}'></style>
<div id='${ID_CONTAINER}'></div>`;
            window.document.body.appendChild(template);
        }
        return window.document.importNode(template.content, true);
    }
    /**
     *
     */
    __getById(id) {
        return DomUtils.getShadowRoot(this).querySelector('#' + id);
    }
    _themeCssFiles() {
        return [ThemeTypes.CssFile.GUI_STACKEDWIDGET];
    }
    //-----------------------------------------------------------------------
    // Override
    appendChild(newNode) {
        const result = super.appendChild(newNode);
        if (DomUtils.getShadowRoot(this) === null) {
            return result;
        }
        this.createPageHolders();
        if (this._currentIndex === -1) {
            this._currentIndex = 0;
        }
        this.showIndex(this._currentIndex);
        return result;
    }
    // Override
    removeChild(oldNode) {
        const result = super.removeChild(oldNode);
        this.createPageHolders();
        if (this._currentIndex >= this.childElementCount) {
            this._currentIndex = this.childElementCount - 1;
        }
        this.showIndex(this._currentIndex);
        return result;
    }
    setCurrentIndex(index) {
        if (index < 0 || index >= this.childElementCount) {
            return;
        }
        this._currentIndex = index;
        this.showIndex(index);
    }
    getCurrentIndex() {
        return this._currentIndex;
    }
    showIndex(index) {
        if (DomUtils.getShadowRoot(this) === null) {
            return;
        }
        const container = this.__getById(ID_CONTAINER);
        for (let i = 0; i < container.children.length; i++) {
            const kid = container.children.item(i);
            if (i === index) {
                kid.classList.add('visible');
                kid.classList.remove('hidden');
            }
            else {
                kid.classList.remove('visible');
                kid.classList.add('hidden');
            }
        }
    }
    createPageHolders() {
        const container = this.__getById(ID_CONTAINER);
        for (let i = 0; i < this.children.length; i++) {
            const kid = this.children.item(i);
            kid.slot = "" + i;
        }
        while (container.childElementCount < this.childElementCount) {
            const holderDiv = this.ownerDocument.createElement('div');
            const contentElement = this.ownerDocument.createElement('slot');
            contentElement.setAttribute('name', "" + container.childElementCount);
            holderDiv.appendChild(contentElement);
            container.appendChild(holderDiv);
        }
        while (container.childElementCount > this.childElementCount) {
            container.removeChild(container.children.item(container.children.length - 1));
        }
    }
}
/**
 * The HTML tag name of this element.
 */
StackedWidget.TAG_NAME = 'ET-STACKEDWIDGET';
exports.StackedWidget = StackedWidget;
//# sourceMappingURL=StackedWidget.js.map