"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/*
 * Copyright 2014-2016 Simon Edwards <simon@simonzone.com>
 *
 * This source code is licensed under the MIT license which is detailed in the LICENSE.txt file.
 */
const ThemeableElementBase_1 = require("../ThemeableElementBase");
const ThemeTypes = require("../Theme");
const StackedWidget_1 = require("./StackedWidget");
const Tab_1 = require("./Tab");
const SnapDropContainer_1 = require("./SnapDropContainer");
const GeneralEvents_1 = require("../GeneralEvents");
const InternalMimeTypes_1 = require("../InternalMimeTypes");
const Util = require("./Util");
const DomUtils = require("../DomUtils");
const _ = require("lodash");
const Logger_1 = require("../Logger");
StackedWidget_1.StackedWidget.init();
SnapDropContainer_1.SnapDropContainer.init();
const ID = "EtTabWidgetTemplate";
const ATTR_TAG = 'data-et-tag';
const ATTR_SHOW_FRAME = "show-frame";
const ATTR_TAG_REST_LEFT = "rest-left";
const ATTR_TAG_REST_RIGHT = "rest";
let registered = false;
const ID_TOP = "ID_TOP";
const ID_TABBAR = "ID_TABBAR";
const ID_TABBAR_CONTAINER = "ID_TABBAR_CONTAINER";
const ID_CONTENTSTACK = "ID_CONTENTSTACK";
const ID_SNAP_DROP_CONTAINER = "ID_SNAP_DROP_CONTAINER";
const ID_CONTENTS = "ID_CONTENTS";
const ID_DROP_INDICATOR = "ID_DROP_INDICATOR";
const CLASS_INDICATOR_HIDE = "CLASS_INDICATOR_HIDE";
const CLASS_INDICATOR_SHOW = "CLASS_INDICATOR_SHOW";
const CLASS_REMAINDER_LEFT = "remainder-left";
const CLASS_REMAINDER_RIGHT = "remainder";
const CLASS_ACTIVE = "active";
const CLASS_TAB = "tab";
/**
 * A widget to display a stack of tabs.
 *
 * See Tab.
 */
class TabWidget extends ThemeableElementBase_1.ThemeableElementBase {
    //-----------------------------------------------------------------------
    //
    //   #                                                         
    //   #       # ###### ######  ####  #   #  ####  #      ###### 
    //   #       # #      #      #    #  # #  #    # #      #      
    //   #       # #####  #####  #        #   #      #      #####  
    //   #       # #      #      #        #   #      #      #      
    //   #       # #      #      #    #   #   #    # #      #      
    //   ####### # #      ######  ####    #    ####  ###### ###### 
    //
    //-----------------------------------------------------------------------
    constructor() {
        super();
        this._initProperties();
        const shadow = this.attachShadow({ mode: 'open', delegatesFocus: false });
        const clone = this.createClone();
        shadow.appendChild(clone);
        this.updateThemeCss();
        this.createTabHolders();
        this.setSelectedIndex(0);
        this._showFrame(this.getShowFrame());
        this._mutationObserver = new MutationObserver((mutations) => {
            this.createTabHolders();
        });
        this._mutationObserver.observe(this, { childList: true });
        this._setupDragAndDrop();
    }
    /**
     * Initialize the TabWidget class and resources.
     *
     * When TabWidget is imported into a render process, this static method
     * must be called before an instances may be created. This is can be safely
     * called multiple times.
     */
    static init() {
        if (registered === false) {
            window.customElements.define(TabWidget.TAG_NAME.toLowerCase(), TabWidget);
            registered = true;
        }
    }
    _initProperties() {
        this._log = Logger_1.getLogger(TabWidget.TAG_NAME, this);
        this._mutationObserver = null;
        this._showTabs = true;
    }
    static get observedAttributes() {
        return [ATTR_SHOW_FRAME];
    }
    /**
     * Custom Element 'attribute changed' hook.
     */
    attributeChangedCallback(attrName, oldValue, newValue) {
        switch (attrName) {
            case ATTR_SHOW_FRAME:
                this._showFrame(Util.toBoolean(newValue));
                break;
            default:
                break;
        }
    }
    _themeCssFiles() {
        return [ThemeTypes.CssFile.GUI_CONTROLS, ThemeTypes.CssFile.GUI_TABWIDGET];
    }
    //-----------------------------------------------------------------------
    update() {
        this.createTabHolders();
    }
    refresh(level) {
        const contentsStack = this._getContentsStack();
        if (contentsStack !== null) {
            super.refresh(level);
            contentsStack.refresh(level);
        }
    }
    createClone() {
        let template = window.document.getElementById(ID);
        if (template === null) {
            template = window.document.createElement('template');
            template.id = ID;
            template.innerHTML = `<style id="${ThemeableElementBase_1.ThemeableElementBase.ID_THEME}"></style>
<div id='${ID_TOP}'>
  <div id='${ID_TABBAR_CONTAINER}'>
    <ul id='${ID_TABBAR}' class="extraterm-tabs"></ul>
  </div>
  <div id='${ID_CONTENTS}'>
    <${SnapDropContainer_1.SnapDropContainer.TAG_NAME} id='${ID_SNAP_DROP_CONTAINER}'>
      <${StackedWidget_1.StackedWidget.TAG_NAME} id='${ID_CONTENTSTACK}'></$ {StackedWidget.TAG_NAME}>
    </${SnapDropContainer_1.SnapDropContainer.TAG_NAME}>
  </div>
</div>
`;
            window.document.body.appendChild(template);
        }
        return window.document.importNode(template.content, true);
    }
    __getById(id) {
        return DomUtils.getShadowRoot(this).querySelector('#' + id);
    }
    _getTop() {
        return this.__getById(ID_TOP);
    }
    _getTabbar() {
        return this.__getById(ID_TABBAR);
    }
    _getContentsStack() {
        return this.__getById(ID_CONTENTSTACK);
    }
    createTabHolders() {
        const tabbar = this._getTabbar();
        const contentsStack = this._getContentsStack();
        let tabCount = 0;
        let stateInTab = false;
        // Tag the source content as tabs or content so that we can distribute it over our shadow DOM.
        let restAttr = ATTR_TAG_REST_LEFT;
        for (let i = 0; i < this.children.length; i++) {
            const kid = this.children.item(i);
            if (kid.nodeName === Tab_1.Tab.TAG_NAME) {
                tabCount++;
                kid.slot = 'tab_' + (tabCount - 1);
                stateInTab = true;
            }
            else if (kid.nodeName === "DIV" && stateInTab) {
                kid.slot = 'content_' + (tabCount - 1);
                stateInTab = false;
                restAttr = ATTR_TAG_REST_RIGHT;
            }
            else if (kid.nodeName === "DIV") {
                kid.slot = restAttr;
                restAttr = ATTR_TAG_REST_RIGHT;
            }
        }
        // Make sure that there is 'rest' element at the front.
        if (tabbar.firstElementChild === null || !tabbar.firstElementChild.classList.contains(CLASS_REMAINDER_LEFT)) {
            const restAllLi = this.ownerDocument.createElement('LI');
            restAllLi.classList.add(CLASS_REMAINDER_LEFT);
            const catchAll = this.ownerDocument.createElement('slot');
            catchAll.setAttribute('name', ATTR_TAG_REST_LEFT);
            restAllLi.appendChild(catchAll);
            tabbar.insertAdjacentElement('afterbegin', restAllLi);
        }
        // Make sure that there is a catch all element at the end.
        const catchAlls = tabbar.querySelectorAll("." + CLASS_REMAINDER_RIGHT);
        let catchAllLi = null;
        if (catchAlls.length === 0) {
            catchAllLi = this.ownerDocument.createElement('LI');
            catchAllLi.classList.add(CLASS_REMAINDER_RIGHT);
            const catchAll = this.ownerDocument.createElement('slot');
            catchAll.setAttribute('name', ATTR_TAG_REST_RIGHT);
            catchAllLi.appendChild(catchAll);
            tabbar.appendChild(catchAllLi);
        }
        else {
            catchAllLi = catchAlls[0];
        }
        if (this._showTabs) {
            // Create tabs.
            let tabElementCount = tabbar.querySelectorAll("." + CLASS_TAB).length;
            while (tabElementCount < tabCount) {
                // The tab part.
                const tabLi = this.ownerDocument.createElement('li');
                tabLi.classList.add(CLASS_TAB);
                const contentElement = this.ownerDocument.createElement('slot');
                contentElement.setAttribute('name', 'tab_' + tabElementCount);
                tabLi.appendChild(contentElement);
                tabLi.addEventListener('click', this._tabClickHandler.bind(this, tabLi, tabElementCount));
                tabbar.insertBefore(tabLi, catchAllLi);
                tabElementCount = tabbar.querySelectorAll("." + CLASS_TAB).length;
            }
            // Delete any excess tab tags.
            const tabElements = DomUtils.toArray(tabbar.querySelectorAll("." + CLASS_TAB));
            if (tabElements.length > tabCount) {
                tabElements.slice(tabCount).forEach(el => tabbar.removeChild(el));
            }
        }
        else {
            // Delete all tab tags. We don't need to show them.
            const tabElements = DomUtils.toArray(tabbar.querySelectorAll("." + CLASS_TAB));
            tabElements.forEach(el => tabbar.removeChild(el));
        }
        // Create content holders
        let contentWrapperElementCount = contentsStack.children.length;
        while (contentWrapperElementCount < tabCount) {
            // Pages for the contents stack.
            const wrapperDiv = this.ownerDocument.createElement('div');
            wrapperDiv.classList.add('wrapper');
            const contentElement = this.ownerDocument.createElement('slot');
            contentElement.setAttribute('name', 'content_' + contentWrapperElementCount);
            wrapperDiv.appendChild(contentElement);
            contentsStack.appendChild(wrapperDiv);
            contentWrapperElementCount++;
        }
        while (contentsStack.children.length > tabCount) {
            contentsStack.removeChild(contentsStack.children[contentsStack.children.length - 1]);
        }
        // Try to find the previously active tab and take note of its index.
        let selectTab = -1;
        const tabElements = DomUtils.toArray(tabbar.querySelectorAll("." + CLASS_TAB));
        for (let i = 0; i < tabElements.length; i++) {
            if (tabElements[i].classList.contains(CLASS_ACTIVE)) {
                selectTab = i;
                break;
            }
        }
        if (tabCount > 0) {
            selectTab = selectTab === -1 ? 0 : selectTab;
            this.setSelectedIndex(selectTab);
            this._showTab(selectTab);
        }
    }
    _tabClickHandler(tabElement, index) {
        // This handler may fire when a tab is removed during the click event bubble procedure. This check
        // supresses the event if the tab has been removed already.
        if (tabElement.parentNode !== null) {
            this.setSelectedIndex(index);
            DomUtils.doLater(this._sendSwitchEvent.bind(this));
        }
    }
    //-----------------------------------------------------------------------
    //
    //   ######                                
    //   #     # #    # #####  #      #  ####  
    //   #     # #    # #    # #      # #    # 
    //   ######  #    # #####  #      # #      
    //   #       #    # #    # #      # #      
    //   #       #    # #    # #      # #    # 
    //   #        ####  #####  ###### #  ####  
    //
    //-----------------------------------------------------------------------
    setSelectedIndex(index) {
        if (this._getContentsStack().children.length === 0) {
            return;
        }
        if (index < 0 || this._getContentsStack().children.length <= index) {
            this._log.warn("Out of range index given to the 'currentIndex' property.");
            return;
        }
        if (this._getContentsStack().getCurrentIndex() === index) {
            return;
        }
        this._getContentsStack().setCurrentIndex(index);
        this._showTab(index);
    }
    getSelectedIndex() {
        return this._getContentsStack().getCurrentIndex();
    }
    _getTabs() {
        return DomUtils.toArray(this.children).filter(element => element.nodeName === Tab_1.Tab.TAG_NAME);
    }
    setSelectedTab(selectTab) {
        const index = _.findIndex(this._getTabs(), tab => tab === selectTab);
        if (index === -1) {
            return;
        }
        this.setSelectedIndex(index);
    }
    getSelectedTab() {
        const currentIndex = this.getSelectedIndex();
        if (currentIndex === -1) {
            return null;
        }
        const currentTab = this._getTabs()[currentIndex];
        return currentTab;
    }
    setShowFrame(value) {
        this.setAttribute(ATTR_SHOW_FRAME, "" + value);
    }
    getShowFrame() {
        if (this.hasAttribute(ATTR_SHOW_FRAME)) {
            return Util.toBoolean(this.getAttribute(ATTR_SHOW_FRAME));
        }
        else {
            return true;
        }
    }
    setShowTabs(show) {
        if (show !== this._showTabs) {
            this._showTabs = show;
            this.update();
        }
    }
    getShowTabs() {
        return this._showTabs;
    }
    //-----------------------------------------------------------------------
    _showFrame(value) {
        if (value) {
            this._getTop().classList.add('show_frame');
        }
        else {
            this._getTop().classList.remove('show_frame');
        }
    }
    _showTab(index) {
        if (this._showTabs) {
            const tabbar = this._getTabbar();
            let tabCounter = 0;
            for (let i = 0; i < tabbar.children.length; i++) {
                const item = tabbar.children.item(i);
                if (item.classList.contains('tab')) {
                    if (tabCounter === index) {
                        item.classList.add(CLASS_ACTIVE);
                    }
                    else {
                        item.classList.remove(CLASS_ACTIVE);
                    }
                    tabCounter++;
                }
            }
        }
    }
    _sendSwitchEvent() {
        const event = new CustomEvent(TabWidget.EVENT_TAB_SWITCH, { detail: null, bubbles: true });
        this.dispatchEvent(event);
    }
    //-----------------------------------------------------------------------
    //
    //  ######                            ##       ######                       
    //  #     # #####    ##    ####      #  #      #     # #####   ####  #####  
    //  #     # #    #  #  #  #    #      ##       #     # #    # #    # #    # 
    //  #     # #    # #    # #          ###       #     # #    # #    # #    # 
    //  #     # #####  ###### #  ###    #   # #    #     # #####  #    # #####  
    //  #     # #   #  #    # #    #    #    #     #     # #   #  #    # #      
    //  ######  #    # #    #  ####      ###  #    ######  #    #  ####  #      
    //                                                                         
    //-----------------------------------------------------------------------
    _setupDragAndDrop() {
        const tabBar = this._getTabbar();
        tabBar.addEventListener("dragstart", this._handleDragStart.bind(this));
        tabBar.addEventListener("dragover", this._handleDragOver.bind(this));
        tabBar.addEventListener("dragenter", this._handleDragEnter.bind(this));
        tabBar.addEventListener("dragexit", this._removeDropIndicator.bind(this));
        tabBar.addEventListener("dragleave", this._removeDropIndicator.bind(this));
        tabBar.addEventListener("dragend", this._handleDragEnd.bind(this));
        tabBar.addEventListener("drop", this._handleDrop.bind(this));
        const snapDropContainer = DomUtils.getShadowId(this, ID_SNAP_DROP_CONTAINER);
        DomUtils.addCustomEventResender(snapDropContainer, SnapDropContainer_1.SnapDropContainer.EVENT_DROPPED, this);
    }
    _handleDragStart(ev) {
        ev.stopPropagation();
        // Only let the 'tab' elements be dragged and not our LIs.
        const target = ev.target;
        const parentElement = target.parentElement;
        if (target.getAttribute("draggable") === null || !(parentElement instanceof Tab_1.Tab)) {
            ev.preventDefault();
            return;
        }
        ev.dataTransfer.setData(InternalMimeTypes_1.ElementMimeType.MIMETYPE, InternalMimeTypes_1.ElementMimeType.elementToData(parentElement));
        ev.dataTransfer.setDragImage(parentElement, -10, -10);
        ev.dataTransfer.effectAllowed = 'move';
        ev.dataTransfer.dropEffect = 'move';
        const dragStartedEvent = new CustomEvent(GeneralEvents_1.EVENT_DRAG_STARTED, { bubbles: true });
        this.dispatchEvent(dragStartedEvent);
    }
    _handleDragEnter(ev) {
        ev.preventDefault();
    }
    _handleDragOver(ev) {
        const pointerTabIndex = this._pointToTabIndex(ev);
        this._showDropIndicator(pointerTabIndex);
        ev.preventDefault();
        ev.stopPropagation();
    }
    _pointToTabIndex(ev) {
        // Figure out which tabs the drop indicator should appear in between.
        const widgetRect = this.getBoundingClientRect();
        const pointXInPageCoords = ev.pageX - widgetRect.left;
        const tabBar = this._getTabbar();
        const childElements = DomUtils.toArray(tabBar.children).filter(kid => kid.classList.contains(CLASS_TAB));
        let index = 0;
        for (const kid of childElements) {
            const kidRect = kid.getBoundingClientRect();
            const midPoint = kidRect.left - widgetRect.left + kidRect.width / 2;
            if (pointXInPageCoords <= midPoint) {
                break;
            }
            index++;
        }
        return index;
    }
    _showDropIndicator(pointerTabIndex) {
        let dropIndicator = DomUtils.getShadowId(this, ID_DROP_INDICATOR);
        if (dropIndicator == null) {
            dropIndicator = document.createElement("li");
            dropIndicator.id = ID_DROP_INDICATOR;
        }
        const tabBar = this._getTabbar();
        const tabElements = DomUtils.toArray(tabBar.children).filter(kid => kid.classList.contains(CLASS_TAB));
        if (tabElements.length === 0) {
            if (tabBar.firstElementChild != null) {
                const position = tabBar.firstElementChild.classList.contains(CLASS_REMAINDER_LEFT) ? 'afterend' : 'beforebegin';
                tabBar.firstElementChild.insertAdjacentElement(position, dropIndicator);
            }
            else {
                tabBar.appendChild(dropIndicator);
            }
        }
        else {
            if (pointerTabIndex < tabElements.length) {
                tabElements[pointerTabIndex].insertAdjacentElement('beforebegin', dropIndicator);
            }
            else {
                tabElements[tabElements.length - 1].insertAdjacentElement('afterend', dropIndicator);
            }
        }
    }
    _handleDrop(ev) {
        this._removeDropIndicator();
        const { mimeType, data } = this._getSupportedDropMimeTypeData(ev);
        if (mimeType != null) {
            if (this.id == null || this.id === "") {
                this._log.warn("A drop occurred on a TabWidget with no ID set.");
                return;
            }
            const pointerTabIndex = this._pointToTabIndex(ev);
            const dragEndedEvent = new CustomEvent(GeneralEvents_1.EVENT_DRAG_ENDED, { bubbles: true });
            this.dispatchEvent(dragEndedEvent);
            const detail = {
                targetTabWidget: this,
                tabIndex: pointerTabIndex,
                mimeType: mimeType,
                dropData: data
            };
            const tabDropEvent = new CustomEvent(TabWidget.EVENT_DROPPED, { bubbles: true, detail: detail });
            this.dispatchEvent(tabDropEvent);
        }
    }
    _getSupportedDropMimeTypeData(ev) {
        for (const mimeType of [InternalMimeTypes_1.ElementMimeType.MIMETYPE, InternalMimeTypes_1.FrameMimeType.MIMETYPE]) {
            const data = ev.dataTransfer.getData(mimeType);
            if (data != null && data !== "") {
                return { mimeType, data };
            }
        }
        return { mimeType: null, data: null };
    }
    _removeDropIndicator() {
        const dropIndicator = DomUtils.getShadowId(this, ID_DROP_INDICATOR);
        if (dropIndicator != null) {
            dropIndicator.parentElement.removeChild(dropIndicator);
        }
    }
    _handleDragEnd(ev) {
        this._removeDropIndicator();
        const dragEndedEvent = new CustomEvent(GeneralEvents_1.EVENT_DRAG_ENDED, { bubbles: true });
        this.dispatchEvent(dragEndedEvent);
    }
}
/**
 * The HTML tag name of this element.
 */
TabWidget.TAG_NAME = "ET-TABWIDGET";
TabWidget.EVENT_TAB_SWITCH = "tab-switch";
TabWidget.EVENT_DROPPED = "tabwidget-dropped";
exports.TabWidget = TabWidget;
//# sourceMappingURL=TabWidget.js.map