/**
 * Copyright 2016 Simon Edwards <simon@simonzone.com>
 */
"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const _ = require("lodash");
const fs = require("fs");
const ViewerElement_1 = require("../ViewerElement");
const ThemeableElementBase_1 = require("../ThemeableElementBase");
const DomUtils = require("../DomUtils");
const ThemeTypes = require("../Theme");
const CodeMirror = require("codemirror");
const CodeMirrorCommands = require("../CodeMirrorCommands");
const CodeMirrorUtils = require("../utils/CodeMirrorUtils");
const ViewerElementTypes = require("../ViewerElementTypes");
const ResizeRefreshElementBase = require("../ResizeRefreshElementBase");
const CommandPaletteRequestTypes_1 = require("../CommandPaletteRequestTypes");
const VirtualScrollArea = require("../VirtualScrollArea");
const Logger_1 = require("../Logger");
const GeneralEvents = require("../GeneralEvents");
const VisualState = ViewerElementTypes.VisualState;
const ID = "EtTextViewerTemplate";
const ID_CONTAINER = "ID_CONTAINER";
const ID_MAIN_STYLE = "ID_MAIN_STYLE";
const CLASS_HIDE_CURSOR = "hide_cursor";
const CLASS_FOCUSED = "terminal-focused";
const CLASS_UNFOCUSED = "terminal-unfocused";
const KEYBINDINGS_CURSOR_MODE = "text-viewer";
const PALETTE_GROUP = "textviewer";
const COMMAND_TYPE_AND_CR_SELECTION = "typeSelectionAndCr";
const COMMAND_TYPE_SELECTION = "typeSelection";
const COMMANDS = [
    COMMAND_TYPE_AND_CR_SELECTION,
    COMMAND_TYPE_SELECTION,
    CommandPaletteRequestTypes_1.COMMAND_OPEN_COMMAND_PALETTE
];
const NO_STYLE_HACK = "NO_STYLE_HACK";
const DEBUG_RESIZE = false;
let registered = false;
CodeMirrorCommands.init();
let cssText = null;
function getCssText() {
    return cssText;
}
const simpleScrollBars = require('codemirror/addon/scroll/simplescrollbars');
// CodeMirror mode management.
const meta = require('codemirror/mode/meta');
const loadedCodeMirrorModes = new Set();
function LoadCodeMirrorMode(modeName) {
    if (loadedCodeMirrorModes.has(modeName)) {
        return;
    }
    require('codemirror/mode/' + modeName + '/' + modeName);
    loadedCodeMirrorModes.add(modeName);
}
class TextViewer extends ViewerElement_1.ViewerElement {
    //-----------------------------------------------------------------------
    //
    //   #                                                         
    //   #       # ###### ######  ####  #   #  ####  #      ###### 
    //   #       # #      #      #    #  # #  #    # #      #      
    //   #       # #####  #####  #        #   #      #      #####  
    //   #       # #      #      #        #   #      #      #      
    //   #       # #      #      #    #   #   #    # #      #      
    //   ####### # #      ######  ####    #    ####  ###### ###### 
    //
    //-----------------------------------------------------------------------
    constructor() {
        super();
        this._initProperties();
    }
    static init() {
        if (registered === false) {
            // Load the CSS resources now.
            cssText = fs.readFileSync(require.resolve('codemirror/lib/codemirror.css'), { encoding: 'utf8' })
                + fs.readFileSync(require.resolve('codemirror/addon/scroll/simplescrollbars.css'), { encoding: 'utf8' });
            window.customElements.define(TextViewer.TAG_NAME.toLowerCase(), TextViewer);
            registered = true;
        }
    }
    /**
     * Type guard for detecting a EtTerminalViewer instance.
     *
     * @param  node the node to test
     * @return      True if the node is a EtTerminalViewer.
     */
    static is(node) {
        return node !== null && node !== undefined && node instanceof TextViewer;
    }
    _initProperties() {
        this._keyBindingManager = null;
        this._text = null;
        this._mimeType = null;
        this._log = Logger_1.getLogger(TextViewer.TAG_NAME, this);
        this._commandLine = null;
        this._returnCode = null;
        this._editable = false;
        this._codeMirror = null;
        this._height = 0;
        this._isEmpty = false;
        this._mode = ViewerElementTypes.Mode.DEFAULT;
        this.document = document;
        this._visualState = VisualState.AUTO;
        this._currentElementHeight = -1;
        this._mainStyleLoaded = false;
        this._resizePollHandle = null;
        this._lastCursorAnchorPosition = null;
        this._lastCursorHeadPosition = null;
        this._viewportHeight = -1;
    }
    //-----------------------------------------------------------------------
    //
    // ######                                
    // #     # #    # #####  #      #  ####  
    // #     # #    # #    # #      # #    # 
    // ######  #    # #####  #      # #      
    // #       #    # #    # #      # #      
    // #       #    # #    # #      # #    # 
    // #        ####  #####  ###### #  ####  
    //
    //-----------------------------------------------------------------------
    setKeyBindingManager(newKeyBindingManager) {
        this._keyBindingManager = newKeyBindingManager;
    }
    setCommandLine(commandLine) {
        this._commandLine = commandLine;
    }
    setReturnCode(returnCode) {
        this._returnCode = returnCode;
    }
    getTitle() {
        if (this._commandLine !== null) {
            return this._commandLine;
        }
        else {
            return "Text";
        }
    }
    getAwesomeIcon() {
        return "file-text-o";
    }
    getSelectionText() {
        const doc = this._codeMirror.getDoc();
        const cursorAnchorPos = doc.getCursor("anchor");
        const cursorHeadPos = doc.getCursor("head");
        if (_.isEqual(cursorHeadPos, cursorAnchorPos)) {
            return null;
        }
        return doc.getSelection("\n");
    }
    focus() {
        this._codeMirror.focus();
    }
    hasFocus() {
        const hasFocus = this._codeMirror.getInputField() === DomUtils.getShadowRoot(this).activeElement;
        return hasFocus;
    }
    setVisualState(newVisualState) {
        if (newVisualState !== this._visualState) {
            if (DomUtils.getShadowRoot(this) !== null) {
                this._applyVisualState(newVisualState);
            }
            this._visualState = newVisualState;
        }
    }
    getVisualState() {
        return this._visualState;
    }
    getText() {
        return this._codeMirror.getDoc().getValue();
    }
    setText(newText) {
        if (this._codeMirror === null) {
            this._text = newText;
        }
        else {
            this._codeMirror.getDoc().setValue(newText);
        }
    }
    // From SupportsClipboardPaste interface.
    canPaste() {
        return this._mode === ViewerElementTypes.Mode.CURSOR;
    }
    // From SupportsClipboardPaste interface.
    pasteText(text) {
        if (!this.canPaste()) {
            return;
        }
        CodeMirrorUtils.pasteText(this._codeMirror.getDoc(), text);
    }
    setMimeType(mimeType) {
        this._mimeType = mimeType;
        const modeInfo = CodeMirror.findModeByMIME(mimeType);
        if (modeInfo != null) {
            if (modeInfo.mode !== null && modeInfo.mode !== "null") {
                LoadCodeMirrorMode(modeInfo.mode);
            }
            if (this._codeMirror !== null) {
                this._codeMirror.setOption("mode", mimeType);
            }
        }
    }
    getMimeType() {
        return this._mimeType;
    }
    getTabSize() {
        return parseInt(this._codeMirror.getOption("tabSize"), 10);
    }
    setTabSize(size) {
        this._codeMirror.setOption("tabSize", size);
    }
    setBytes(buffer, mimeType) {
        let charset = "utf-8";
        let cleanMimeType = mimeType;
        if (mimeType.indexOf(';') !== -1) {
            charset = mimeType.slice(mimeType.indexOf(';') + 1);
            cleanMimeType = mimeType.slice(0, mimeType.indexOf(';'));
        }
        const decodedString = buffer.toString(charset);
        this.setText(decodedString);
        this.setMimeType(cleanMimeType);
    }
    setMode(newMode) {
        if (newMode !== this._mode) {
            switch (newMode) {
                case ViewerElementTypes.Mode.CURSOR:
                    // Enter cursor mode.
                    this._enterCursorMode();
                    break;
                case ViewerElementTypes.Mode.DEFAULT:
                    this._exitCursorMode();
                    break;
            }
            this._mode = newMode;
        }
    }
    getMode() {
        return this._mode;
    }
    setEditable(editable) {
        this._editable = editable;
        if (this._mode === ViewerElementTypes.Mode.CURSOR) {
            this._codeMirror.setOption("readOnly", !editable);
        }
    }
    getEditable() {
        return this._editable;
    }
    /**
     * Gets the height of this element.
     *
     * @return {number} [description]
     */
    getHeight() {
        return this._height;
    }
    // VirtualScrollable
    getMinHeight() {
        return 0;
    }
    // VirtualScrollable
    getVirtualHeight(containerHeight) {
        const result = this.getVirtualTextHeight();
        if (DEBUG_RESIZE) {
            this._log.debug("getVirtualHeight: ", result);
        }
        return result;
    }
    // VirtualScrollable
    getReserveViewportHeight(containerHeight) {
        if (DEBUG_RESIZE) {
            this._log.debug("getReserveViewportHeight: ", 0);
        }
        return 0;
    }
    // VirtualScrollable
    setDimensionsAndScroll(setterState) {
        if (setterState.heightChanged || setterState.yOffsetChanged) {
            if (DEBUG_RESIZE) {
                this._log.debug(`setDimensionsAndScroll(): ` +
                    `height=${setterState.height}, heightChanged=${setterState.heightChanged}, ` +
                    `yOffset=${setterState.yOffset}, yOffsetChanged=${setterState.yOffsetChanged}, ` +
                    `physicalTop=${setterState.physicalTop}, physicalTopChanged=${setterState.physicalTopChanged}, ` +
                    `containerHeight=${setterState.containerHeight}, ` +
                    `containerHeightChanged=${setterState.containerHeightChanged}, ` +
                    `visibleBottomOffset=${setterState.visibleBottomOffset}, ` +
                    `visibleBottomOffsetChanged=${setterState.visibleBottomOffsetChanged}`);
            }
            // FIXME the commented code makes it go faster but breaks the pop-out frame function and hangs the whole app.
            // const op = () => {
            this._adjustHeight(setterState.height);
            this.scrollTo(0, setterState.yOffset);
            // };
            // if (this._codeMirror !== null) {
            //   this._codeMirror.operation(op);
            // } else {
            //   op();
            // }
        }
        if (setterState.visibleBottomOffsetChanged) {
            const shadowRoot = DomUtils.getShadowRoot(this);
            if (shadowRoot !== null) {
                const horizontalScrollbar = shadowRoot.querySelector("DIV.CodeMirror-overlayscroll-horizontal");
                const offsetFromBottom = Math.max(0, -1 * setterState.visibleBottomOffset);
                horizontalScrollbar.style.bottom = "" + offsetFromBottom + "px";
            }
        }
    }
    isFontLoaded() {
        return this._effectiveFontFamily().indexOf(NO_STYLE_HACK) === -1;
    }
    lineCount() {
        const doc = this._codeMirror.getDoc();
        return this._isEmpty ? 0 : doc.lineCount();
    }
    getCursorPosition() {
        const cursorPos = this._codeMirror.cursorCoords(true, "local");
        const scrollInfo = this._codeMirror.getScrollInfo();
        const detail = {
            left: cursorPos.left,
            top: cursorPos.top,
            bottom: cursorPos.bottom,
            viewPortTop: scrollInfo.top
        };
        return detail;
    }
    clearSelection() {
        const doc = this._codeMirror.getDoc();
        if (!doc.somethingSelected()) {
            return;
        }
        doc.setCursor(doc.getCursor());
    }
    setCursorPositionTop(ch) {
        const doc = this._codeMirror.getDoc();
        doc.setCursor({ line: 0, ch: ch });
        return true;
    }
    setCursorPositionBottom(ch) {
        const doc = this._codeMirror.getDoc();
        doc.setCursor({ line: doc.lineCount() - 1, ch: ch });
        return true;
    }
    // From viewerelementtypes.SupportsMimeTypes
    static supportsMimeType(mimeType) {
        const mode = CodeMirror.findModeByMIME(mimeType);
        return mode !== null && mode !== undefined;
    }
    deleteTopPixels(topPixels) {
        const defaultTextHeight = this._codeMirror.defaultTextHeight();
        const linesToDelete = Math.min(Math.floor(topPixels / defaultTextHeight), this.lineCount());
        const pos = { line: 0, ch: 0 };
        const endPos = { line: linesToDelete, ch: 0 };
        this._codeMirror.getDoc().replaceRange("", pos, endPos);
        this._emitVirtualResizeEvent();
    }
    connectedCallback() {
        super.connectedCallback();
        if (DomUtils.getShadowRoot(this) !== null) {
            return;
        }
        const shadow = this.attachShadow({ mode: 'open', delegatesFocus: true });
        const clone = this.createClone();
        shadow.appendChild(clone);
        this._initFontLoading();
        this.installThemeCss();
        const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
        this.style.height = "0px";
        this._exitCursorMode();
        this._mode = ViewerElementTypes.Mode.DEFAULT;
        const codeMirrorOptions = {
            value: "",
            readOnly: true,
            lineNumbers: true,
            scrollbarStyle: "overlay",
            cursorScrollMargin: 0,
            showCursorWhenSelecting: true,
            theme: "text",
            keyMap: this._codeMirrorKeyMap()
        };
        if (this._mimeType !== null) {
            codeMirrorOptions.mode = this._mimeType;
        }
        // Create the CodeMirror instance
        this._codeMirror = CodeMirror((el) => {
            containerDiv.appendChild(el);
        }, codeMirrorOptions);
        this._codeMirror.on("cursorActivity", () => {
            const effectiveFocus = this._visualState === ViewerElementTypes.VisualState.FOCUSED ||
                (this._visualState === ViewerElementTypes.VisualState.AUTO && this.hasFocus());
            if (this._mode !== ViewerElementTypes.Mode.DEFAULT && effectiveFocus) {
                this._lastCursorHeadPosition = this._codeMirror.getDoc().getCursor("head");
                this._lastCursorAnchorPosition = this._codeMirror.getDoc().getCursor("anchor");
                const event = new CustomEvent(ViewerElement_1.ViewerElement.EVENT_CURSOR_MOVE, { bubbles: true });
                this.dispatchEvent(event);
            }
        });
        this._codeMirror.on("scroll", () => {
            // Over-scroll bug/feature fix
            const scrollInfo = this._codeMirror.getScrollInfo();
            // this._log.debug("codemirror event scroll:", scrollInfo);
            const clientYScrollRange = this._getClientYScrollRange();
            if (scrollInfo.top > clientYScrollRange) {
                this._codeMirror.scrollTo(0, clientYScrollRange);
            }
        });
        this._codeMirror.on("focus", (instance) => {
            if (this._visualState === VisualState.AUTO) {
                const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
                containerDiv.classList.add(CLASS_FOCUSED);
                containerDiv.classList.remove(CLASS_UNFOCUSED);
            }
        });
        this._codeMirror.on("blur", (instance) => {
            if (this._visualState === VisualState.AUTO) {
                containerDiv.classList.add(CLASS_UNFOCUSED);
                containerDiv.classList.remove(CLASS_FOCUSED);
            }
        });
        this._codeMirror.on("beforeSelectionChange", (instance, obj) => {
            // obj: { ranges: {anchor: CodeMirror.Position; head: CodeMirror.Position; }[]; }
            if (obj.ranges.length === 0) {
                return;
            }
            if (obj.ranges.length === 1) {
                const pair = obj.ranges[0];
                if (_.isEqual(pair.anchor, pair.head)) {
                    return;
                }
            }
            this._emitBeforeSelectionChangeEvent(obj.origin === "*mouse");
        });
        this._codeMirror.on("keyHandled", (instance, name, event) => {
            const isUp = name === "PageUp" || name === "Up";
            const isDown = name === "PageDown" || name === "Down";
            if (isUp || isDown) {
                const cursorAnchorPos = this._codeMirror.getDoc().getCursor("anchor");
                const cursorHeadPos = this._codeMirror.getDoc().getCursor("head");
                if (this._lastCursorHeadPosition !== null && this._lastCursorAnchorPosition !== null
                    && _.isEqual(this._lastCursorHeadPosition, this._lastCursorAnchorPosition) // check for no selection
                    && _.isEqual(cursorHeadPos, cursorAnchorPos) // check for no selection
                    && this._lastCursorHeadPosition.line === cursorHeadPos.line) {
                    // The last action didn't move the cursor.
                    const ch = this._lastCursorAnchorPosition.ch; // _lastCursorAnchorPosition can change before the code below runs.
                    DomUtils.doLater(() => {
                        const detail = { edge: isUp
                                ? ViewerElementTypes.Edge.TOP
                                : ViewerElementTypes.Edge.BOTTOM,
                            ch: ch };
                        const event = new CustomEvent(ViewerElement_1.ViewerElement.EVENT_CURSOR_EDGE, { bubbles: true, detail: detail });
                        this.dispatchEvent(event);
                    });
                }
            }
        });
        this._codeMirror.on('viewportChange', (instance, from, to) => {
            if (this._mode !== ViewerElementTypes.Mode.CURSOR) {
                return;
            }
            const height = to - from;
            if (height !== this._viewportHeight) {
                this._viewportHeight = height;
                DomUtils.doLater(this._emitVirtualResizeEvent.bind(this));
            }
        });
        // Filter the keyboard events before they reach CodeMirror.
        containerDiv.addEventListener('keydown', this._handleContainerKeyDownCapture.bind(this), true);
        containerDiv.addEventListener('keydown', this._handleContainerKeyDown.bind(this));
        containerDiv.addEventListener('keyup', this._handleContainerKeyUpCapture.bind(this), true);
        containerDiv.addEventListener('contextmenu', this._handleContextMenuCapture.bind(this), true);
        const codeMirrorElement = this._codeMirror.getWrapperElement();
        this._codeMirror.on("scrollCursorIntoView", (instance, ev) => {
            ev.preventDefault();
        });
        this._applyVisualState(this._visualState);
        if (this._text !== null) {
            this._codeMirror.getDoc().setValue(this._text);
            DomUtils.doLater(this._emitVirtualResizeEvent.bind(this));
            this._text = null;
        }
        this._adjustHeight(this._height);
    }
    _themeCssFiles() {
        return [ThemeTypes.CssFile.TEXT_VIEWER];
    }
    refresh(level) {
        if (this._codeMirror !== null) {
            if (DEBUG_RESIZE) {
                this._log.debug("calling codeMirror.refresh()");
            }
            if (level === ResizeRefreshElementBase.RefreshLevel.RESIZE) {
                this._codeMirror.setSize(null, null);
            }
            else {
                this._codeMirror.refresh();
            }
        }
        VirtualScrollArea.emitResizeEvent(this);
    }
    //-----------------------------------------------------------------------
    //
    // ######                                      
    // #     # #####  # #    #   ##   ##### ###### 
    // #     # #    # # #    #  #  #    #   #      
    // ######  #    # # #    # #    #   #   #####  
    // #       #####  # #    # ######   #   #      
    // #       #   #  #  #  #  #    #   #   #      
    // #       #    # #   ##   #    #   #   ###### 
    //
    //-----------------------------------------------------------------------
    /**
     *
     */
    createClone() {
        let template = window.document.getElementById(ID);
        if (template === null) {
            template = window.document.createElement('template');
            template.id = ID;
            template.innerHTML = `<style id="${ID_MAIN_STYLE}">

        /* The idea is that this rule will be quickly applied. We can then monitor
           the computed style to see when the proper theme font is applied and
           NO_STYLE_HACK disappears from the reported computed style. */
        .terminal {
          font-family: sans-serif, ${NO_STYLE_HACK};
        }

        ${getCssText()}
        </style>
        <style id="${ThemeableElementBase_1.ThemeableElementBase.ID_THEME}"></style>
        <div id="${ID_CONTAINER}" class="terminal_viewer ${CLASS_UNFOCUSED}"></div>`;
            window.document.body.appendChild(template);
        }
        return window.document.importNode(template.content, true);
    }
    _applyVisualState(visualState) {
        const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
        if ((visualState === VisualState.AUTO && this.hasFocus()) ||
            visualState === VisualState.FOCUSED) {
            containerDiv.classList.add(CLASS_FOCUSED);
            containerDiv.classList.remove(CLASS_UNFOCUSED);
        }
        else {
            containerDiv.classList.add(CLASS_UNFOCUSED);
            containerDiv.classList.remove(CLASS_FOCUSED);
        }
    }
    _enterCursorMode() {
        const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
        containerDiv.classList.remove(CLASS_HIDE_CURSOR);
        const doc = this._codeMirror.getDoc();
        doc.setCursor({ line: doc.lineCount() - 1, ch: 0 });
        this._lastCursorHeadPosition = this._codeMirror.getDoc().getCursor("head");
        this._lastCursorAnchorPosition = this._codeMirror.getDoc().getCursor("anchor");
        if (this._editable) {
            this._codeMirror.setOption("readOnly", false);
        }
    }
    _exitCursorMode() {
        if (this._codeMirror !== null) {
            this._codeMirror.setOption("readOnly", true);
        }
        const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
        containerDiv.classList.add(CLASS_HIDE_CURSOR);
    }
    _emitVirtualResizeEvent() {
        if (DEBUG_RESIZE) {
            this._log.debug("_emitVirtualResizeEvent");
        }
        VirtualScrollArea.emitResizeEvent(this);
    }
    _emitBeforeSelectionChangeEvent(originMouse) {
        const event = new CustomEvent(ViewerElement_1.ViewerElement.EVENT_BEFORE_SELECTION_CHANGE, { detail: { originMouse: originMouse },
            bubbles: true });
        this.dispatchEvent(event);
    }
    _getMimeTypeName() {
        for (const info of CodeMirror.modeInfo) {
            if (info.mime === this._mimeType) {
                return info.name;
            }
        }
        return this._mimeType;
    }
    scrollTo(optionsOrX, y) {
        let xCoord = 0;
        let yCoord = 0;
        if (typeof optionsOrX === "number") {
            xCoord = optionsOrX;
            if (y !== undefined) {
                yCoord = y;
            }
        }
        else {
            xCoord = optionsOrX.left;
            yCoord = optionsOrX.top;
        }
        if (DomUtils.getShadowRoot(this) === null) {
            return;
        }
        this._codeMirror.scrollTo(xCoord, yCoord);
    }
    // ----------------------------------------------------------------------
    //
    //   #    #                                                 
    //   #   #  ###### #   # #####   ####    ##   #####  #####  
    //   #  #   #       # #  #    # #    #  #  #  #    # #    # 
    //   ###    #####    #   #####  #    # #    # #    # #    # 
    //   #  #   #        #   #    # #    # ###### #####  #    # 
    //   #   #  #        #   #    # #    # #    # #   #  #    # 
    //   #    # ######   #   #####   ####  #    # #    # #####  
    //                                                        
    // ----------------------------------------------------------------------
    _codeMirrorKeyMap() {
        if (this._keyBindingManager === null || this._keyBindingManager.getKeyBindingContexts() === null) {
            return {}; // empty keymap
        }
        const keyBindings = this._keyBindingManager.getKeyBindingContexts().context(KEYBINDINGS_CURSOR_MODE);
        if (keyBindings === null) {
            return {};
        }
        const codeMirrorKeyMap = keyBindings.keyBindings
            .filter((binding) => COMMANDS.indexOf(binding.command) === -1)
            .reduce((accu, binding) => {
            accu[binding.normalizedShortcut] = binding.command;
            return accu;
        }, {});
        return codeMirrorKeyMap;
    }
    dispatchEvent(ev) {
        if (ev.type === 'keydown' || ev.type === 'keypress') {
            const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
            return containerDiv.dispatchEvent(ev);
        }
        else {
            return super.dispatchEvent(ev);
        }
    }
    _scheduleSyntheticKeyDown(ev) {
        DomUtils.doLater(() => {
            const fakeKeyDownEvent = DomUtils.newKeyboardEvent('keydown', {
                bubbles: true,
                key: ev.key,
                code: ev.code,
                location: ev.location,
                repeat: ev.repeat,
                keyCode: ev.keyCode,
                charCode: ev.charCode,
                which: ev.which,
                ctrlKey: ev.ctrlKey,
                shiftKey: ev.shiftKey,
                altKey: ev.altKey,
                metaKey: ev.metaKey
            });
            super.dispatchEvent(fakeKeyDownEvent);
        });
    }
    _handleContainerKeyDown(ev) {
        if (this._mode !== ViewerElementTypes.Mode.DEFAULT) {
            ev.stopPropagation();
        }
    }
    _handleContainerKeyDownCapture(ev) {
        let command = null;
        if (this._keyBindingManager !== null && this._keyBindingManager.getKeyBindingContexts() !== null &&
            this._mode === ViewerElementTypes.Mode.CURSOR) {
            const keyBindings = this._keyBindingManager.getKeyBindingContexts().context(KEYBINDINGS_CURSOR_MODE);
            if (keyBindings !== null) {
                command = keyBindings.mapEventToCommand(ev);
                if (this._executeCommand(command)) {
                    ev.stopPropagation();
                    return;
                }
                else {
                    if (command !== null) {
                        return;
                    }
                    if (ev.shiftKey) {
                        const evWithoutShift = {
                            shiftKey: false,
                            metaKey: ev.metaKey,
                            altKey: ev.altKey,
                            ctrlKey: ev.ctrlKey,
                            key: ev.key,
                            keyCode: ev.keyCode
                        };
                        command = keyBindings.mapEventToCommand(evWithoutShift);
                        if (command !== null && command.startsWith("go")) {
                            // CodeMirror will handle this key.
                            return;
                        }
                    }
                }
            }
        }
        if (this._mode === ViewerElementTypes.Mode.DEFAULT) {
            ev.stopPropagation();
            // Emit a key down event which our parent elements can catch.
            this._scheduleSyntheticKeyDown(ev);
        }
    }
    _handleContainerKeyUpCapture(ev) {
        if (this._mode === ViewerElementTypes.Mode.DEFAULT) {
            ev.stopPropagation();
            ev.preventDefault();
        }
    }
    _handleContextMenuCapture(ev) {
        // Prevent CodeMirror from seeing this event and messing with the hidden textarea and the focus.
        ev.stopImmediatePropagation();
        ev.preventDefault();
        this.executeCommand(CommandPaletteRequestTypes_1.COMMAND_OPEN_COMMAND_PALETTE);
    }
    getCommandPaletteEntries(commandableStack) {
        let commandList = [
            { id: COMMAND_TYPE_SELECTION, group: PALETTE_GROUP, iconRight: "terminal", label: "Type Selection", commandExecutor: this },
            { id: COMMAND_TYPE_AND_CR_SELECTION, group: PALETTE_GROUP, iconRight: "terminal", label: "Type Selection & Execute", commandExecutor: this }
        ];
        if (this._mode === ViewerElementTypes.Mode.CURSOR) {
            const cmCommandList = CodeMirrorCommands.commandDescriptions(this._codeMirror).map((desc) => {
                return { id: desc.command,
                    group: PALETTE_GROUP,
                    iconLeft: desc.iconLeft,
                    iconRight: desc.iconRight,
                    label: desc.label,
                    commandExecutor: this };
            });
            commandList = [...commandList, ...cmCommandList];
        }
        const keyBindings = this._keyBindingManager.getKeyBindingContexts().context(KEYBINDINGS_CURSOR_MODE);
        if (keyBindings !== null) {
            commandList.forEach((commandEntry) => {
                const shortcut = keyBindings.mapCommandToKeyBinding(commandEntry.id);
                commandEntry.shortcut = shortcut === null ? "" : shortcut;
            });
        }
        return commandList;
    }
    executeCommand(commandId) {
        this._executeCommand(commandId);
    }
    _executeCommand(command) {
        switch (command) {
            case COMMAND_TYPE_AND_CR_SELECTION:
            case COMMAND_TYPE_SELECTION:
                const text = this._codeMirror.getDoc().getSelection();
                if (text !== "") {
                    if (command === COMMAND_TYPE_AND_CR_SELECTION) {
                        // Exit cursor mode.
                        const setModeDetail = { mode: ViewerElementTypes.Mode.DEFAULT };
                        const setModeEvent = new CustomEvent(GeneralEvents.EVENT_SET_MODE, { detail: setModeDetail });
                        setModeEvent.initCustomEvent(GeneralEvents.EVENT_SET_MODE, true, true, setModeDetail);
                        this.dispatchEvent(setModeEvent);
                    }
                    const typeTextDetail = { text: text + (command === COMMAND_TYPE_AND_CR_SELECTION ? "\n" : "") };
                    const typeTextEvent = new CustomEvent(GeneralEvents.EVENT_TYPE_TEXT, { detail: typeTextDetail });
                    typeTextEvent.initCustomEvent(GeneralEvents.EVENT_TYPE_TEXT, true, true, typeTextDetail);
                    this.dispatchEvent(typeTextEvent);
                }
                break;
            case CommandPaletteRequestTypes_1.COMMAND_OPEN_COMMAND_PALETTE:
                CommandPaletteRequestTypes_1.dispatchCommandPaletteRequest(this);
                break;
            default:
                if (this._mode === ViewerElementTypes.Mode.CURSOR && CodeMirrorCommands.isCommand(command)) {
                    CodeMirrorCommands.executeCommand(this._codeMirror, command);
                    return true;
                }
                else {
                    return false;
                }
        }
        return true;
    }
    //-----------------------------------------------------------------------
    //
    // #######                        #                                            
    // #        ####  #    # #####    #        ####    ##   #####  # #    #  ####  
    // #       #    # ##   #   #      #       #    #  #  #  #    # # ##   # #    # 
    // #####   #    # # #  #   #      #       #    # #    # #    # # # #  # #      
    // #       #    # #  # #   #      #       #    # ###### #    # # #  # # #  ### 
    // #       #    # #   ##   #      #       #    # #    # #    # # #   ## #    # 
    // #        ####  #    #   #      #######  ####  #    # #####  # #    #  ####  
    //
    //-----------------------------------------------------------------------
    _initFontLoading() {
        this._mainStyleLoaded = false;
        DomUtils.getShadowId(this, ID_MAIN_STYLE).addEventListener('load', () => {
            this._mainStyleLoaded = true;
            this._handleStyleLoad();
        });
    }
    _cleanUpFontLoading() {
        if (this._resizePollHandle !== null) {
            this._resizePollHandle.cancel();
            this._resizePollHandle = null;
        }
    }
    _handleStyleLoad() {
        if (this._mainStyleLoaded) {
            // Start polling the term for application of the font.
            this._resizePollHandle = DomUtils.doLaterFrame(this._resizePoll.bind(this));
        }
    }
    _effectiveFontFamily() {
        const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
        const cs = window.getComputedStyle(containerDiv, null);
        return cs.getPropertyValue("font-family");
    }
    _resizePoll() {
        if (this._mainStyleLoaded) {
            if (!this.isFontLoaded()) {
                // Font has not been correctly applied yet.
                this._resizePollHandle = DomUtils.doLaterFrame(this._resizePoll.bind(this));
            }
            else {
                // Yay! the font is correct. Resize the term soon.
                // FIXME do we need to do anything here?
            }
        }
    }
    getVirtualTextHeight() {
        if (DomUtils.getShadowRoot(this) === null) {
            return 8;
        }
        return this._isEmpty ? 0 : this._codeMirror.defaultTextHeight() * this.lineCount();
    }
    _getClientYScrollRange() {
        return Math.max(0, this.getVirtualHeight(this.getHeight()) - this.getHeight() + this.getReserveViewportHeight(this.getHeight()));
    }
    _adjustHeight(newHeight) {
        this._height = newHeight;
        if (this.parentNode === null || DomUtils.getShadowRoot(this) === null) {
            return;
        }
        const elementHeight = this.getHeight();
        if (elementHeight !== this._currentElementHeight) {
            this._currentElementHeight = elementHeight;
            this.style.height = "" + elementHeight + "px";
            const totalTextHeight = this.getVirtualTextHeight();
            let codeMirrorHeight;
            codeMirrorHeight = elementHeight;
            const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
            containerDiv.style.height = "" + codeMirrorHeight + "px";
            this._codeMirror.setSize("100%", "" + codeMirrorHeight + "px");
        }
    }
    _themeCssSet() {
        // const themeTag = <HTMLStyleElement> util.getShadowId(this, ID_THEME_STYLE);
        // if (themeTag !== null) {
        //   themeTag.innerHTML = this.getThemeCss();
        // }
    }
}
TextViewer.TAG_NAME = "ET-TEXT-VIEWER";
exports.TextViewer = TextViewer;
function px(value) {
    if (value === null || value === undefined || value === "") {
        return 0;
    }
    return parseInt(value.slice(0, -2), 10);
}
//# sourceMappingURL=TextViewer.js.map