/**
 * Copyright 2016 Simon Edwards <simon@simonzone.com>
 */
"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const Electron = require("electron");
const shell = Electron.shell;
const _ = require("lodash");
const fs = require("fs");
const path = require("path");
const he = require("he");
const SourceDir = require("../SourceDir");
const ViewerElement_1 = require("../ViewerElement");
const ResizeRefreshElementBase = require("../ResizeRefreshElementBase");
const ThemeableElementBase_1 = require("../ThemeableElementBase");
const ThemeTypes = require("../Theme");
const DomUtils = require("../DomUtils");
const ViewerElementTypes = require("../ViewerElementTypes");
const VirtualScrollArea = require("../VirtualScrollArea");
const Logger_1 = require("../Logger");
const VisualState = ViewerElementTypes.VisualState;
const ID = "EtTipViewerTemplate";
const ID_CONTAINER = "ID_CONTAINER";
const ID_CONTENT = "ID_CONTENT";
const ID_CONTROLS = "ID_CONTROLS";
const ID_PREVIOUS_BUTTON = "ID_PREVIOUS_BUTTON";
const ID_NEXT_BUTTON = "ID_NEXT_BUTTON";
const ID_SHOW_TIPS = "ID_SHOW_TIPS";
const CLASS_KEYCAP = "CLASS_KEYCAP";
const KEYBINDINGS_SELECTION_MODE = "image-viewer";
const DEBUG_SIZE = false;
let registered = false;
let instanceIdCounter = 0;
/**
 * Load in the array of tips from src/tips/tips.html.
 *
 * Each separate tip in the HTML file must be enclosed by an article tag.
 * @return the array of tip HTMLs.
 */
function loadTipFile() {
    const tipPath = path.join(SourceDir.path, "tips/tips.html");
    const tipHtml = fs.readFileSync(tipPath, { encoding: 'utf8' });
    const parts = tipHtml.split(/<article>([^]*?)<\/article>/m);
    return parts.filter((p, i) => {
        return i % 2 === 1;
    });
}
const tipData = loadTipFile();
class TipViewer extends ViewerElement_1.ViewerElement {
    //-----------------------------------------------------------------------
    //
    //   #                                                         
    //   #       # ###### ######  ####  #   #  ####  #      ###### 
    //   #       # #      #      #    #  # #  #    # #      #      
    //   #       # #####  #####  #        #   #      #      #####  
    //   #       # #      #      #        #   #      #      #      
    //   #       # #      #      #    #   #   #    # #      #      
    //   ####### # #      ######  ####    #    ####  ###### ###### 
    //
    //-----------------------------------------------------------------------
    constructor() {
        super();
        this._initProperties();
    }
    static init() {
        if (registered === false) {
            window.customElements.define(TipViewer.TAG_NAME.toLowerCase(), TipViewer);
            registered = true;
        }
    }
    /**
     * Type guard for detecting a EtTipViewer instance.
     *
     * @param  node the node to test
     * @return      True if the node is a EtTipViewer.
     */
    static is(node) {
        return node !== null && node !== undefined && node instanceof TipViewer;
    }
    _initProperties() {
        this._log = Logger_1.getLogger(TipViewer.TAG_NAME, this);
        this._configManager = null;
        this._keyBindingManager = null;
        this._height = 0;
        this._tipIndex = 0;
    }
    //-----------------------------------------------------------------------
    //
    // ######                                
    // #     # #    # #####  #      #  ####  
    // #     # #    # #    # #      # #    # 
    // ######  #    # #####  #      # #      
    // #       #    # #    # #      # #      
    // #       #    # #    # #      # #    # 
    // #        ####  #####  ###### #  ####  
    //
    //-----------------------------------------------------------------------
    setConfigManager(newConfigManager) {
        if (this._configManager !== null) {
            this._configManager.unregisterChangeListener(this);
        }
        this._configManager = newConfigManager;
        if (this._configManager !== null) {
            this._configManager.registerChangeListener(this, this._configChanged.bind(this));
        }
    }
    setKeyBindingManager(newKeyBindingManager) {
        if (this._keyBindingManager !== null) {
            this._keyBindingManager.unregisterChangeListener(this);
        }
        this._keyBindingManager = newKeyBindingManager;
        if (this._keyBindingManager !== null) {
            this._keyBindingManager.registerChangeListener(this, this._keyBindingChanged.bind(this));
        }
    }
    getTitle() {
        return "Tip";
    }
    getAwesomeIcon() {
        return "idea";
    }
    getSelectionText() {
        return null;
    }
    focus() {
        if (DomUtils.getShadowRoot(this) === null) {
            return;
        }
        const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
        DomUtils.focusWithoutScroll(containerDiv);
    }
    hasFocus() {
        const hasFocus = this === DomUtils.getShadowRoot(this).activeElement; // FIXME
        return hasFocus;
    }
    // VirtualScrollable
    getHeight() {
        return this._height;
    }
    // VirtualScrollable
    setDimensionsAndScroll(setterState) {
        if (DEBUG_SIZE) {
            this._log.debug("setDimensionsAndScroll(): ", setterState.height, setterState.heightChanged, setterState.yOffset, setterState.yOffsetChanged);
        }
    }
    // VirtualScrollable
    getMinHeight() {
        return this._height;
    }
    // VirtualScrollable
    getVirtualHeight(containerHeight) {
        if (DEBUG_SIZE) {
            this._log.debug("getVirtualHeight: ", this._height);
        }
        return this._height;
    }
    // VirtualScrollable
    getReserveViewportHeight(containerHeight) {
        if (DEBUG_SIZE) {
            this._log.debug("getReserveViewportHeight: ", 0);
        }
        return 0;
    }
    // From viewerelementtypes.SupportsMimeTypes
    static supportsMimeType(mimeType) {
        return [TipViewer.MIME_TYPE].indexOf(mimeType) !== -1;
    }
    connectedCallback() {
        super.connectedCallback();
        this._tipIndex = this._configManager.getConfig().tipCounter % this._getTipCount();
        if (DomUtils.getShadowRoot(this) !== null) {
            return;
        }
        const shadow = this.attachShadow({ mode: 'open', delegatesFocus: false });
        const clone = this.createClone();
        shadow.appendChild(clone);
        this.updateThemeCss();
        const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
        // Intercept link clicks and open them in an external browser.
        containerDiv.addEventListener('click', (ev) => {
            const source = ev.target;
            if (source.tagName === "A") {
                ev.preventDefault();
                shell.openExternal(source.href);
            }
        });
        containerDiv.addEventListener('focus', (ev) => {
            if (ev.target instanceof HTMLSelectElement) {
                ev.stopPropagation();
                return;
            }
        }, true);
        this._setTipHTML(this._getTipHTML(this._tipIndex));
        const nextButton = DomUtils.getShadowId(this, ID_NEXT_BUTTON);
        nextButton.addEventListener('click', () => {
            this._tipIndex = (this._tipIndex + 1) % this._getTipCount();
            this._setTipHTML(this._getTipHTML(this._tipIndex));
        });
        const previousButton = DomUtils.getShadowId(this, ID_PREVIOUS_BUTTON);
        previousButton.addEventListener('click', () => {
            this._tipIndex = (this._tipIndex + this._getTipCount() - 1) % this._getTipCount();
            this._setTipHTML(this._getTipHTML(this._tipIndex));
        });
        const showTipsSelect = DomUtils.getShadowId(this, ID_SHOW_TIPS);
        showTipsSelect.value = this._configManager.getConfig().showTips;
        showTipsSelect.addEventListener('change', () => {
            const newConfig = _.cloneDeep(this._configManager.getConfig());
            newConfig.showTips = showTipsSelect.value;
            this._configManager.setConfig(newConfig);
        });
    }
    /**
     * Custom Element 'disconnected' life cycle hook.
     */
    disconnectedCallback() {
        super.disconnectedCallback();
        if (this._configManager !== null) {
            this._configManager.unregisterChangeListener(this);
        }
    }
    _themeCssFiles() {
        return [ThemeTypes.CssFile.TIP_VIEWER, ThemeTypes.CssFile.FONT_AWESOME, ThemeTypes.CssFile.GUI_CONTROLS];
    }
    refresh(level) {
        this._processRefresh(level);
    }
    //-----------------------------------------------------------------------
    //
    // ######                                      
    // #     # #####  # #    #   ##   ##### ###### 
    // #     # #    # # #    #  #  #    #   #      
    // ######  #    # # #    # #    #   #   #####  
    // #       #####  # #    # ######   #   #      
    // #       #   #  #  #  #  #    #   #   #      
    // #       #    # #   ##   #    #   #   ###### 
    //
    //-----------------------------------------------------------------------
    /**
     *
     */
    createClone() {
        let template = window.document.getElementById(ID);
        if (template === null) {
            template = window.document.createElement('template');
            template.id = ID;
            template.innerHTML = `<style id="${ThemeableElementBase_1.ThemeableElementBase.ID_THEME}">
        </style>
        <div id="${ID_CONTAINER}" class="container-fluid">
          <div id="${ID_CONTENT}"></div>
          <div id="${ID_CONTROLS}" class="form-inline">
            <div class="btn-group">
              <button id="${ID_PREVIOUS_BUTTON}" title="Previous Tip" class="btn btn-default btn-sm"><i class="fa fa-chevron-left" aria-hidden="true"></i></button>
              <button id="${ID_NEXT_BUTTON}" title="Next Tip" class="btn btn-default btn-sm"><i class="fa fa-chevron-right" aria-hidden="true"></i></button>
            </div>
            <div class="form-group form-group-sm">
              <label for="">Show tips: </label>
              <select id="${ID_SHOW_TIPS}" class="form-control">
                <option value="always">Everytime</option>
                <option value="daily">Daily</option>
                <option value="never">Never</option>
              </select>
            </div>
          </div>
        </div>`;
            window.document.body.appendChild(template);
        }
        return window.document.importNode(template.content, true);
    }
    _configChanged() {
        const showTipsSelect = DomUtils.getShadowId(this, ID_SHOW_TIPS);
        showTipsSelect.value = this._configManager.getConfig().showTips;
    }
    _keyBindingChanged() {
        this._setTipHTML(this._getTipHTML(this._tipIndex));
    }
    _setTipHTML(html) {
        const contentDiv = DomUtils.getShadowId(this, ID_CONTENT);
        contentDiv.innerHTML = html;
        this._substituteKeycaps(contentDiv);
        this._fixImgRelativeUrls(contentDiv);
        this._processRefresh(ResizeRefreshElementBase.RefreshLevel.RESIZE);
    }
    _substituteKeycaps(contentDiv) {
        // Replace the kbd elements with the requested keyboard short cuts.
        const kbdElements = contentDiv.querySelectorAll("span." + CLASS_KEYCAP);
        DomUtils.toArray(kbdElements).forEach((kbd) => {
            const dataContext = kbd.getAttribute("data-context");
            const dataCommand = kbd.getAttribute("data-command");
            if (dataContext !== null && dataCommand !== null) {
                const keyBindings = this._keyBindingManager.getKeyBindingContexts().context(dataContext);
                if (keyBindings != null) {
                    const shortcut = keyBindings.mapCommandToKeyBinding(dataCommand);
                    if (shortcut !== null) {
                        kbd.innerHTML = `<span>${he.encode(shortcut)}</span>`;
                    }
                    else {
                        kbd.parentNode.removeChild(kbd);
                    }
                }
            }
        });
    }
    _fixImgRelativeUrls(contentDiv) {
        const imgElements = contentDiv.querySelectorAll("img");
        const prefix = "file:///" + SourceDir.path + "/tips/";
        DomUtils.toArray(imgElements).forEach((element) => {
            const img = element;
            img.src = prefix + img.getAttribute("src");
        });
    }
    _processRefresh(level) {
        const containerDiv = DomUtils.getShadowId(this, ID_CONTAINER);
        if (containerDiv !== null) {
            // --- DOM Read ---
            const rect = containerDiv.getBoundingClientRect();
            this._height = rect.height;
            this._adjustHeight(this._height);
            VirtualScrollArea.emitResizeEvent(this);
        }
    }
    _getTipHTML(tipNumber) {
        return tipData[tipNumber];
    }
    _getTipCount() {
        return tipData.length;
    }
    _adjustHeight(newHeight) {
        this._height = newHeight;
        if (this.parentNode === null || DomUtils.getShadowRoot(this) === null) {
            return;
        }
        this.style.height = "" + newHeight + "px";
    }
}
TipViewer.TAG_NAME = "ET-TIP-VIEWER";
TipViewer.MIME_TYPE = "application/x-extraterm-tip";
exports.TipViewer = TipViewer;
//# sourceMappingURL=TipViewer.js.map